function math.clamp (x, a, b)
	return ((x < a) and a) or ((x > b) and b) or x
end

function math.round (a)
	if a >= 0 then
		return math.floor (a + 0.5)
	else
		return math.ceil (a - 0.5)
	end
end

function math.lerp (a, b, t)
	return a + t * (b - a)
end

function math.bezier_2 (t, p0, p1, p2)
	local inv_t = 1 - t
	return (inv_t * inv_t) * p0 + 2 * t * inv_t * p1 + t * t * p2
end

function math.bezier_3 (t, p0, p1, p2, p3)
	local inv_t = 1 - t
	return p0 * (inv_t * inv_t * inv_t) + p1 * 3 * t * (inv_t * inv_t) + p2 * 3 * (t * t) * inv_t + p3 * (t * t * t)
end

-- interpolation with piecewise linear function
-- xx - for a given set of values should be xx[i] < xx[i + 1]
function math.pwlinear (xx, yy)
	assert (#xx == #yy)

	local segments = { }
	local n = #xx
	for i = 1, n do
		segments[i] = { x = xx[i], y = yy[i] }
	end

	return setmetatable ({ segments = segments }, { __call = function (t, x)
		local segments = t.segments
		
		if x <= segments[1].x then
			return segments[1].y
		elseif x >= segments[n].x then
			return segments[n].y
		end

		local i, right = 1, n
		while (i < right) do
			local mid = i + math.floor ((right - i) / 2)
			if x <= segments[mid].x then
				right = mid
			else
				i = mid + 1
			end
		end 
		
		local s1, s0 = segments[i], segments[i - 1]
		return math.lerp (s0.y, s1.y, (x - s0.x) / (s1.x - s0.x))
	end})
end

-- cubic spline
-- xx - for a given set of values should be xx[i] < xx[i + 1]
function math.cspline (xx, yy)
	assert (#xx == #yy)

	local n = #xx

	local splines = { }
	for i = 1, n do
		splines[i] = { x = xx[i], a = yy[i] }
	end
	splines[1].c = 0

	local alpha = { 0 }
	local beta = { 0 }
	local A, B, C, F, h_i, h_i1, z
	for i = 2, n - 1 do
		h_i = xx[i] - xx[i - 1]
		h_i1 = xx[i + 1] - xx[i]
		A = h_i
		C = 2.0 * (h_i + h_i1)
		B = h_i1
		F = 6.0 * ((yy[i + 1]- yy[i]) / h_i1 - (yy[i] - yy[i - 1]) / h_i)
		z = (A * alpha[i - 1] + C)
		alpha[i] = -B / z
		beta[i] = (F - A * beta[i - 1]) / z
	end

	splines[n].c = (F - A * beta[n - 1]) / (C + A * alpha[n - 1])

	for i = n - 1, 2, -1 do
		splines[i].c = alpha[i] * splines[i + 1].c + beta[i]
	end

	for i = n, 2, -1 do
		local h_i = xx[i] - xx[i - 1]
		splines[i].d = (splines[i].c - splines[i - 1].c) / h_i
		splines[i].b = h_i * (2.0 * splines[i].c + splines[i - 1].c) / 6 + (yy[i] - yy[i - 1]) / h_i
	end

	return setmetatable ({ splines = splines }, { __call = function (t, x)
		local splines = t.splines
		local s
		if x <= splines[1].x then
			s = splines[2]
		elseif x >= splines[n].x then
			s = splines[n]
		else
			local i, j = 1, n
			while (i + 1 < j) do
				local k = i + math.floor ((j - i) / 2)
				if (x <= splines[k].x) then
					j = k
				else
					i = k
				end
			end
			s = splines[j]
		end

		local dx = x - s.x
		return s.a + (s.b + (s.c * 0.5 + s.d * dx / 6.0) * dx) * dx
	end})
end
