require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/enter_name_view.lua"
require "underworld2/shared.lua"
require "ui/modal/confirm_window.lua"

local view = setmetatable ({ }, { __index = ui.core.window })

local function _process_reset_game()
	ui.event_view.show()
	underworld2.network.new_game(function()
		ui.event_view.close()
		underworld2.openid_show_relogin (true)
		underworld2.inbox.clear()
		underworld2.full_reset()
	end)
end

local function _reset_profile_handler()
	local info_view = ui.confirm_window.create (L"KEY_ACCOUNT_SETTINGS_RESET_PROGRESS", L"KEY_ACCOUNT_SETTINGS_RESET_CONFIRMATION", L"KEY_ACCOUNT_SETTINGS_RESET_CONFIRM", L"KEY_ACCOUNT_SETTINGS_RESET_CANCEL")
	ui.core.view_controller:show_modal_view (info_view)
	info_view.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (info_view)
			if info_view.result == ui.confirm_window.RESULT_CONFIRM then
				_process_reset_game()
			end
		end
	}).build(info_view)
end

local function _send_feedback_handler()
	underworld2.shared.send_feedback()
end

local function _create_button (text, handler, tex_id)
	local hlayout = ui.HorizontalLayout()

	local btn = ui.TextButton()
	btn:setTexture (tex_id or "ui/settings/btn_green_normal:filtered")
	btn:setTextStyle (ui.text_styles.normal_hl.build())
	btn:setText (text)
	btn:setHorizontalSizePolicy (ui.SizePolicyFixed)
	btn:setWidth(0.5)
	events.connect (btn:getClickedSignalProxy(), handler)

	hlayout:addWidget (ui.Widget())
	hlayout:addWidget (btn)
	hlayout:addWidget (ui.Widget())

	hlayout:setVerticalPolicy (ui.SizePolicyMinimum)

	return hlayout
end

local function _create_visual_separator()
	local hlayout = ui.HorizontalLayout()

	local sep = ui.Picture()
	sep:setTexture("ui/settings/line_spacing_settings:filtered")
	sep:setHeight (sep:getPreferredSize().y)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	sep:setHorizontalSizePolicy (ui.SizePolicyFixed)
	sep:setWidth(0.5)

	hlayout:addWidget(ui.Widget())
	hlayout:addWidget(sep)
	hlayout:addWidget(ui.Widget())

	return hlayout
end

local function _create_l_r(l, r)
	local hlayout = ui.HorizontalLayout()

	local hgroup_wrapper = ui.Widget()
	hgroup_wrapper:setHorizontalSizePolicy(ui.SizePolicyFixed)
	hgroup_wrapper:setWidth(0.5)

	local hhlayout = ui.HorizontalLayout()
	hgroup_wrapper:setLayout(hhlayout)

	local h_shift = ui.Widget()
	h_shift:setHorizontalSizePolicy(ui.SizePolicyMaximum)

	hhlayout:addItem(l)
	hhlayout:addWidget(h_shift)
	hhlayout:addItem(r)

	hlayout:addWidget (ui.Widget())
	hlayout:addWidget (hgroup_wrapper)
	hlayout:addWidget (ui.Widget())

	return hlayout
end

local function _create_labeled_switch(text_on, text_off, img_on, img_off, cb)
	local vlayout = ui.VerticalLayout()
	vlayout:setItemSpacing (ui.config.vspacing)
	vlayout:setHorizontalPolicy(ui.SizePolicyMinimum)

	local default_switch_state = cb ()
	local default_text = default_switch_state and text_on or text_off
	local default_img = default_switch_state and img_on or img_off

	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.normal.build())
	title:setText (default_text)
	title:setTextAlignment (ui.AlignmentHorizontalCenter)
	title:setVerticalSizePolicy (ui.SizePolicyMinimum)

	local btn = ui.ImageButton()
	btn:setTexture(default_img)

	events.connect (btn:getClickedSignalProxy(), function()
		local st = cb(true)

		local txt = st and text_on or text_off
		local img = st and img_on or img_off

		title:setText(txt)
		btn:setTexture(img)
	end)

	vlayout:addWidget(title)
	vlayout:addWidget(btn)

	return vlayout	
end

local function _toggle_music(toggle)
	local st = underworld2.settings:get ("background_music_enabled", true)
	if not toggle then return st end

	if st then get_music_player():stop()
	else get_music_player():play()
	end

	underworld2.settings:set ("background_music_enabled", not st)
	underworld2.settings:flush ()

	return not st
end

local function _toggle_sound(toggle)
	local st = underworld2.settings:get ("sound_fx_enabled", true)
	if not toggle then return st end

	if st then get_sound_pool():mute()
	else get_sound_pool():unmute()
	end

	underworld2.settings:set ("sound_fx_enabled", not st)
	underworld2.settings:flush ()

	return not st
end

function view:build_ui( )

	local W = ui.Picture()
	W:setPlacingMode (ui.Picture_PlacingModeStretch)
	W:setSize (1, 1)
	W:setOrigin (0, 0)
	W:setTexture("ui/solid_background:filtered")
	W:setColor(render.color(35,39,45,255))

	local scroll = ui.ScrollWidget()
	scroll:setScrollType (ui.ScrollTypeVertical)
	scroll:setOrigin (0.0, underworld2.ui.status_panel:get_size().y)
	scroll:setSize (1.0, 1.0 - underworld2.ui.status_panel:get_size().y)
	scroll:setScrollBarTexture ("ui/scroll:filtered")
	W:addChild (scroll)

	local layout = ui.VerticalLayout()
	layout:setItemSpacing (ui.config.hspacing)

	--f, l = _create_group (L ("KEY_ACCOUNT_SETTINGS_ACCOUNT_ID", tostring (underworld2.player:id())))
	--layout:addWidget (f)

	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.normal.build())
	title:setText (L ("KEY_ACCOUNT_SETTINGS_ACCOUNT_ID", tostring (underworld2.player:id())))
	title:setTextAlignment (ui.AlignmentHorizontalCenter)
	title:setVerticalSizePolicy (ui.SizePolicyMinimum)

	layout:addWidget (ui.Widget())

	local hlayout = _create_l_r(
		 _create_labeled_switch(L"KEY_ACCOUNT_SETTINGS_SOUND_ON", L"KEY_ACCOUNT_SETTINGS_SOUND_OFF", "ui/settings/btn_sound:filtered", "ui/settings/btn_sound_off:filtered", _toggle_sound),
		 _create_labeled_switch(L"KEY_ACCOUNT_SETTINGS_MUSIC_ON", L"KEY_ACCOUNT_SETTINGS_MUSIC_OFF", "ui/settings/btn_music:filtered", "ui/settings/btn_music_off:filtered", _toggle_music)
	)

	layout:addItem (hlayout)

	layout:addWidget (title)

	layout:addItem (_create_button (L"KEY_ACCOUNT_SETTINGS_SEND_FEEDBACK", _send_feedback_handler))

	layout:addItem (_create_visual_separator())

	layout:addItem (_create_button (L "KEY_ACCOUNT_SETTINGS_RESET_PROGRESS", _reset_profile_handler, "ui/settings/btn_red_normal:filtered"))

	layout:addWidget (ui.Widget())

	scroll:setLayout (layout)

	return W
end

function view:title()
	return L("KEY_ACCOUNT_SETTINGS_WINDOW_NAME")
end

function view:icon()
end

ui.account_settings_view = view
