require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/widgets/upgrade_widget.lua"
require "ui/widgets/skill_widget.lua"
require "ui/modal/crew_member_confirmation.lua"
require "ui/modal/not_enough_cash.lua"
require "underworld2/timed_events.lua"
require "ui/modal/not_enough_gold.lua"

local view = setmetatable ({ }, { __index = ui.core.window })
local delivery_events = { }

local function show_add_to_crew (skill_id, level_id, event)
	local v = ui.core.view.builder (ui.crew_member_confirmation).build (skill_id, level_id)
	ui.core.view.delegate_builder (
	{
		finished = function (v)
			if event then event:finalize() end
			ui.core.view_controller:close_modal_view (v)
			ui.event_view.show()
			underworld2.network.equip_upgrade ( {skill_id = skill_id, level = level_id }, function (err, success)
				ui.event_view.close()
				if err ~= 0 or not success then
					get_sound_pool():play ("ui/UI_Error.wav")
					underworld2.ui.show_message (L "KEY_NETWORK_STATE_ERROR_SYNC")
					return
				end
				get_sound_pool():play ("ui/SFX_Cash_Full.wav")
				if underworld2.tutorial:current_stage() == tutorial.crew[9] then
					underworld2.tutorial:next_stage()
				end
			end)
		end
	}).build (v)
	ui.core.view_controller:show_modal_view (v)

	get_sound_pool():play ("ui/SFX_Notice.wav")
	return true
end

local function run_delivery_game_event (skill_id, level_id)
	local e = underworld2.crew_delivery_event.create (function (e)
		if ui.core.view_controller:has_modal_views() then
			return false
		end
		show_add_to_crew (skill_id, level_id, e)
		return true
	end)
	underworld2.game_events:queue (e)
end

local function create_delivery_timed_event (skill_id, level_id, when)
	
	local event_id = underworld2.timed_events:queue ("upgrade_delivery", when, function()
		delivery_events[skill_id] = nil
		run_delivery_game_event (skill_id, level_id)
	end)
	return event_id
end

function view.update_upgrades_delivery_queue()
	if not underworld2.timed_events:queue_exists ("upgrade_delivery") then
		underworld2.timed_events:create_queue ("upgrade_delivery")
	end
	underworld2.timed_events:clear_queue ("upgrade_delivery")
	delivery_events = { }

	local now = underworld2.server_time()
	local player = underworld2.player
	local timed_events = underworld2.timed_events
	
	local pending_skills = player:getPendingSkills()
	
	for i = 0, pending_skills:size() - 1 do
		local skill = pending_skills[i]
		local when = skill:getDeliveryEndTime()
		
		local skill_id = skill:getSkillId()
		local level_id = skill:getSkillLevelId()
		if when <= now then
			run_delivery_game_event (skill_id, level_id)
		else
			local event_id = create_delivery_timed_event (skill_id, level_id, when)
			delivery_events[skill_id] = event_id
		end
	end
end

function view:build_ui ()
	underworld2.timed_events:clear_queue ("upgrade_delivery")

	local root_widget = ui.Background()
	root_widget:setSize (1.0, 1.0)
	root_widget:setOrigin (0.0, 0.0)
	root_widget:setTexture("ui/bg_dark", false)

	local container = ui.Widget()
	container:setSize(1.0, 1.0 - underworld2.ui.status_panel:get_size().y)
	container:setOrigin(0.0, underworld2.ui.status_panel:get_size().y)
	container:setMargins (0, 0, ui.config.vspacing, 0)
	root_widget:addChild (container)

	local container_layout = ui.VerticalLayout()
	container_layout:setItemSpacing (ui.config.vspacing)

	local skills_tab = ui.CrewNavigationBar()
	skills_tab:setVerticalSizePolicy (ui.SizePolicyMinimum)
	local skills_layout = ui.HorizontalLayout()
	skills_layout:setItemSpacing (ui.config.hspacing)
	skills_tab:setLayout (skills_layout)
	
	local skill_group = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP")

	self.skills = 
	{
		ui.skill_widget.create (underworld2.player, skill_group:getSkill ("COURIERS"), 1),
		ui.skill_widget.create (underworld2.player, skill_group:getSkill ("ITEM"), 2),
		ui.skill_widget.create (underworld2.player, skill_group:getSkill ("WALLET"), 3)
	}
	
	skills_layout:addWidget (self.skills[1].widget)
	skills_layout:addWidget (self.skills[2].widget)
	skills_layout:addWidget (self.skills[3].widget)
	self.skills_tab = skills_tab

	self.skills[1].overlay:setHidden(true)
	
	do
		--do upscale if necessarry
		local sizes = { 
			self.skills[1].button:getPreferredSize(),
			self.skills[2].button:getPreferredSize(),
			self.skills[3].button:getPreferredSize()
		}

		local max_width = container:getWidth() - skills_tab:getLeftMargin() - skills_tab:getRightMargin() - ui.config.hspacing * 2
		local sum_width = sizes[1].x + sizes[2].x + sizes[3].x
		local scale_factor = max_width / sum_width
		
		local btn_handler = function (btn) 
			local index = btn:getTag()
			skills_tab:setFocusItemIndex (index - 1)
			local skill = self.skills[index].skill

			self:show_skill (skill)
			
			self.skills[index].notification_badge:setPendingNotifications(0)

			for i,v in ipairs(self.skills) do
				v.overlay:setHidden(i == index)
			end
		end
		
		for i, v in pairs (self.skills) do
			local function update_size( widget )
				widget:setHorizontalSizePolicy (ui.SizePolicyFixed)
				widget:setVerticalSizePolicy (ui.SizePolicyFixed)
				widget:setSize (sizes[i] * scale_factor)
			end

			update_size(v.button)
			update_size(v.overlay)
			update_size(v.widget)

			v.state.widget:setPlacingMode (ui.Picture_PlacingModeFill)
			
			events.connect (v.button:getClickedSignalProxy(), btn_handler)
			
		end
	end
	
	container_layout:addWidget (skills_tab)
	
	local scroll = ui.ScrollWidget()
	scroll:setScrollType (ui.ScrollTypeVertical)
	local scroll_layout = ui.VerticalLayout()
	scroll_layout:setItemSpacing (ui.config.vspacing)
	scroll:setLayout (scroll_layout)
	scroll:setScrollBarTexture ("ui/scroll:filtered")
	self.upgrades_scroll = scroll
	
	self.scroll_layout = scroll_layout
	
	container_layout:addWidget (scroll)
	
	container:setLayout (container_layout)
	
	self.rows = { }
	
	self.handlers = 
	{
		buy = function (btn)
			local level_id = btn:getTag()
			local wrapper = self.rows[level_id]
			local skill = wrapper.skill
			local skill_id = skill:getId()
			local level = skill:getLevel (level_id)
			
			local cost = level:getCost()
			
			if wrapper.state == ui.upgrade_widget.STATE_NOT_AVAILABLE then
				wrapper:harlem_shake()
				get_sound_pool():play ("ui/UI_Error.wav")
				return 
			end

			if underworld2.player:getMoney() < cost then
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_cash.show (cost - underworld2.player:getMoney(), false)
				return
			end
						
			ui.event_view.show()
			underworld2.network.buy_upgrade ( { skill_id = skill_id, level = level_id, cost = cost }, function (err, success)
				ui.event_view.close()
				if err ~= 0 or not success then
					underworld2.ui.show_message (L "KEY_NETWORK_STATE_ERROR_SYNC")
					get_sound_pool():play ("ui/UI_Error.wav")
				else
					local pending_skill = underworld2.player:getPendingSkill (skill_id, level_id)
					wrapper:set_pending_skill (pending_skill)
					wrapper:set_state (ui.upgrade_widget.STATE_IN_DELIVERY)
					if underworld2.tutorial:current_stage() == tutorial.crew[7] then
						underworld2.timed_events:queue (underworld2.server_time() + 0.5, function() underworld2.tutorial:next_stage() end)
					end
				end
			end)
		end,
		boost = function (btn)
			local level_id = btn:getTag()
			local wrapper = self.rows[level_id]
			
			local pending_skill = wrapper.pending_skill
			local price = pending_skill:getDeliveryPrice (underworld2.server_time())
			
			if underworld2.player:getGold() < price then
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_gold.show(price - underworld2.player:getGold())
				return
			end

			get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")

			ui.event_view.show()
			underworld2.network.boost ({ type = 3, id = pending_skill:getId(), gold = price }, function (success) 
				ui.event_view.close()
				local skill_id = pending_skill:getSkillId()
				local level_id = pending_skill:getSkillLevelId()

				if success then
					wrapper:set_state (ui.upgrade_widget.STATE_OWNED)
					if underworld2.tutorial:current_stage() == tutorial.crew[9] then
						show_add_to_crew (skill_id, level_id)
					else
						run_delivery_game_event (skill_id, level_id)
					end
				else
					underworld2.ui.show_message (L "KEY_NETWORK_STATE_ERROR_SYNC")
					get_sound_pool():play ("ui/UI_Error.wav")
				end
			end)
			
		end,

		-- obsolete: confirmation through modal
		confirm = function (btn) end
	}
	
	self:show_skill (skill_group:getSkill ("COURIERS"))
	self.skills[1].notification_badge:setPendingNotifications(0)
	
	self.player_stats_conn = events.connect (underworld2.player:getStatsChangedSignalProxy(), function (player)
		for i, v in pairs (self.skills) do
			v:refresh (player)
		end
	end)
	
	return root_widget
end

function view:set_selected_skill (index)
	self.skills_tab:setFocusItemIndex (index)
end

function view:show_skill (skill)
	local tinsert = table.insert
	local scroll_layout = self.scroll_layout
		
	for i, wrapper in pairs (self.rows) do
		scroll_layout:removeWidget (wrapper.widget)
	end
	self.rows = { }
	
	local player = underworld2.player
	
	local skill_id = skill:getId()
	local player_level = player:getLevel()
	local skill_level = player:getSkillLevel (skill_id)
	
	local new_items = underworld2.upgrades_tracker.new_items {
		level = player:getLevel(),
		category = "skills",
		subcategory = skill:getId()
	}

	for level_id = 1, skill:getLevelsCount() do	
		local level = skill:getLevel (level_id)
		if level:get() then
			local wrapper = ui.upgrade_widget.create (skill, level_id, self.handlers)
			local state = ui.upgrade_widget.STATE_NOT_AVAILABLE
			
			local pending_skill = player:getPendingSkill (skill_id, level_id)
			
			if skill_level >= level_id then
				state = ui.upgrade_widget.STATE_OWNED
			elseif player_level < level:getLevelRequired() then
				state = ui.upgrade_widget.STATE_NOT_AVAILABLE
			elseif pending_skill:get() then
				state = ui.upgrade_widget.STATE_IN_DELIVERY
				wrapper:set_pending_skill (pending_skill)
			else
				state = ui.upgrade_widget.STATE_AVAILABLE
			end
			wrapper:set_state (state, new_items[level_id])
			scroll_layout:addWidget (wrapper.widget)
			self.rows[level_id] = wrapper

			if state == ui.upgrade_widget.STATE_NOT_AVAILABLE then
				break
			end
		end
	end	

	underworld2.upgrades_tracker.visited_subcategory( underworld2.player, "skills", skill:getId() )

	self.upgrades_scroll:setRelativeScroll(math.vec2(0,0))
end

function view:update ()
	local server_time = underworld2.server_time()
	for i, v in pairs (self.rows) do
		v:update (server_time)
	end
end

function view:destroy()
	events.disconnect (underworld2.player:getStatsChangedSignalProxy(), self.player_stats_conn)
	view.update_upgrades_delivery_queue()
end

function view:title()
	return L "KEY_CREW_TITLE"
end

function view:tag()
	return "ui.crew_view"
end

ui.crew_view = view
