require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/enter_name_view.lua"

local dev_view = setmetatable ({ }, { __index = ui.core.window })

local function _create_button (text)
	local button = ui.TextButton()
	button:setTextStyle (ui.text_styles.normal.build())
	button:setTexture ("ui/btn_neutral:filtered")
	button:setText (text)
	button:setTrigger (true)
end 

local function _display_metrics_handler (state)
	if state == 0 then
		underworld2.settings:set ("dev.debug.disable_metrics", true)
	elseif state == 1 then
		underworld2.settings:set ("dev.debug.disable_metrics", false)
	end
	underworld2.settings:flush()
end

_debug_show_touch_zones = 0

local function _touch_zones_handler (state)
	if state == 0 then
		ui.setDebugTouchZones (false)
	elseif state == 1 then
		ui.setDebugTouchZones (true)
	end 
	_debug_show_touch_zones = state
end

local function _hd_tiles_handler (state)
	local value = state == 1
	underworld2.settings:set ("dev.game.hd_tiles", value)
	underworld2.settings:flush()

	local info_view = ui.information_view.create ("In order to apply changes the game should be restarted.")
	ui.core.view_controller:show_modal_view (info_view)
	info_view.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (info_view)
		end
	}).build(info_view)
end

local function _font_geometry_handler (state)
end

local function _set_server_url (url)
	if url == underworld2.network.get_server_base_url() then
		return
	end
	underworld2.network.set_server_base_url (url)
	underworld2.network.reset_session()
	force_relogin = true
end

local function _tutorial_handler()
	underworld2.settings:set ("tutorial", { force_start_initial_tutorial = true })
	underworld2.settings:set ("tutorial.passive_trading_completed", false)
	underworld2.settings:flush()
end

local function _start_tutorial (btn)
	local tutorial_map = { "initial", "passive_trading", "crew" }
	underworld2.settings:set ("tutorial.selected_tutorial_name", tutorial_map[btn:getTag()])
	underworld2.settings:flush()
end

local function _process_reset_game()
	ui.event_view.show()
	underworld2.network.new_game(function() 
		ui.event_view.close()
		ui.core.view_controller:close_modal_view (wait_view)
		underworld2.settings:flush()
	end)
end

local function _check_name (name)
	return true
end

local function _process_set_name (name)
	if not _check_name (name) then
		_show_enter_name_dialog (name)
		return
	end

	ui.event_view.show()
	underworld2.network.change_name( { name = name },  function(success)
		ui.event_view.close()
		if not success then
			_show_enter_name_dialog (name)
		end
	end)
end

local function _change_nickname_handler()
	local default = underworld2.player:getNickname()
	local enter_name_view = ui.enter_name_view.create (default)
	enter_name_view.delegate_builder (
	{
		finished = function()
			local name = enter_name_view.name
			ui.core.view_controller:close_modal_view (enter_name_view)
			if name ~= default then
				_process_set_name (name)
			end
		end
	}).build (enter_name_view)
	ui.core.view_controller:show_modal_view (enter_name_view)
	get_sound_pool():play ("ui/SFX_Notice.wav")
end

local function _reset_profile_handler()
	local info_view = ui.information_view.create ("Are you sure you want to start a new game?", { ui.information_view.BUTTON_YES, ui.information_view.BUTTON_CANCEL })
	ui.core.view_controller:show_modal_view (info_view)
	info_view.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (info_view)
			if info_view.result == ui.information_view.BUTTON_YES then
				_process_reset_game()
			end
		end
	}).build(info_view)
	get_sound_pool():play ("ui/SFX_Notice.wav")
end

local function _tile_scale_handler (editor, text)
	local scale = tonumber (text)
	if scale and scale >= 1.0 then
		underworld2.settings:set ("dev.game.tile_scale", scale)
		underworld2.settings:flush()
		--editor:setText (tostring (scale))
		
		local info_view = ui.information_view.create ("In order to apply changes the game should be restarted.")
		ui.core.view_controller:show_modal_view (info_view)
		info_view.delegate_builder (
		{
			finished = function()
				ui.core.view_controller:close_modal_view (info_view)
			end
		}).build(info_view)
		get_sound_pool():play ("ui/SFX_Notice.wav")
	else
	end
end

local function _create_on_off_switch (text, modes, handler, default)
	local modes = modes or { "On", "Off" }
	default = default or 0

	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.normal.build())
	title:setText (text)

	local group = ui.ButtonGroup ("", true)

	local button_layout = ui.HorizontalLayout()
	button_layout:setItemSpacing (ui.config.hspacing)

	local on = ui.TextButton()
	on:setTextStyle (ui.text_styles.normal_hl.build())
	on:setTexture ("ui/btn_offers:filtered")
	on:setText (modes[1])
	on:setTag (1)
	on:setToggleButton (true)
	on:setButtonGroup (group)

	local off = ui.TextButton()
	off:setTextStyle (ui.text_styles.normal_hl.build())
	off:setTexture ("ui/btn_offers:filtered")
	off:setText (modes[2])
	off:setTag (0)
	off:setToggleButton (true)
	off:setButtonGroup (group)

	on:setToggled (default == 1)
	off:setToggled (default == 0)

	local __ = function (button) handler (button:getTag()) end

	events.connect (on:getClickedSignalProxy(), __)
	events.connect (off:getClickedSignalProxy(), __)

	local layout = ui.HorizontalLayout()
	layout:setItemSpacing (ui.config.hspacing)

	button_layout:addWidget (on)
	button_layout:addWidget (off)

	layout:addWidget (title)
	layout:addItem (button_layout)

	layout:setVerticalPolicy (ui.SizePolicyMinimum)

	return layout
end

local function _create_action (text, action_text, handler)
	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.normal.build())
	title:setText (text)

	local action = ui.TextButton()
	action:setTextStyle (ui.text_styles.normal_hl.build())
	action:setTexture ("ui/btn_offers:filtered")
	action:setText (action_text)

	events.connect (action:getClickedSignalProxy(), handler)

	local layout = ui.HorizontalLayout()
	layout:setItemSpacing (ui.config.hspacing)

	layout:addWidget (title)
	layout:addWidget (action)

	layout:setVerticalPolicy (ui.SizePolicyMinimum)

	return layout
end

local function _create_group (text)
	local frame = ui.Frame()
	frame:setBordersTexture ("ui/frame_grey")

	local vlayout = ui.VerticalLayout()
	vlayout:setItemSpacing (ui.config.vspacing)

	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.large.build())
	title:setText (text)
	title:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalRight)
	vlayout:addWidget (title)

	frame:setVerticalSizePolicy (ui.SizePolicyMinimum)
	frame:setLayout (vlayout)

	return frame, vlayout
end

local function _create_string (text)
	local label = ui.TextLabel ()
	label:setTextStyle (ui.text_styles.small.build())
	label:setText (text)
	return label
end

local function _create_button (text, handler, align, tag)
	local hlayout = ui.HorizontalLayout()
	align = align or 2
	if align == 2 or align == 3 then
		hlayout:addWidget (ui.Widget())
	end

	local btn = ui.TextButton()
	btn:setTexture ("ui/btn_lower_normal:filtered")
	btn:setTextStyle (ui.text_styles.normal_hl.build())
	btn:setText (text)
	btn:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	btn:setTag (tag or 0)
	events.connect (btn:getClickedSignalProxy(), handler)
	hlayout:addWidget (btn)

	if align == 1 or align == 2 then
		hlayout:addWidget (ui.Widget())
	end
	hlayout:setVerticalPolicy (ui.SizePolicyMinimum)

	return hlayout
end

local function _toggle_music()
	local enabled = underworld2.settings:get ("background_music_enabled", true)
	underworld2.settings:set ("background_music_enabled", not enabled)
	underworld2.settings:flush ()

	if get_music_player():isBackgroundMusicPlaying() then
		return
	end

	if enabled then
		get_music_player():stop()
	--else
		--get_music_player():resume()
	end
end

function dev_view:build_ui( )

	local W = ui.Background()
	W:setSize (1, 1)
	W:setOrigin (0, 0)
	W:setTexture("ui/card_bg", false)

	local scroll = ui.ScrollWidget()
	scroll:setScrollType (ui.ScrollTypeVertical)
	scroll:setOrigin (0.0, underworld2.ui.status_panel:get_size().y)
	scroll:setSize (1.0, 1.0 - underworld2.ui.status_panel:get_size().y)
	scroll:setScrollBarTexture ("ui/scroll:filtered")
	W:addChild (scroll)

	local layout = ui.VerticalLayout()
	layout:setItemSpacing (ui.config.vspacing)
	scroll:setLayout (layout)

	do
		local player = underworld2.player
		local tile_index = player:getTileInfo():getTileIndex()
		
		local f, l = _create_group ("Profile data")
		layout:addWidget (f)

		l:addWidget (_create_string (string.format ("ID: %s", underworld2.player:id())))
		l:addWidget (_create_string (string.format ("Tile: %dx%d", tile_index.x, tile_index.y)))
		l:addItem (_create_button ("Reset progress", _reset_profile_handler))
	end

	do
		local f, l = _create_group ("Tutorial")
		layout:addWidget (f)

		l:addItem (_create_button ("Initital", _start_tutorial, 2, 1))
		l:addItem (_create_button ("Passive Trading", _start_tutorial, 2, 2))
		l:addItem (_create_button ("Crew", _start_tutorial, 2, 3))
	end


	do
		local f, l = _create_group ("Settings")
		layout:addWidget (f)

		l:addItem (_create_button ("Toggle music", _toggle_music))
	end

	do
		local f, l = _create_group ("Server")
		layout:addWidget (f)

		local url_string = _create_string (string.format ("URL: %s", underworld2.network.get_server_base_url()))
		l:addWidget (url_string)

		local button_layout = ui.HorizontalLayout()
		button_layout:setItemSpacing (ui.config.hspacing)

		local change = ui.TextButton()
		change:setTexture ("ui/btn_yellow")
		change:setTextStyle (ui.text_styles.normal_hl.build())
		change:setText ("Change")
		button_layout:addWidget (change)

		local default = ui.TextButton()
		default:setTexture ("ui/btn_yellow")
		default:setTextStyle (ui.text_styles.normal_hl.build())
		default:setText ("Default")
		button_layout:addWidget (default)

		local share = ui.TextButton()
		share:setTexture ("ui/btn_yellow")
		share:setTextStyle (ui.text_styles.normal_hl.build())
		share:setText ("Test share")
		button_layout:addWidget (share)

		events.connect (share:getClickedSignalProxy(), function() underworld2.network.share_reward {
				callback = function() 
					ui.event_view.close()
					--handler(true) 
				end,
				type = "crew",
				upgradeId = 37,
				level = 1
			}
		end)

		l:addItem (button_layout)


		events.connect (change:getClickedSignalProxy(), function()
			get_sound_pool():play ("ui/SFX_Runner_Complete.wav")
			local url = s3eOSReadStringUTF8WithDefault ("url:", underworld2.network.get_server_base_url())
			if url then
				url_string:setText (string.format ("URL: %s", url))
				_set_server_url (url)
			end
		end)

		events.connect (default:getClickedSignalProxy(), function()
			local url = underworld2.network.get_default_server_base_url()
			url_string:setText (string.format ("URL: %s", url))
			_set_server_url (url)
		end)

	end
	
	do
		local f, l = _create_group ("Map Tiles")
		layout:addWidget (f)
		
		local hlayout = ui.HorizontalLayout()
		
		local label = ui.TextLabel()
		label:setTextStyle (ui.text_styles.small.build())
		label:setText ("Scale")
		hlayout:addWidget (label)
		
		local editor = ui.LineEdit()
		editor:setTextStyle (ui.text_styles.normal_hl.build())
		editor:setText (tostring (underworld2.settings:get ("dev.game.tile_scale", 2.0)))
		editor:setInputFormat (ui.InputFormatNumber)
		editor:setTextFieldTexture ("ui/messaging/btn_send_dark:filtered")
		events.connect (editor:getEditingFinishedSignalProxy(), _tile_scale_handler)
		
		hlayout:addWidget (editor)
		l:addItem (hlayout)
	end
	
	return W
end

function dev_view:destroy()
end

function dev_view:title()
	return string.format ("Developer Settings")
end

function dev_view:icon()
end

ui.dev_view = ui.dev_view or dev_view
