
require "ui/core/view.lua"
require "ui/event_view.lua"
require "ui/text_styles.lua"

local view = setmetatable ({ name = "" }, { __index = ui.core.view})

local builder = ui.core.view.builder (view)

function view.create (default_name, change_name_routine)
	local v = builder.build (default_name, change_name_routine)
	return v
end

function view:back()
	self:finalize()
end

function view.show (default_name, change_name_routine)
	local instance = view.create (default_name, change_name_routine)
	instance.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (instance)
		end
	}).build (instance)
	ui.core.view_controller:show_modal_view (instance)
end 

local function __validate_name (name)
	if not name or #name == 0 then
		return false, L "KEY_ENTER_NAME_ERROR_EMPTY"
	elseif #name > 15 then
		return false, L "KEY_ENTER_NAME_ERROR_TOO_LONG"
	elseif string.find (name, "%c") then
		return false, L "KEY_ENTER_NAME_ERROR_BR_OR_TAB"
	elseif string.sub (name, 1, 1) == " " then
		return false, L "KEY_ENTER_NAME_ERROR_LEADING_SPACE"
	elseif string.sub (name, -1) == " " then
		return false, L "KEY_ENTER_NAME_ERROR_TRAILING_SPACE"
	end
	return true
end

function __update_geometry (self)
	local widget = self.ui.widget
	local height = widget:getPreferredSize().y
	widget:setSize (1, height)
	widget:setOrigin (0, (1 - height) * 0.5)
end

local function __try_change_name (self, default_name)
	local new_name = s3eOSReadStringUTF8WithDefault ("", default_name, 0)
	if new_name == nil then
		return false
	end
	local is_valid, message = __validate_name (new_name)
	if not is_valid then
		self.ui.error:setText (message)
		return false
	end
	self.name = new_name
	self.ui.error:setText ("")
	self.ui.text:setText (new_name)
	return true
end

local function __switch_to_confirmation (self)
	self.ui.title:setText ("Confirm Your New Name")
	self.ui.hint:setText ("Your new name is:")
	self.ui.new_name_pane:setHidden (true)
	self.ui.confirm_name_pane:setHidden (false)
end

local function __switch_to_updated (self)
	self.ui.title:setText ("Name Has Been Updated")
	self.ui.new_name_pane:setHidden (true)
	self.ui.confirm_name_pane:setHidden (true)
	self.ui.separator:setHidden (true)
	__update_geometry (self)
	get_sound_pool():play ("ui/SFX_Notice.wav")
end

local function __default_change_name_routine (self)
	ui.event_view.show()
	underworld2.network.push_back 
	{
		url = "/api/changeprofile.do",
		body = 
		{
			name = "request",
			children = 
			{
				{ name = "nickname", text = self.name }
			}
		},

		on_completed = function (result)
			self:set_completed()
			ui.event_view.close()
		end
	}
end

function view:build_ui (default_name, change_name_routine)
	local widget = ui.load_ui ("ui/change_name.ui")
	
	change_name_routine = change_name_routine or __default_change_name_routine
	
	self.default_name = default_name
	self.name = default_name or ""


	self.ui =
	{
		widget				= widget,
		title 				= runtime.cast_to (widget:findChild ("title"), "ui::TextLabel"),
		cancel 				= runtime.cast_to (widget:findChild ("cancel"), "ui::Button"),
		hint 				= runtime.cast_to (widget:findChild ("hint"), "ui::TextLabel"),
		text 				= runtime.cast_to (widget:findChild ("text"), "ui::TextLabel"),
		new_name_pane 		= widget:findChild ("new_name_pane"),
		separator			= widget:findChild ("separator"),
		confirm_name_pane 	= widget:findChild ("confirm_name_pane"),
		new 				= runtime.cast_to (widget:findChild ("new"), "ui::Button"),
		confirm 			= runtime.cast_to (widget:findChild ("confirm"), "ui::Button"),
		change 				= runtime.cast_to (widget:findChild ("change"), "ui::Button"),
		error 				= runtime.cast_to (widget:findChild ("error"), "ui::TextLabel")
	}

	widget:setMargins (ui.config.hspacing * 3, ui.config.hspacing * 3, ui.config.vspacing * 2, ui.config.vspacing * 2)


	self.ui.confirm_name_pane:setHidden (true)
	self.ui.text:setText (self.name)

	events.connect (self.ui.new:getClickedSignalProxy(), function() 
		if __try_change_name (self, self.name) then
			__switch_to_confirmation (self)
		end 
	end)

	events.connect (self.ui.change:getClickedSignalProxy(), function()
		__try_change_name (self, self.name)
	end)

	events.connect (self.ui.confirm:getClickedSignalProxy(), function()
		change_name_routine (self)
	end)

	events.connect (self.ui.cancel:getClickedSignalProxy(), function()
		self.name = self.default_name
		self:finalize()
	end)

	__update_geometry (self)

	return widget
end

function view:set_failed (message)
	self.ui.widget:setHidden (false)
	self.ui.error:setText (message or "unknown error")
end

function view:set_completed()
	self.ui.widget:setHidden (false)
	self.default_name = name
	__switch_to_updated (self)
end

ui.enter_name_view = ui.enter_name_view or view
