require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/widgets/item_widget.lua"
require "ui/modal/item_sell_confirmation.lua"
require "ui/modal/item_delivery_confirmation.lua"
require "underworld2/timed_events.lua"

require "ui/modal/not_enough_cash.lua"
require "ui/modal/not_enough_gold.lua"

require "underworld2/equipment_view_state.lua"
require "underworld2/upgrades_tracker.lua"

local view = setmetatable ({ }, { __index = ui.core.window })

local function create_slot (player, slot_id, btn_group, handler, width)
	local slot = underworld2.SlotsDatabase():get (slot_id)
	local item = player:getEquipedItem (slot_id)
	local widget = ui.ItemSlotWidget (slot)
	widget:setTag (slot_id)
	widget:setButtonGroup (btn_group)
	events.connect (widget:getClickedSignalProxy(), handler)
	if width then
		widget:setWidth (width)
		widget:setHorizontalSizePolicy (ui.SizePolicyFixed)
	end
	local locked = slot:getUnlockLevel() > player:getLevel()
	widget:setToggleButton (true)
	widget:setLocked (locked)
	widget:setNeedsDisplayTypeName(true)

	if locked then
		widget:setSound ("ui/UI_Error.wav")
	else
		widget:setSound ("ui/UI_Click_05_Light.wav")
	end

	do
		local hhlayout = ui.HorizontalLayout()
		widget:setLayout(hhlayout)

		local waiting_widget = ui.WaitingWidget()
		waiting_widget:setHorizontalSizePolicy (ui.SizePolicyMinimum)

		hhlayout:addWidget(ui.Widget())
		hhlayout:addWidget(waiting_widget)
		hhlayout:addWidget(ui.Widget())

		waiting_widget:setHidden(locked)

		events.connect(widget:getDownloadedSignalProxy(), function()
			waiting_widget:setHidden(true)
		end)
	end

	widget:setItem (item)

	return widget
end

function view:build_ui( equipment_view )
	self.equipment_view = equipment_view

	underworld2.timed_events:clear_queue ("item_delivery")
		
	local root_widget = ui.Background()
	root_widget:setSize (1.0, 1.0)
	root_widget:setOrigin (0.0, 0.0)
	root_widget:setTexture("ui/bg_dark", false)

	local container = ui.Widget()
	container:setSize(1.0, 1.0 - underworld2.ui.status_panel:get_size().y)
	container:setOrigin(0.0, underworld2.ui.status_panel:get_size().y)
	container:setMargins (0, 0, ui.config.vspacing, 0)
	root_widget:addChild (container)

	local root_layout = ui.VerticalLayout()
	root_layout:setItemSpacing (ui.config.vspacing)

	local player = underworld2.player

	local btn_group = ui.ButtonGroup ("slots_btn_group", true)

	local narrow_slot_width = (container:getWidth() - ui.config.hspacing * 4.0) * 0.25
	local wide_slot_width = (container:getWidth() - ui.config.hspacing * 4.0) - narrow_slot_width * 2.0

	self.slot_order_map = { 7, 8, 4, 5, 6, 2, 1, 3}
	self.slot_index_map = { }
		
	local slots_tab = ui.ItemSlotsTab()
	slots_tab:setHorizontalSizePolicy (ui.SizePolicyFixed)
	slots_tab:setWidth (1.0)
	local slots_layout = ui.HorizontalLayout()
	slots_layout:setItemSpacing (ui.config.hspacing)
	self.slots_tab = slots_tab
	events.connect (slots_tab:getItemChangedSignalProxy(), function (index)
		local slot_id = self.slot_index_map[index + 1]
		if slot_id then
			self:slot_selected (underworld2.player, slot_id)
		end
	end)
		
	local btn_handler = function (btn)
		self:set_selected_slot_id (underworld2.player, btn:getTag(), true)
	end
	
	local player_level = player:getLevel()
	self.slot_btns = { }
	
	for slot_id = 1,8 do
		local slot = underworld2.SlotsDatabase():get (slot_id)
		if slot and slot:getUnlockLevel() <= player_level then
			self.slot_btns[slot_id] = create_slot (player, slot_id, btn_group, btn_handler, wide_slot_width)
		end
	end
	
	for i, slot_id in ipairs (self.slot_order_map) do
		local slot_widget = self.slot_btns[slot_id]
		if slot_widget then
			slots_layout:addWidget (slot_widget)
			table.insert (self.slot_index_map, slot_id)
		end
	end
	
	slots_tab:setLayout (slots_layout)
	
	local slot_height = 0.0
	
	for i, v in pairs (self.slot_btns) do
		local h = v:getPreferredSize().y
		slot_height = math.max (slot_height, h)
	end
	
	slots_tab:setVerticalSizePolicy (ui.SizePolicyMinimum)

	local items_scroll = ui.ScrollWidget()
	items_scroll:setScrollType (ui.ScrollTypeVertical)
	local items_layout = ui.VerticalLayout()
	items_layout:setItemSpacing (ui.config.vspacing)
	items_scroll:setLayout (items_layout)
	items_scroll:setScrollBarTexture ("ui/scroll:filtered")

	self.items_layout = items_layout
	self.items_scroll = items_scroll

	self.rows = { }

	root_layout:addWidget (slots_tab)
	root_layout:addWidget (items_scroll)
	container:setLayout (root_layout)

	self.handlers = 
	{	
		buy = function (btn)
			local wrapper = self.rows[btn:getTag()]
			if not wrapper then return end

			local item 		= wrapper.item
			local item_id 	= item:getId()

			local player = underworld2.player

			if item:getLevelRequired() > player:getLevel() then
				-- some Harlem shaking here
				wrapper:harlem_shake()
				get_sound_pool():play ("ui/UI_Error.wav")
				return
			end

			if underworld2.player:getMoney() < item:getPrice() then
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_cash.show (item:getPrice() - underworld2.player:getMoney(), false)
				return
			end

			ui.event_view.show()
			underworld2.network.buy_item (item, function (result)
				ui.event_view.close()
				if not result then
					get_sound_pool():play ("ui/UI_Error.wav")
					underworld2.ui.show_message ("Synchronization error")
					return
				end
				local item = player:findInBackback (item_id)
				if item:get() then
					wrapper:set_delivery_time (player:getItemDeliveryTime (item_id), player:getItemDeliveryDuration (item_id))
					wrapper:set_state (ui.widgets.item_widget.STATE_IN_DELIVERY)
					underworld2.equipment_view_state.delivering.set_best_item_id (player, self.current_slot:getId(), item_id)
				end
				self:reload_page( player, self.current_slot )
			end)
		end,

		sell = function (btn)
			local wrapper = self.rows[btn:getTag()]
			if not wrapper then return end

			local item = wrapper.item
			local slot = wrapper.slot

			local v = ui.item_sell_confirmation.create (item)
			ui.core.view_controller:show_modal_view (v)
			v.delegate_builder (
			{
				finished = function (v)
					ui.core.view_controller:close_modal_view (v)
					if (v.result == ui.item_sell_confirmation.RESULT_SELL) then
						local player = underworld2.player
						wrapper:set_state (ui.widgets.item_widget.STATE_NOT_PURCHASED)

						get_sound_pool():play ("ui/SFX_Cash_Out.wav")

						ui.event_view.show()
						underworld2.network.sell_item (item:getId(), function (result)
							ui.event_view.close()
							player:removeItem (item:getId())
							wrapper:set_state (ui.widgets.item_widget.STATE_NOT_PURCHASED)
							local equiped_item = player:getEquipedItem (slot:getId())
							self.slot_btns[slot:getId()]:setItem (equiped_item)
							self:reload_page( player, self.current_slot )
						end)
					end
				end
			}).build (v)

			get_sound_pool():play ("ui/SFX_Notice.wav")
		end,

		equip = function (btn)
			local wrapper = self.rows[btn:getTag()]
			if not wrapper then return end

			local player = underworld2.player
			local item, slot = wrapper.item, wrapper.slot
			local item_id, slot_id = item:getId(), slot:getId()

			wrapper:set_state (ui.widgets.item_widget.STATE_EQUIPED)
			self.slot_btns[slot_id]:setItem (item)

			local equiped_item = player:getEquipedItem (slot_id)
			if equiped_item:get() then
				local equiped_item_id = equiped_item:getId()
				for i, v in pairs (self.rows) do
					if v.item:getId() == equiped_item_id then
						v:set_state (ui.widgets.item_widget.STATE_OWNED)
						break
					end
				end
			end
			player:equipItem (slot_id, item_id)

			ui.event_view.show()

			underworld2.network.equip_item (item_id, slot_id, function (result)
				get_sound_pool():play ("ui/SFX_Loot.wav")
				ui.event_view.close()				
				self:reload_page( player, self.current_slot )
				self:update_stats_diff()
			end)
		end,

		unequip = function (btn)
			local wrapper = self.rows[btn:getTag()]
			if not wrapper then return end

			local player 	= underworld2.player
			local slot 		= wrapper.slot
			local slot_id 	= slot:getId()

			wrapper:set_state (ui.widgets.item_widget.STATE_OWNED)
			self.slot_btns[slot_id]:setItem (underworld2.ItemPtr())
			
			player:unequipItem (slot_id)

			self:update_stats_diff()
			
			ui.event_view.show()
			underworld2.network.unequip_item (slot_id, function (result)
				ui.event_view.close()
				self:reload_page( player, self.current_slot )
			end)
		end,

		equip_now = function (btn)
			local wrapper = self.rows[btn:getTag()]
			if not wrapper then return end

			local item 		= wrapper.item
			local item_id 	= item:getId()

			local player = underworld2.player
			local price = player:getItemDeliveryPrice (item_id, underworld2.server_time())
			
			if price > player:getGold() then
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_gold.show(price - player:getGold())
				return
			end
			
			get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")
			ui.event_view.show()

			underworld2.network.boost ({ type = 2, gold = price, id = item_id}, function (result)
				get_sound_pool():play ("ui/SFX_Loot.wav")
				ui.event_view.close()
				if not result then
					get_sound_pool():play ("ui/UI_Error.wav")
					underworld2.ui.show_message ("Synchronization error")
					return
				end
				player:setItemDelivered (item_id)
				wrapper:set_state (ui.widgets.item_widget.STATE_OWNED)
				self:reload_page( player, self.current_slot )

				underworld2.equipment_view_state.delivering.mark_item_boosted(player, item_id)
			end)
		end
	}
	
	return root_widget
end

function view.update_item_delivery_queue()
	if not underworld2.timed_events:queue_exists ("item_delivery") then
		underworld2.timed_events:create_queue ("item_delivery", function()
			if ui.core.view_controller:has_modal_views() then
				return false
			elseif not underworld2.player then
				return false
			elseif #underworld2.player:getActiveTrigger() > 0 then
				return false
			end
			return true
		end)
	end
	underworld2.timed_events:clear_queue ("item_delivery")

	local now = underworld2.server_time()
	local player = underworld2.player
	local pending_items = player:getPendingItems()
	local timed_events = underworld2.timed_events
	for i = 0, pending_items:size() - 1 do
		local item_id = pending_items[i]
		local when = player:getItemDeliveryTime (item_id)
		if when > now then
			local event_id = timed_events:queue ("item_delivery", when, function()			
				local item = underworld2.ItemsDatabase():getItem (item_id)
				local v = ui.core.view.builder (ui.item_delivery_confirmation).build (item)
				ui.core.view.delegate_builder (
				{
					finished = function (v)
						ui.core.view_controller:close_modal_view (v)

						get_sound_pool():play ("ui/SFX_Cash_Full.wav")

						if v.result == ui.item_delivery_confirmation.RESULT_GOTO_SHOP then
							local slot		= underworld2.SlotsDatabase():getSlotByType (item:getSlotType())
							local equipment = ui.core.view.builder (ui.equipment_shop_view).build()
							equipment:set_selected_slot_id (underworld2.player, slot:getId())
							underworld2.ui.stack:push_window (equipment)
							underworld2.ui.status_panel:hide_profile_info(true)
							underworld2.equipment_view_state.delivering.reset_delivered(underworld2.player, {[slot:getId()] = false})
						end
					end
				}).build (v)
				ui.core.view_controller:show_modal_view (v)

				get_sound_pool():play ("ui/SFX_Notice.wav")
			end)
		end
	end
end

function view:destroy()
	if not underworld2.tutorial:current_stage() then
		view.update_item_delivery_queue()
	end
end

function view:update_stats_diff()
	local player = underworld2.player
	local equiped_item = player:getEquipedItem (self.slot:getId())
	for i, wrapper in pairs (self.rows) do
		wrapper:compare (equiped_item)
	end
end

function view:set_selected_slot_id (player, slot_id, animated)
	animated = animated or false
	for i, id in pairs (self.slot_index_map) do
		if slot_id == id then
			self.slots_tab:setFocusItemIndex (i - 1, animated)
			break
		end
	end
	self:slot_selected (player, slot_id)
end

function view:slot_selected (player, slot_id)
	local slot = underworld2.SlotsDatabase():get (slot_id)
	for i, id in pairs (self.slot_index_map) do
		local btn = self.slot_btns[id]
		if (id == slot_id) and btn then
			--error "[ui"
			btn:setColorClamp (1.0)
			btn:setToggled (true)
		elseif btn then
			btn:setColorClamp (0.5)
			btn:setToggled (false)
		end
	end
	self:reload_page (player, slot)
	self.current_slot = slot

	underworld2.equipment_view_state.delivering.reset_delivered(underworld2.player, {[slot_id] = false})
end

local function lex(...)
	local args = {...}
	if #args % 2 ~= 0 then
		error "Arguments count must be even"
	end 

	for i=1,#args,2 do
		local lhs = args[i]
		local rhs = args[i+1]

		if lhs < rhs then 
			return true
		elseif lhs > rhs then
			return false
		end
	end

	return false
end

function view:reload_page (player, slot, scroll_to)
	--do return end
	self.rows_cache = self.rows_cache or {}
	local tinsert = table.insert

	local build_widget = function (index, player, slot, item, handlers, tag, width)
		local cached = self.rows_cache[index]
		if cached then 
			cached:set_visible(true)
			return cached 
		end
		local wrapper = ui.widgets.item_widget.create (player, slot, item, handlers, tag, width)
		for i=1,index-1 do
			if not self.rows_cache[i] then
				self.rows_cache[i] = ui.widgets.item_widget.create (player, slot, item, handlers, tag, width)
				self.rows_cache[i]:set_visible(false)
			end 
		end
		self.rows_cache[index] = wrapper
		self.items_layout:addWidget (wrapper.widget)
		return wrapper
	end

	for i, v in pairs (self.rows_cache) do
		--self.items_layout:removeWidget (v.widget)
		v:set_visible(false)
	end

	self.rows = { }
	self.slot = slot
	
	local items_db = underworld2.ItemsDatabase()
	local level = player:getLevel()
	local cpp_db_items = items_db:select (slot:getType(), math.max(0,level - 3), level + 3)

	local now = underworld2.server_time()

	local items = {}
	local added = {}


	local slot_id 				= slot:getType()
	local equiped_item 			= player:getEquipedItem (slot:getId())
	local equiped_item_id		= equiped_item:get() and equiped_item:getId() 
	local equiped_item_found	= false

	if equiped_item_id then 
		tinsert( items, { id = equiped_item_id, item = equiped_item } )
		added[equiped_item_id] = true
		equiped_item_found = true
	end

	local pending_items = {}
	local backpack_items = {}

	local pending_items_native = player:getPendingItems()
	for i = 0, pending_items_native:size() - 1 do
		local item_id = pending_items_native[i]
		local item = items_db:getItem(item_id)

		delivery_time = player:getItemDeliveryTime (item_id)
		local delivered = (delivery_time - now) <= 0

		if not added[item_id] and item:get() and (item:getSlotType() == slot:getType()) then
			if not delivered then
				tinsert( pending_items, { id = item_id, item = item } )
			else
				if not player:isItemEquiped (item_id) then
					tinsert( backpack_items, { id = item_id, item = item } )
				end
				print("adding lost item " .. item_id)
			end

			added[item_id] = true
		end
	end

	table.sort( pending_items, function(a,b) return lex(a.item:getPrice(), b.item:getPrice(), a.id, b.id)  end )
	for i,v in ipairs( pending_items ) do tinsert(items, v) end

	local bp_items_count = player:getBackpackItemsCount()
	
	for i=0,bp_items_count-1 do
		local item = player:getBackpackItem(i)
		if item:get() and (item:getSlotType() == slot:getType()) then
			local item_id = item:getId()
			if not added[item_id] and not player:isItemEquiped (item_id) then
				tinsert(backpack_items, { id = item_id, item = item })
				added[item_id] = true
			end
		end
	end

	table.sort( backpack_items, function(a,b) return lex(b.item:getPrice(), a.item:getPrice(), a.id, b.id) end )
	for i,v in ipairs( backpack_items ) do tinsert(items, v) end

	local db_items = {}
	for i = 0, cpp_db_items:size() - 1 do
		local item_id   = cpp_db_items[i]
		if not added[item_id] and not player:isItemEquiped (item_id) then
			local item 		= items_db:getItem (item_id)

			if item:isAvailable() then
				tinsert( db_items, { id = item_id, item = item } )
			end
		end
	end

	table.sort( db_items, function(a,b) return lex(a.item:getLevelRequired(), b.item:getLevelRequired(),a.item:getPrice(), b.item:getPrice(), a.id, b.id) end )

	print("==========================================")
	for i,v in ipairs(db_items) do 
		
		print(v.item:getLevelRequired(), " ", v.item:getPrice(), " ", v.id)
		tinsert( items, v ) 
	end
	print("==========================================")

	local new_items = underworld2.upgrades_tracker.new_items {
		level = player:getLevel(),
		category = "items",
		subcategory = slot:getType()
	}

	for i,v in ipairs(items) do
		local item_id   = v.id
		local item 		= v.item

		print( i, " ", item_id, " ", item:getLevelRequired(), " ", item:getPrice(), " " )
		
		local delivery_time = 0
		local delivery_duration = 1
		local state = ui.widgets.item_widget.STATE_NOT_PURCHASED
		local available = true
		
		if item_id == equiped_item_id then
			state = ui.widgets.item_widget.STATE_EQUIPED
			equiped_item_found = true
		elseif player:isItemEquiped (item_id) then
			-- item equipped into another slot
			available = false
		elseif player:findInBackback (item_id):get() then
			delivery_time = player:getItemDeliveryTime (item_id)
			local delivered = (delivery_time - now) <= 0
			if delivered then
				state = ui.widgets.item_widget.STATE_OWNED
			else
				delivery_duration = underworld2.ItemsDeliveryTimeDatabase():get (item:getLevelRequired())
				state = ui.widgets.item_widget.STATE_IN_DELIVERY
			end
		end

		if available then
			print("Caching widget ", i)
			local wrapper = build_widget ( i, player, slot, item, self.handlers, item_id, 1.0 - self.items_scroll:getLeftMargin() - self.items_scroll:getRightMargin())
			wrapper:set_item (player, slot, item, item_id, new_items[item_id] or false)
			wrapper:set_state (state)
			wrapper:set_delivery_time (delivery_time, delivery_duration)
			wrapper:compare (equiped_item)
			self.rows[item_id] = wrapper
		end
	end

	underworld2.upgrades_tracker.visited_subcategory( player, "items", slot:getType() )
	if self.equipment_view then
		self.equipment_view:update_items_new_state()
	end
		
	self.items_scroll:setRelativeScroll (math.vec2 (0, 0))
end

function view:update()
	local now = underworld2.server_time()
	for i, wrapper in pairs (self.rows) do
		wrapper:update (now)
	end
end

function view:title()
	return L ("KEY_EQUIPMENT_VIEW_TITLE")
end

ui.equipment_shop_view = ui.equipment_shop_view or view