require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/equipment_shop_view.lua"
require "ui/modal/info_modal.lua"
require "ui/animations/factory.lua"
require "underworld2/equipment_view_state.lua"

local view = setmetatable ({ }, { __index = ui.core.window })
local view_state = underworld2.equipment_view_state

local function create_slot (player, slot_id, handler, width, delay)
	local slot = underworld2.SlotsDatabase():get (slot_id)

	if not slot:get() then
		error("Invalid slot: "..slot_id)
	end

	local item = player:getEquipedItem (slot_id)
	local widget = ui.ItemSlotWidget (slot)
	if width then
		widget:setWidth (width)
		widget:setHorizontalSizePolicy (ui.SizePolicyFixed)
	end

	local locked = slot:getUnlockLevel() > player:getLevel()
	widget:setTag (slot_id)
	events.connect (widget:getClickedSignalProxy(), handler)
	widget:setHasNewItems( underworld2.upgrades_tracker.has_new_items {
		level = player:getLevel(),
		category = "items",
		subcategory = slot:getType()
	})
	if locked then
		widget:setSound ("ui/UI_Error.wav")
	else
		widget:setSound ("ui/UI_Click_05_Light.wav")
		widget:setNeedsShowDeliveryBar(true)
		widget:setDeliveringItem(view_state.delivering.get_item_id(underworld2.player, slot_id))
	end

	do
		local hhlayout = ui.HorizontalLayout()
		widget:setLayout(hhlayout)

		local waiting_widget = ui.WaitingWidget()
		waiting_widget:setHorizontalSizePolicy (ui.SizePolicyMinimum)

		hhlayout:addWidget(ui.Widget())
		hhlayout:addWidget(waiting_widget)
		hhlayout:addWidget(ui.Widget())

		waiting_widget:setHidden(locked)

		events.connect(widget:getDownloadedSignalProxy(), function()
			waiting_widget:setHidden(true)
		end)
	end

	widget:setItem (item)


	local session = view_state.session
	if session.slot(slot_id) > 0 then
		widget:startUnlockAnimation(delay)
	else
		widget:setLocked (locked)
	end

	return widget
end

function build_parameter_animation(icon,value_label,title_label,lock_icon)
	local dark_color = render.color(100, 100, 100, 255)
	local greyed_color = render.color(60, 60, 60, 255)

	return ui.animations.build {
		from = 0,
		to = 2,
		duration = 0.8,

		step = function(self, value)
			local t = math.clamp(value, 0.0, 1.0)
			if icon then
				if value > 1.0 then
					icon:setScale(math.lerp(1.2,1.0,value-1.0))
				else
					icon:setScale(math.lerp(1.0,1.2,t))
				end
			end
			if icon then
				icon:setColor(render.color(
					math.lerp(dark_color.r, 255, t),
					math.lerp(dark_color.g, 255, t),
					math.lerp(dark_color.b, 255, t),
					dark_color.a))
				icon:setSaturation(t)
			end
			if value_label then
				value_label:setHidden(false)
				
				if value > 1.0 then
					value_label:setColor(render.color(255,255,255,255*(value-1)))
				else
					value_label:setColor(render.color(255,255,255,0))
				end
			end
			if title_label then
				title_label:setColor(render.color(
					math.lerp(greyed_color.r, 255, t),
					math.lerp(greyed_color.g, 255, t),
					math.lerp(greyed_color.b, 255, t),
					greyed_color.a))
			end
			if lock_icon then
				lock_icon:setHidden(false)

				local lock_icon_offset
				if value > 1.0 then
					lock_icon_offset = math.lerp(-0.01, 0.02, value-1.0)
					lock_icon:setColor(render.color(255,255,255,255*(2-value)))
				else
					lock_icon_offset = math.lerp(0,-0.01,value)
					lock_icon:setColor(render.color(255,255,255,math.lerp(200,255,value)))
				end
				
				lock_icon:setOffset(math.vec2(0,lock_icon_offset))
			end
		end
	}
end

function create_stat_row (icon_id, title, value)
	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing (ui.config.hspacing)

	local icon = ui.Picture()
	icon:setTexture (icon_id)
	hlayout:addWidget (icon)

	local title_label = ui.TextLabel()
	title_label:setText (title)
	title_label:setTextStyle (ui.text_styles.xsmall_hl.build())
	title_label:setTextAlignment (ui.AlignmentHorizontalLeft + ui.AlignmentVerticalCenter)
	title_label:setHorizontalSizePolicy (ui.SizePolicyMaximum)
	hlayout:addWidget (title_label)

	local value_label = ui.TextLabel()
	value_label:setText (tostring (value))
	value_label:setTextStyle (ui.text_styles.xsmall_hl.build())
	value_label:setTextAlignment (ui.AlignmentHorizontalCenter + ui.AlignmentVerticalCenter)
	value_label:setWordWrap (false)
	hlayout:addWidget (value_label)

	hlayout:addWidget (ui.Widget())

	return hlayout, value_label
end

function create_commodity_row (icon_id, value, locked)
	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing (ui.config.hspacing)

	local icon = ui.Picture()
	icon:setTexture (icon_id)
	icon:setHorizontalSizePolicy (ui.SizePolicyMinimum)

	hlayout:addWidget (icon)

	local value_label = ui.TextLabel()
	if value == 0 then
		value_label:setText ('-')
	else
		value_label:setText (tostring(value))
	end
	value_label:setTextStyle (ui.text_styles.xsmall_hl.build())
	value_label:setTextAlignment (ui.AlignmentHorizontalLeft + ui.AlignmentVerticalCenter)
	value_label:setWordWrap (false)
	value_label:setSize (value_label:getPreferredSize())

	local container = ui.Widget()

	local lock_icon = ui.Picture()
	lock_icon:setTexture ("ui/icons/ico_item_lock_small:filtered")
	lock_icon:setHorizontalSizePolicy (ui.SizePolicyFixed)
	lock_icon:setVerticalSizePolicy (ui.SizePolicyFixed)
	lock_icon:setSize (lock_icon:getPreferredSize())
	lock_icon:setColor (render.color(255,255,255,200))

	container:setHorizontalSizePolicy (ui.SizePolicyFixed)
	container:setVerticalSizePolicy (ui.SizePolicyFixed)
	container:setSize (lock_icon:getSize())

	container:addChild (value_label)
	container:addChild (lock_icon)

	--value_label:setOrigin(math.vec2(container:getWidth()/2-value_label:getWidth()/2,0))

	if locked then
		value_label:setHidden(true)
		lock_icon:setHidden(false)
	else
		value_label:setHidden(false)
		lock_icon:setHidden(true)
	end

	hlayout:addWidget (container)

	local animation = build_parameter_animation(icon,value_label,nil,lock_icon)
	if locked then
		animation:step(0)
	end

	return hlayout, value_label, animation
end

local percents_string = function (value)
	--[[if value % 1 < 0.01 then
		return string.format ("%.0%%f", value)
	else
		return string.format ("%.1f", value)
	end]]
	if value == 0 then
		return '-'
	end

	return L( "KEY_EQUIPEMENT_PERCENT", value)
end

function view:build_ui()

	self.animations = ui.animations.factory {}
	self.statistics = {}
	self.ready = false

	local statistics = self.statistics

	underworld2.player:validateItemsIconCache()

	ui.preloadEffect ("grayscale")

	local root_widget = ui.Background()
	root_widget:setSize (1.0, 1.0)
	root_widget:setOrigin (0.0, 0.0)
	root_widget:setTexture("ui/bg_dark", false)

	local container = ui.Widget()
	container:setSize(1.0, 1.0 - underworld2.ui.status_panel:get_size().y)
	container:setOrigin(0.0, underworld2.ui.status_panel:get_size().y)
	container:setMargins (ui.config.hspacing, ui.config.hspacing, ui.config.vspacing, ui.config.vspacing)
	root_widget:addChild (container)

	local root_layout = ui.VerticalLayout()
	root_layout:setItemSpacing (ui.config.vspacing)

	local player = underworld2.player
	local stats = player:getStats()

	local function animate_player_widget(name, from, to)
		local difference = to - from
		local widget = self.stat_widgets[name]

		self.animations:bind(name .. "_animation", ui.animations.build {
			from = 0,
			to = 2,
			duration = 0.9,
			delay = 0.6,

			step = function(self, value)
				if value > 1 then
					widget:setText(tostring(math.floor(math.lerp(from,to,value-1))))
					widget:setZoom(1.0)
				else
					if difference > 0 then
						widget:setText("+" .. difference)
					else
						widget:setText(difference)
					end
					widget:setZoom(math.lerp(1.0,1.5,value))
				end
			end
		})

		widget:setText(tostring(from))
	end

	local slot_btn_handler = function (btn)
		local slot_id = btn:getTag()
		
		local slot = underworld2.SlotsDatabase():get (slot_id)
		if slot:getUnlockLevel() > underworld2.player:getLevel() then
			local slot_widget = runtime.cast_to (btn, "ui::ItemSlotWidget")
			slot_widget:startShakeAnimation()
		else
			local v = ui.core.view.builder (ui.equipment_shop_view).build( self )
			v:set_selected_slot_id (underworld2.player, slot_id)
			underworld2.ui.stack:push_window (v)
			view_state.delivering.reset_delivered(underworld2.player, {[slot_id] = self.slots[slot_id]})
		end
	end

	local narrow_slot_width = (container:getWidth() - ui.config.hspacing * 4.0) * 0.25
	
	local session = view_state.session
	local delay = function(slot_id)
		local index = session.slot(slot_id)
		if index > 0 then return (index-1) * 1/3 end
		return 0
	end

	local skill_group = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP")
	local skill = skill_group:getSkill ("ITEM")
	local skill_level = underworld2.player:getSkillLevel (skill:getId())
	
	-- 1-st and 2-nd commodities are available by default
	local unlocked_commodities = { 1, 2 }
	for i = 1, skill_level do
		table.insert (unlocked_commodities, i + 1)
	end

	local function is_commodity_unlocked(commodity_id)
		for i = 1, #unlocked_commodities do
			if unlocked_commodities[i] == commodity_id then
				return true
			end
		end

		return false
	end

	self.slots = {
		[7] = create_slot (player, 7, slot_btn_handler, narrow_slot_width, delay(7)),
		[2] = create_slot (player, 2, slot_btn_handler, nil, delay(2)),
		[5] = create_slot (player, 5, slot_btn_handler, narrow_slot_width, delay(5)),

		[8] = create_slot (player, 8, slot_btn_handler, narrow_slot_width, delay(8)),
		[4] = create_slot (player, 4, slot_btn_handler, nil, delay(4)),
		[6] = create_slot (player, 6, slot_btn_handler, narrow_slot_width, delay(6)),

		[1] = create_slot (player, 1, slot_btn_handler, nil, delay(1)),
		[3] = create_slot (player, 3, slot_btn_handler, nil, delay(3))
	}

	do
		local row = ui.Widget()
		local layout = ui.HorizontalLayout()
		layout:setItemSpacing (ui.config.hspacing)
		layout:addWidget (self.slots[7])
		layout:addWidget (self.slots[2])
		layout:addWidget (self.slots[5])
		row:setLayout (layout)
		root_layout:addWidget (row)
	end

	do
		local row = ui.Widget()
		local layout = ui.HorizontalLayout()
		layout:setItemSpacing (ui.config.hspacing)
		layout:addWidget (self.slots[8])
		layout:addWidget (self.slots[4])
		layout:addWidget (self.slots[6])
		row:setLayout (layout)
		root_layout:addWidget (row)
	end

	do
		local row = ui.Widget()
		local layout = ui.HorizontalLayout()
		layout:setItemSpacing (ui.config.hspacing)
		layout:addWidget (self.slots[1])
		layout:addWidget (self.slots[3])
		row:setLayout (layout)
		root_layout:addWidget (row)
	end
	do

		local stat_widgets =
		{
			safety_bonus = { },
			lab_sales = { },
		}
		self.stat_widgets = stat_widgets

		local hlayout = ui.HorizontalLayout()
		hlayout:setVerticalPolicy (ui.SizePolicyMinimum)
		hlayout:setItemSpacing (ui.config.hspacing)

		do
			local frame 	= ui.Frame()
			frame:setBordersTexture ("ui/equipment/frame_nav_darken:filtered")
			local layout 	= ui.VerticalLayout()
			layout:setItemSpacing (ui.config.vspacing)

			do
				local hhlayout = ui.HorizontalLayout()

				local title_label = ui.TextLabel()
				title_label:setTextStyle (ui.text_styles.xsmall_hl.build())
				title_label:setText (L"KEY_EQUIPEMENT_TITLE_YOUR_STATS")
				title_label:setTextAlignment (ui.AlignmentVerticalCenter)
				--title_label:setVerticalSizePolicy (ui.SizePolicyMinimum)
				title_label:setHorizontalSizePolicy (ui.SizePolicyMaximum)
				
				local info_btn = ui.ImageButton ()
				info_btn:setTexture ("ui/equipment/equimpent_improvement/btn_info_small:filtered")
				events.connect (info_btn:getClickedSignalProxy(), function()
					ui.info_modal.show(L"KEY_PARAM_MAIN_STATS_DESCRIPTION", L"KEY_PARAM_MAIN_STATS_TITLE")
				end)

				hhlayout:addWidget(title_label)
				hhlayout:addWidget(info_btn)

				layout:addItem (hhlayout)
			end

			local l
			l, stat_widgets.delivery_time = create_stat_row ("ui/equipment/ico_32_time:filtered", L"KEY_EQUIPEMENT_TITLE_DELIVERY_TIME", stats:getDeliveryTimeLevel())
			layout:addItem (l)

			l, stat_widgets.heat_drop = create_stat_row ("ui/equipment/ico_32_cooldown:filtered", L"KEY_EQUIPEMENT_TITLE_HEAT_COOLDOWN", stats:getHeatDropLevel())
			layout:addItem (l)

			l, stat_widgets.storage = create_stat_row ("ui/equipment/ico_32_storage:filtered", L"KEY_EQUIPEMENT_TITLE_STORAGE_SPACE", stats:getStorageLevel())
			layout:addItem (l)

			l, stat_widgets.heat_scale = create_stat_row ("ui/equipment/ico_32_heat:filtered", L"KEY_EQUIPEMENT_TITLE_HEAT_LIMIT", stats:getHeatScaleLevel())
			layout:addItem (l)

			frame:setLayout (layout)
			hlayout:addWidget (frame)
		end

		do
			local frame 	= ui.Frame()
			frame:setBordersTexture ("ui/equipment/frame_nav_darken:filtered")
			frame:setVerticalSizePolicy (ui.SizePolicyMinimum)
			local layout 	= ui.VerticalLayout()
			layout:setItemSpacing (ui.config.vspacing)

			do
				local title_layout = ui.HorizontalLayout()
				title_layout:setItemSpacing (ui.config.hspacing)

				--title_layout:addWidget (ui.Widget())

				local title_ico = ui.Picture()
				title_ico:setTexture ("ui/equipment/ico_32_discount:filtered")
				title_ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
				title_layout:addWidget (title_ico)

				local title_label = ui.TextLabel()
				title_label:setTextStyle (ui.text_styles.xsmall_hl.build())
				title_label:setText (L"KEY_EQUIPEMENT_TITLE_LAB_DISCOUNT")
				title_label:setTextAlignment (ui.AlignmentVerticalCenter)
				title_label:setHorizontalSizePolicy (ui.SizePolicyMaximum)
				title_label:setWordWrap (false)
				title_layout:addWidget (title_label)
				
				local info_btn = ui.ImageButton ()
				info_btn:setTexture ("ui/equipment/equimpent_improvement/btn_info_small:filtered")
				events.connect (info_btn:getClickedSignalProxy(), function()
					ui.info_modal.show(L"KEY_PARAM_DISCOUNT_DESCRIPTION", L"KEY_PARAM_DISCOUNT_TITLE")
				end)

				title_layout:addWidget (info_btn)

				layout:addItem (title_layout)
			end

			do

				local columns_layout = ui.HorizontalLayout()
				columns_layout:setItemSpacing (ui.config.hspacing)

				local column0_layout = ui.VerticalLayout()
				column0_layout:setHorizontalPolicy(ui.SizePolicyMaximum)
				column0_layout:setItemSpacing (ui.config.vspacing)

				local lab_sales = { }
				for i = 1, 8 do
					lab_sales[i] = stats:getLabSales (i) * 100
				end

				local anims = {}

				local l
				l, stat_widgets.lab_sales[1], anims[1] = create_commodity_row (ui.config.commodity_small_icons[1], lab_sales[1], not is_commodity_unlocked(1))
				column0_layout:addItem (l)

				l, stat_widgets.lab_sales[2], anims[2] = create_commodity_row (ui.config.commodity_small_icons[2], lab_sales[2], not is_commodity_unlocked(2))
				column0_layout:addItem (l)

				l, stat_widgets.lab_sales[3], anims[3] = create_commodity_row (ui.config.commodity_small_icons[3], lab_sales[3], not is_commodity_unlocked(3))
				column0_layout:addItem (l)

				l, stat_widgets.lab_sales[4], anims[4] = create_commodity_row (ui.config.commodity_small_icons[4], lab_sales[4], not is_commodity_unlocked(4))
				column0_layout:addItem (l)

				columns_layout:addItem (column0_layout)

				local column1_layout = ui.VerticalLayout()
				column1_layout:setItemSpacing (ui.config.vspacing)

				l, stat_widgets.lab_sales[5], anims[5] = create_commodity_row (ui.config.commodity_small_icons[5], lab_sales[5], not is_commodity_unlocked(5))
				column1_layout:addItem (l)

				l, stat_widgets.lab_sales[6], anims[6] = create_commodity_row (ui.config.commodity_small_icons[6], lab_sales[6], not is_commodity_unlocked(6))
				column1_layout:addItem (l)

				l, stat_widgets.lab_sales[7], anims[7] = create_commodity_row (ui.config.commodity_small_icons[7], lab_sales[7], not is_commodity_unlocked(7))
				column1_layout:addItem (l)

				l, stat_widgets.lab_sales[8], anims[8] = create_commodity_row (ui.config.commodity_small_icons[8], lab_sales[8], not is_commodity_unlocked(8))
				column1_layout:addItem (l)

				columns_layout:addItem (column1_layout)

				layout:addItem (columns_layout)

				for k, v in pairs(anims) do
					local index = view_state.session.skill(k)
					if index > 0 then
						v.delay = v.duration * (index-1)
						self.animations:bind("discount_row_" .. k, v)
					end
				end
			end

			frame:setLayout (layout)
			hlayout:addWidget (frame)

		end

		do
			local frame 	= ui.Frame()
			frame:setBordersTexture ("ui/equipment/frame_nav_darken:filtered")
			frame:setVerticalSizePolicy (ui.SizePolicyMinimum)
			local layout 	= ui.VerticalLayout()
			layout:setItemSpacing (ui.config.vspacing)


			do
				local title_layout = ui.HorizontalLayout()
				title_layout:setItemSpacing (ui.config.hspacing)

				--title_layout:addWidget (ui.Widget())

				local title_ico = ui.Picture()
				title_ico:setTexture ("ui/equipment/ico_32_police_protection32:filtered")
				title_ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
				title_layout:addWidget (title_ico)

				local title_label = ui.TextLabel()
				title_label:setTextStyle (ui.text_styles.xsmall_hl.build())
				title_label:setText (L"KEY_EQUIPEMENT_TITLE_COP_PROTECTION")
				title_label:setTextAlignment (ui.AlignmentVerticalCenter)
				title_label:setHorizontalSizePolicy (ui.SizePolicyMaximum)
				title_label:setWordWrap (false)
				title_layout:addWidget (title_label)
				
				local info_btn = ui.ImageButton ()
				info_btn:setTexture ("ui/equipment/equimpent_improvement/btn_info_small:filtered")
				events.connect (info_btn:getClickedSignalProxy(), function()
					ui.info_modal.show(L"KEY_PARAM_PROTECTION_DESCRIPTION", L"KEY_PARAM_PROTECTION_TITLE")
				end)

				title_layout:addWidget (info_btn)

				layout:addItem (title_layout)
			end

			do

				local columns_layout = ui.HorizontalLayout()
				columns_layout:setItemSpacing (ui.config.hspacing)

				local column0_layout = ui.VerticalLayout()
				column0_layout:setHorizontalPolicy(ui.SizePolicyMaximum)
				column0_layout:setItemSpacing (ui.config.vspacing)

				local safety_bonus = { }
				for i = 1, 8 do
					safety_bonus[i] = stats:getSafetyBonus (i) * 100
				end

				local anims = {}

				local l
				l, stat_widgets.safety_bonus[1], anims[1] = create_commodity_row (ui.config.commodity_small_icons[1], safety_bonus[1], not is_commodity_unlocked(1))
				column0_layout:addItem (l)

				l, stat_widgets.safety_bonus[2], anims[2] = create_commodity_row (ui.config.commodity_small_icons[2], safety_bonus[2], not is_commodity_unlocked(2))
				column0_layout:addItem (l)

				l, stat_widgets.safety_bonus[3], anims[3] = create_commodity_row (ui.config.commodity_small_icons[3], safety_bonus[3], not is_commodity_unlocked(3))
				column0_layout:addItem (l)

				l, stat_widgets.safety_bonus[4], anims[4] = create_commodity_row (ui.config.commodity_small_icons[4], safety_bonus[4], not is_commodity_unlocked(4))
				column0_layout:addItem (l)

				columns_layout:addItem (column0_layout)

				local column1_layout = ui.VerticalLayout()
				column1_layout:setItemSpacing (ui.config.vspacing)

				l, stat_widgets.safety_bonus[5], anims[5] = create_commodity_row (ui.config.commodity_small_icons[5], safety_bonus[5], not is_commodity_unlocked(5))
				column1_layout:addItem (l)

				l, stat_widgets.safety_bonus[6], anims[6] = create_commodity_row (ui.config.commodity_small_icons[6], safety_bonus[6], not is_commodity_unlocked(6))
				column1_layout:addItem (l)

				l, stat_widgets.safety_bonus[7], anims[7] = create_commodity_row (ui.config.commodity_small_icons[7], safety_bonus[7], not is_commodity_unlocked(7))
				column1_layout:addItem (l)

				l, stat_widgets.safety_bonus[8], anims[8] = create_commodity_row (ui.config.commodity_small_icons[8], safety_bonus[8], not is_commodity_unlocked(8))
				column1_layout:addItem (l)

				columns_layout:addItem (column1_layout)

				layout:addItem (columns_layout)

				for k, v in pairs(anims) do
					local index = view_state.session.skill(k)
					if index > 0 then
						v.delay = v.duration * (index-1)
						self.animations:bind("protection_row_" .. k, v)
					end
				end
			end

			frame:setLayout (layout)
			hlayout:addWidget (frame)
		end

		root_layout:addItem (hlayout)
	end

	container:setLayout (root_layout)
	local function update_stats(animate_slots)
		local player = underworld2.player
		for slot_id, slot_widget in pairs (self.slots) do
			local item = player:getEquipedItem (slot_id)
			if item:get() then
				local slot_item = slot_widget:getItem()
				if (not slot_item:get()) or (slot_item:getId() ~= item:getId()) then
					slot_widget:setItem (item)
					if animate_slots then
						slot_widget:startEquipAnimation()
					end
				end
			else
				slot_widget:setItem (item)
			end
		end
		local stats, stat_widgets = player:getStats(), self.stat_widgets

		stat_widgets.delivery_time:setText (tostring (stats:getDeliveryTimeLevel()))
		stat_widgets.heat_drop:setText (tostring (stats:getHeatDropLevel()))
		stat_widgets.storage:setText (tostring (stats:getStorageLevel()))
		stat_widgets.heat_scale:setText (tostring (stats:getHeatScaleLevel()))

		statistics.delivery_time = (stats:getDeliveryTimeLevel())
		statistics.heat_drop = (stats:getHeatDropLevel())
		statistics.storage = (stats:getStorageLevel())
		statistics.heat_scale = (stats:getHeatScaleLevel())

		for commodity_id, label in pairs (stat_widgets.lab_sales) do
			local value = stats:getLabSales (commodity_id) * 100
			label:setText (percents_string (value))
			label:setSize (label:getPreferredSize())
		end
		for commodity_id, label in pairs (stat_widgets.safety_bonus) do
			local value = stats:getSafetyBonus (commodity_id) * 100
			label:setText (percents_string (value))
			label:setSize (label:getPreferredSize())
		end
	end

	update_stats()
	self.equipment_signal_conn = events.connect (player:getEquipmentChangedSignalProxy(), function()
		-- update only in foreground
		self.update_statistics = function()
			local previous_statistics = {}
			for k, v in pairs(statistics) do
				previous_statistics[k] = v
			end

			update_stats(true)

			for k, v in pairs(statistics) do
				local difference = v - previous_statistics[k]
				if difference ~= 0 then
					animate_player_widget(k, previous_statistics[k], v)
				end
			end
		end
	end)

	view_state.delivering.update_slots(underworld2.player, self.slots)
	view_state.session.clear('slots')
	view_state.session.clear('skills')

	self.animations:update(0, true)

	return root_widget
end

function view:update_items_new_state()
	local level = underworld2.player:getLevel()
	for k, widget in pairs(self.slots) do
		local slot = underworld2.SlotsDatabase():get (widget:getTag())

		widget:setHasNewItems( underworld2.upgrades_tracker.has_new_items { 
			level = level, 
			category = "items",
			subcategory = slot:getType() 
		})
	end
end

function view:update(dt)
	if self.ready then
		-- update only in foreground
		if self.update_statistics then
			self.update_statistics()
			self.update_statistics = nil
		end

		self.animations:update(dt)
	end

	self.ready = true
end

function view:title()
	return L ("KEY_EQUIPMENT_VIEW_TITLE")
end

function view:destroy()
	local player = underworld2.player
	events.disconnect (player:getEquipmentChangedSignalProxy(), self.equipment_signal_conn)
	ui.removeEffectFromPreloadList ("grayscale")
end

function view:moved_to_foreground(from_bg)
	if from_bg then 
		view_state.delivering.update_slots(underworld2.player, self.slots)
		self.ready = false
	end
end

ui.equipment_view = ui.equipment_view or view