require "network/online_controller.lua"

require "util/recursive_lock.lua"
require "ui/core/view.lua"
require "ui/text_styles.lua"
require "ui/config.lua"
require "ui/format.lua"

require "ui/my_order_widget.lua"
require "ui/my_offer_widget.lua"

require "ui/information_view.lua"

require "ui/storage_alert_view.lua"
require "ui/deal_alert_view.lua"

require "ui/dealers_list_view.lua"

require "ui/modal/levelup.lua"
require "ui/modal/inbox.lua"
require "ui/modal/vault_warning.lua"

require "ui/modal/busted_modal.lua"
require "ui/modal/rent_modal.lua"

require "ui/modal/not_enough_cash.lua"
require "ui/modal/not_enough_gold.lua"

require "ui/modal/crew_locked.lua"
require "ui/modal/couriers.lua"
require "ui/modal/runners_busy.lua"

require "ui/modal/news_modal.lua"
require "ui/modal/top_hustlers.lua"
require "ui/modal/chase_view.lua"
require "ui/modal/chase_runaway_view.lua"

require "ui/modal/quests_view.lua"
require "ui/modal/quest_failed_view.lua"

require "anim/queue.lua"

require "ui/modal/wore_off_cash_boost.lua"
require "ui/modal/wore_off_rp_boost.lua"
require "ui/modal/wore_off_bust_protection.lua"

local view = setmetatable (
	{
		ui = { },
	},
	{ __index = ui.core.window }
)

local items = { }

local commodity_images		= ui.config.commodity_images


local preloaded_textures = {
	"ui/top_hustler/tophustler_top_bg:filtered",
	"ui/icons/ico_fame_hustler_header:filtered",
	"ui/backlight_storagefull:filtered",
	"ui/btn_close:filtered",
	"ui/top_hustler/btn_glossy_blue:filtered",
	"ui/top_hustler/btn_glossy_disabled:filtered",
	"ui/top_hustler/tophustler_nav_top_bg:filtered",
	"ui/top_hustler/nav_indicator_active:filtered",
	"ui/top_hustler/nav_indicator_passive:filtered",
	"ui/top_hustler/tophustler_nav_bottom_bg:filtered",
	"ui/top_hustler/btn_arrow_left:filtered",
	"ui/top_hustler/btn_arrow:filtered",
	"ui/loan_shark/ico_warning:filtered",
	"ui/scroll:filtered"
}

local function show_message (...)
	underworld2.ui.show_message (...)
end

local function check_levelup_routine (view, player)
	local prev_level_id = player:getLevel()
	return { 
		check = function()
			local new_level_id = player:getLevel()
			if (new_level_id > 2) and (new_level_id > prev_level_id) then
				view:show_levelup (new_level_id)
				-- update upvalue, this will make sequenced check() calls behave consistently  
				prev_level_id = new_level_id
			end
		end,
	}
end

local function __get_map_drop_items (loot)
	local items = { }
	local pos = loot.pos
	
	if loot.cash then
		table.insert (items, {
			pos = pos,
			slot = 1,
			icon = "ui/map/items/item_cash:filtered",
			text = string.format ("+%d", loot.cash.amount),
			hint = L "KEY_CASH",
			sound = "ui/SFX_Get_Money.wav",
			autocollect_delay = 1.0,
			cbfn = loot.cash.cbfn
		})
	end

	if loot.cash_bonus then
		table.insert (items, {
			pos = pos,
			slot = 3,
			icon = "ui/map/items/item_cash:filtered",
			text = string.format ("+%d", loot.cash_bonus.amount),
			hint = L "KEY_BONUS",
			sound = "ui/SFX_Get_Money.wav",
			autocollect_delay = 1.05,
			cbfn = loot.cash_bonus.cbfn
		})
	end

	if loot.commodity then
		local commodity = underworld2.CommoditiesDatabase():get (loot.commodity.id)
		table.insert (items, {
			pos = pos,
			slot = 1,
			icon = ui.config.commodity_icons[loot.commodity.id],
			text = string.format ("+%d", loot.commodity.amount),
			hint = underworld2.StringsDatabase():get (commodity:getNameKey()),
			sound = "ui/SFX_Loot.wav",
			autocollect_delay = 1.0,
			cbfn = loot.commodity.cbfn
		})
	end

	if loot.rp then
		table.insert (items, {
			pos = pos,
			slot = 2,
			icon = "ui/main/ico_topbar_rp:filtered",
			text = string.format ("+%d", loot.rp.amount),
			hint = L "KEY_RESPECT",
			sound = "ui/SFX_Loot.wav",
			autocollect_delay = 1.1,
			cbfn = loot.rp.cbfn
		})
	end

	if loot.rp_bonus then
		table.insert (items, {
			pos = pos,
			slot = 4,
			icon = "ui/main/ico_topbar_rp:filtered",
			text = string.format ("+%d", loot.rp_bonus.amount),
			hint = L "KEY_BONUS",
			sound = "ui/SFX_Loot.wav",
			autocollect_delay = 1.15,
			cbfn = loot.rp_bonus.cbfn
		})
	end

	if loot.fame then
		table.insert (items, {
			pos = pos,
			slot = 0,
			icon = "ui/icons/ico_fame_56:filtered",
			text = string.format ("+%d", loot.fame.amount),
			hint = L "KEY_FAME",
			sound = "ui/SFX_Loot.wav",
			autocollect_delay = 1.2,
			cbfn = loot.fame.cbfn
		})
	end
	return items
end

local function __eval_cash_amount_to_collect (player, courier_money, premium)
	local modifier = 0.0
	local bonus 			= player:getBonus ("CashBoostBonus")
	if bonus:get() then
		local modifier_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		modifier = modifier_bonus:getCurrentModifierValue (underworld2.server_time())
	end

	if premium then
		local cash = courier_money
		local cash_bonus = courier_money * modifier
		--print (string.format ("courier_money=%d; cash=%d; cash_bonus=%d(premium)", courier_money, cash, cash_bonus))
		return cash, cash_bonus
	end

	local total_cash = courier_money + courier_money * modifier
	local wallet_available = math.max (0, player:getMaxMoney() - player:getMoney())
	local max_to_collect = math.min (wallet_available, total_cash)
	local cash = math.floor (max_to_collect / (modifier + 1))
	local cash_bonus = math.floor (cash * modifier)

	--print (string.format ("courier_money=%d; cash=%d; cash_bonus=%d(free)", courier_money, cash, cash_bonus))

	return cash, cash_bonus
end

local function __eval_bonus (player, value, bonus_type)
	local bonus	= player:getBonus (bonus_type)
	if bonus:get() then
		local modifier_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		local modifier = modifier_bonus:getCurrentModifierValue (underworld2.server_time())
		if modifier > 0 then
			return math.floor (value * modifier)
		end
	end
end

local function calculate_payoff_price( cash )
	local priceReduce = 0.7 
	return underworld2.BalanceSettingsGoldDatabase():get(cash) * priceReduce
end

function show_crew()
	local v = ui.core.view.builder (ui.crew_view).build()
	underworld2.ui.stack:push_window (v)
	underworld2.ui.status_panel:hide_profile_info()
end

function show_equipment()
	local v = ui.core.view.builder (ui.equipment_view).build()
	underworld2.ui.stack:push_window (v)
	underworld2.ui.status_panel:hide_profile_info()
end

function show_store()
	local store = ui.core.view.builder (ui.store_view).build()
	underworld2.ui.stack:push_window (store)
	underworld2.ui.status_panel:hide_profile_info()
end

function show_leaderboard()
	ui.top_hustlers.show()
end

function show_crew_locked()
	local view = ui.core.view.builder (ui.crew_locked).build()
	ui.core.view.delegate_builder ({
		finished = function (self)
			ui.core.view_controller:close_modal_view (self)
		end
	}).build (view)
	ui.core.view_controller:show_modal_view (view)
end

function view:build_ui()
	self.first_launch = true
	self.tile_center = { }
	self.ui = { }
	self.ui.animations = anim.queue()
	self.editor_offset = 0

	local root_widget = ui.Picture ("root_widget")
	root_widget:setTexture ("ui/bg_light:filtered")
	root_widget:setPlacingMode (ui.Picture_PlacingModeStretch)
	root_widget:setOrigin (0.0, 0.0)
	root_widget:setSize (1.0, 1.0)
	self.ui.root_widget = root_widget

	local screen_ratio  = ui.DeviceMetrics():getScreenRatio()
	
	local map_tile = ui.MapWidget("map_widget")--ui.MapTile ("map_tile")
	map_tile:setOrigin (0, 0)
	map_tile:setSize (1.0, 1.0)
	map_tile:setEnabled (false)
	self.ui.map_tile = map_tile

	self.user_location_pin = ui.map.LocationPinPtr (runtime.disown (ui.map.UserLocationPin (math.geopoint (0, 0))))
	map_tile:setUserLocationPin (self.user_location_pin)

	local geoname_frame = ui.Frame ()
	geoname_frame:setBordersTexture ("ui/bar_location:filtered")
	local geoname_card_sz_y = geoname_frame:getTopMargin() + geoname_frame:getBottomMargin()
	geoname_frame:setTopMargin (0)
	geoname_frame:setBottomMargin (0)
	self.ui.geoname_frame = geoname_frame

	do

		local hlayout = ui.HorizontalLayout()

		local geoname = ui.TextLabel()
		geoname:setTextStyle (ui.text_styles.geoname.build())
		geoname:setText ("ABC")

		hlayout:addWidget (geoname)

		geoname_frame:setLayout (hlayout)

		local margin = math.max ((geoname_card_sz_y - geoname_frame:getPreferredSize().y) / 2, 0)
		geoname_frame:setTopMargin (margin)
		geoname_frame:setBottomMargin (margin)

		self.ui.geoname = geoname

	end

	do
		local context_payoff_frame = ui.Frame()
		context_payoff_frame:setBordersTexture ("ui/drop_heat_offer/frame_heat_offer:filtered")
		context_payoff_frame:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.vspacing * 2.0, ui.config.vspacing * 2.0)

		self.ui.context_payoff_frame = context_payoff_frame

		local frame_layout = ui.VerticalLayout()

		local hlayout = ui.HorizontalLayout()
		hlayout:setItemSpacing(ui.config.hspacing)

		local avatar = ui.Picture()

		avatar:setVerticalSizePolicy(ui.SizePolicyFixed)
		avatar:setHorizontalSizePolicy(ui.SizePolicyFixed)

		avatar:setTexture("ui/runners/ch_runner_avatar_1:filtered")
		local avatar_size = avatar:getPreferredSize()

		avatar:setPlacingMode(ui.Picture_PlacingModeFill)

		local text_layout = ui.VerticalLayout()
		--text_layout:setItemSpacing(ui.config.vspacing)

		local title_label = ui.TextLabel()
		title_label:setTextStyle(ui.text_styles.context_payoff_character_name.build())
		title_label:setText("Title")

		text_layout:addWidget(title_label)

		local content_label_1 = ui.TextLabel()
		content_label_1:setTextStyle(ui.text_styles.context_payoff_text.build())
		content_label_1:setText(L"KEY_CONTEXT_PAYOFF_LINE_1")

		local content_label_2 = ui.TextLabel()
		content_label_2:setTextStyle(ui.text_styles.context_payoff_text.build())
		content_label_2:setText(L"KEY_CONTEXT_PAYOFF_LINE_2")

		text_layout:addWidget(content_label_1)
		text_layout:addWidget(content_label_2)

		local text_height = title_label:getPreferredSize().y + content_label_1:getPreferredSize().y + content_label_2:getPreferredSize().y

		local avatar_scale = text_height / avatar_size.y
		avatar_size.y = text_height
		avatar_size.x = avatar_size.x * avatar_scale

		avatar:setSize(avatar_size)

		local avatar_hl = ui.HorizontalLayout()
		avatar_hl:addWidget(ui.Widget())
		local avatar_vl = ui.VerticalLayout()
		avatar_vl:addWidget(ui.Widget())

		local warning_sing = ui.Picture()

		warning_sing:setTexture("ui/ico_warning_modal:filtered")
		warning_sing:setVerticalSizePolicy(ui.SizePolicyFixed)
		warning_sing:setHorizontalSizePolicy(ui.SizePolicyFixed)
		local warning_sing_size = warning_sing:getPreferredSize()
		warning_sing:setPlacingMode(ui.Picture_PlacingModeFill)

		warning_sing_size.x = warning_sing_size.x * avatar_size.y / warning_sing_size.y * 0.4
		warning_sing_size.y = avatar_size.y * 0.4

		warning_sing:setSize(warning_sing_size)

		avatar_vl:addWidget(warning_sing)
		avatar_hl:addItem(avatar_vl)
		avatar:setLayout(avatar_hl)

		hlayout:addWidget(avatar)
		hlayout:addItem(text_layout)

		frame_layout:addItem(hlayout)

		hlayout = ui.HorizontalLayout()

		local price_button = ui.PriceButton()
		price_button:setStateTexture ("ui/drop_heat_offer/btn_glossy_premium_small:filtered")
		price_button:setTextStyle (ui.text_styles.button_small.build())
		price_button:setPrice (10)
		price_button:setText (L "KEY_HEAT_MODAL_LOWER_BUTTON")
		price_button:setCurrencyIcon ("ui/ico_pbtn_gold:filtered")
		price_button:setSize(0.4, 1.0)
		price_button:setHorizontalSizePolicy(ui.SizePolicyFixed)
		price_button:setVerticalSizePolicy(ui.SizePolicyMinimum)

		events.connect(price_button:getClickedSignalProxy(), function()
			local player = underworld2.player
			local exposure = player:getPoliceExposure()
			local price = exposure:getResetPrice (underworld2.server_time())
			if price == 0 then
				get_sound_pool():play ("ui/UI_Click_05.wav")
			elseif price <= player:getGold() then
				get_sound_pool():play ("ui/SFX_Heat_Drop.wav")
				underworld2.network.reset_heat (price, function() end)
				player:removeGold (price)
				player:resetHeat()
				self.ui.hide_context_payoff()
			else
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_gold.show(price - underworld2.player:getGold())
			end
		end)

		self.ui.payoff = price_button

		hlayout:addWidget(price_button)
		hlayout:addWidget(ui.Widget())

		frame_layout:addItem(hlayout)

		frame_layout:setItemSpacing(ui.config.vspacing)

		context_payoff_frame:setLayout(frame_layout)

		context_payoff_frame:setLeftMargin(context_payoff_frame:getLeftMargin() / 2)
		context_payoff_frame:setTopMargin(context_payoff_frame:getTopMargin() / 2)
		context_payoff_frame:setBottomMargin(context_payoff_frame:getBottomMargin() / 2)

		local size = context_payoff_frame:getPreferredSize()
		size.x = 1.0 - ui.config.hspacing * 4
		context_payoff_frame:setSize(size)
		context_payoff_frame:setOrigin(ui.config.hspacing * 2, 0.4)

		local size_difference = size.y - geoname_frame:getPreferredSize().y
		self.ui.context_payoff_frame_offset = size_difference

		context_payoff_frame:setHidden(true)

		self.ui.show_context_payoff = function()
			local courier = underworld2.player:getFirstFreeCourier()

			if not courier:get() then
				courier = underworld2.player:getCourierAt(0)
			end

			avatar:setTexture(string.format("ui/runners/ch_runner_avatar_%i:filtered", courier:getTag()))
			title_label:setText(L ("KEY_CONTEXT_PAYOFF_TITLE", { name = underworld2.couriers.names[courier:getTag()] } ) )

			context_payoff_frame:setHidden(false)
			geoname_frame:setHidden(true)

			price_button:startShaking()
		end

		self.ui.hide_context_payoff = function()
			context_payoff_frame:setHidden(true)
			geoname_frame:setHidden(false)
			price_button:stopShaking()
		end
	end


	local commodities = ui.CommoditiesWidget()
	commodities:setSwipeSound ("")
	commodities:setTabStateTexture ("ui/nav_button:filtered")
	commodities:setOrigin (0.0, 0.0)
	commodities:setSize (1.0, 1.0)
	events.connect (commodities:getTabChangedSignalProxy(), 
		function(index)
			self.current_tab_index = index
			if tutorial.initial[4] == underworld2.tutorial:current_stage() and index == 0 then
				underworld2.tutorial:next_stage()
			end
		end)
	self.ui.commodities = commodities

	local footer
	do
		local offset_x
		local offset_y

		footer = ui.Widget ()
		footer:setMargins (ui.config.hspacing, ui.config.hspacing, 0, ui.config.vspacing)
		local footer_layout = ui.HorizontalLayout()
		footer_layout:setItemSpacing (ui.config.hspacing)

		do
			local vlayout = ui.VerticalLayout()
			vlayout:setHorizontalPolicy (ui.SizePolicyMinimum)
			vlayout:setItemSpacing (ui.config.vspacing)

			do
				local hlayout = ui.HorizontalLayout()
				hlayout:setVerticalPolicy (ui.SizePolicyMinimum)

				local btn_top = ui.ImageButton()
				btn_top:setTexture ("ui/main/btn_bottombar_fame:filtered")
				btn_top:setHorizontalSizePolicy (ui.SizePolicyMinimum)
				events.connect (btn_top:getClickedSignalProxy(), show_leaderboard)
				hlayout:addWidget (btn_top)

				hlayout:addWidget (ui.Widget())

				vlayout:addItem (hlayout)
			end

			local btn_eqp = ui.LockButton()
			btn_eqp:setNormalTexture ("ui/main/btn_bottombar_equipment:filtered")
			btn_eqp:setLockedTexture ("ui/main/btn_bottombar_equipment_locked:filtered")
			btn_eqp:setRequiredLevel (underworld2.BalanceValuesDatabase():get ("equipmentUnlockLevel", 2))
			btn_eqp:setVerticalSizePolicy (ui.SizePolicyMinimum)
			btn_eqp:setDisplayCount(false)
			events.connect (btn_eqp:getClickedSignalProxy(), function()
				if btn_eqp:isLocked() then
					btn_eqp:startShakeAnimation()
				else
					show_equipment()
				end
			end)
			self.ui.btn_equipment = btn_eqp
			vlayout:addWidget (btn_eqp)

			vlayout:addWidget (ui.Widget())

			offset_x = btn_eqp:getPreferredSize().x + footer:getRightMargin() * 2.0

			footer_layout:addItem (vlayout)
		end

		footer_layout:addWidget (ui.Widget())

		do
			local vlayout = ui.VerticalLayout()
			vlayout:addWidget (ui.Widget())

			local btn_market = ui.TextButton()
			btn_market:setTexture ("ui/main/btn_bottombar_market:filtered")
			btn_market:setTextStyle (ui.text_styles.button_regular.build())
			btn_market:setText ( L ("KEY_BTN_MAP_FOOTER_MARKET"))
			btn_market:setVerticalSizePolicy (ui.SizePolicyMinimum)
			btn_market:setHorizontalSizePolicy (ui.SizePolicyMinimum)
			events.connect (btn_market:getClickedSignalProxy(), function()
				self:hide_map (true)
			end)
			self.ui.btn_market = btn_market
			vlayout:addWidget (btn_market)

			offset_y = footer:getBottomMargin() * 2.0 + btn_market:getPreferredSize().y

			footer_layout:addItem (vlayout)
		end

		footer_layout:addWidget (ui.Widget())

		do
			local vlayout = ui.VerticalLayout()
			vlayout:setHorizontalPolicy (ui.SizePolicyMinimum)
			vlayout:setItemSpacing (ui.config.vspacing)

			do
				local hlayout = ui.HorizontalLayout()
				hlayout:setVerticalPolicy (ui.SizePolicyMinimum)

				local btn_store = ui.ImageButton()
				btn_store:setTexture ("ui/main/btn_bottombar_bank:filtered")
				btn_store:setHorizontalSizePolicy (ui.SizePolicyMinimum)
				events.connect (btn_store:getClickedSignalProxy(), show_store)
				hlayout:addWidget (ui.Widget())
				hlayout:addWidget (btn_store)
				vlayout:addItem (hlayout)
			end

			local btn_crew = ui.LockButton()
			btn_crew:setNormalTexture ("ui/main/btn_bottombar_crew:filtered")
			btn_crew:setLockedTexture ("ui/main/btn_bottombar_crew_locked:filtered")
			btn_crew:setVerticalSizePolicy (ui.SizePolicyMinimum)
			btn_crew:setRequiredLevel (underworld2.BalanceValuesDatabase():get ("crewUnlockLevel", 4))
			btn_crew:setDisplayCount(false)
			events.connect (btn_crew:getClickedSignalProxy(), function()
				if btn_crew:isLocked() then
					show_crew_locked()
				else
					show_crew()
				end
			end)
			self.ui.btn_crew = btn_crew
			vlayout:addWidget (btn_crew)

			vlayout:addWidget (ui.Widget())

			footer_layout:addItem (vlayout)
		end

		footer:setLayout (footer_layout)

		local footer_size = footer:getPreferredSize()
		footer_size.x = 1.0
		self.default_footer_size = footer_size
		self.extended_footer_size = math.vec2 (footer_size.x + offset_x * 2.0, footer_size.y + offset_y * 2.0)

		footer:setSize (footer_size)
		footer:setOrigin (0.0, 1.0 - footer_size.y)

		self.ui.footer = footer
	end

	map_tile:setViewport (0, underworld2.ui.status_panel.size.y, 1.0, 1.0 - underworld2.ui.status_panel.size.y - self.ui.footer:getSize().y)
	map_tile:setViewCenter (math.vec2 (0.5, 0.5))

	root_widget:addChild (map_tile)
	root_widget:addChild (geoname_frame)
	root_widget:addChild (self.ui.context_payoff_frame)
	root_widget:addChild (footer)
	root_widget:addChild (commodities)

	self.ui.home = home

	self.ui.map_tile = map_tile
	self.ui.commodities = commodities
	self:_calc_commodities_height()

	local btn_show_map = ui.ImageButton()
	btn_show_map:setOrigin (0, 0)
	btn_show_map:setSize (1.0, self.commodities_y)
	root_widget:addChild (btn_show_map)
	self.ui.btn_show_map = btn_show_map
	events.connect (btn_show_map:getClickedSignalProxy(), function() 
		if self.ui.animations:empty() then
			self:show_map (true)
		end
	end)

	geoname_frame:setSize (1.0 - ui.config.hspacing * 4, geoname_frame:getPreferredSize().y)
	geoname_frame:setOrigin (ui.config.hspacing * 2, 1.0)

	local dealers_overlay = ui.DealersOverlay()
	map_tile:addOverlay (dealers_overlay)
	self.dealers_overlay = dealers_overlay

	events.connect (dealers_overlay:getDealSelectedSignalProxy(), function (id)
		self:on_deal_selected (id)
	end)

	events.connect (dealers_overlay:getQuestSelectedSignalProxy(), function (id)
		self:on_quest_selected (id)
	end)

	events.connect (dealers_overlay:getContactSelectedSignalProxy(), function (id)
		local profile = underworld2.ProfileDatabase():hasProfile (id) and underworld2.ProfileDatabase():get (id)
		if profile and profile:get() and profile:isPlayer() then
			ui.inbox.show_chat (id)
		end
	end)

	self.show_map_lock = util.recursive_lock.create (
		function() self.ui.btn_show_map:setEnabled (false) end,
		function() self.ui.btn_show_map:setEnabled (true) end)

	self.footer_lock = util.recursive_lock.create (
		function() self.ui.footer:setEnabled (false) end,
		function() self.ui.footer:setEnabled (true) end)

	self.card_handlers = 
	{
		on_offer_selected = function (w, index) self:on_card_offer_selected (w, index) end,
		on_order_selected = function (w, index) self:on_card_order_selected (w, index) end,
		on_mode_changed = function (w, mode) self:on_card_mode_changed (w, mode) end,
		on_edit_offer_price = function (w, price) self:on_card_edit_offer_price (w, price) end,
		on_edit_order_price = function (w, price) self:on_card_edit_order_price (w, price) end,
		on_animation_started = function (w) self:on_card_animation_started (w) end,
		on_animation_finished = function (w) self:on_card_animation_finished (w) end
	}

	for i = 1, #preloaded_textures do
		ui.preloadTexture(preloaded_textures[i])
	end
	
	return root_widget
end

function view:lock_trade_gesture()
	self.show_map_lock:lock()
	self.footer_lock:lock()
end

function view:unlock_trade_gesture()
	self.show_map_lock:release()
	self.footer_lock:release()
end

function view:_calc_commodities_height(h)
	local geoname_frame = ui.Frame ()
	geoname_frame:setBordersTexture ("ui/bar_location")
	local geoname_card_sz_y = geoname_frame:getTopMargin() + geoname_frame:getBottomMargin()
	geoname_frame:setTopMargin (0)
	geoname_frame:setBottomMargin (0)

	do
		local hlayout = ui.HorizontalLayout()

		local geoname = ui.TextLabel()
		geoname:setTextStyle (ui.text_styles.geoname.build())
		geoname:setText ("ABC")

		hlayout:addWidget (geoname)

		geoname_frame:setLayout (hlayout)
		geoname_card_sz_y = math.max (geoname_card_sz_y, geoname_frame:getPreferredSize().y)
	end

	--local w = ui.CommoditiesWidget()

	local geoname_card_sz_y = geoname_frame:getPreferredSize().y

	local frame_sz_y = h

	--if not h then
		local w = self.ui.commodities:addCommodityTab ()
		local tab = self.ui.commodities:getTabAt (0)
		frame_sz_y = w:getSize().y
	--end

	self.commodities_offset = geoname_card_sz_y + ui.config.vspacing + self.ui.commodities:getSize().y - tab:getSize().y
	local sz_y = frame_sz_y + self.commodities_offset

	self.commodities_y = 1.0 - sz_y
	self.frame_sz_y = frame_sz_y
	self.ui.commodities:setSize (1.0, 1.0)
	self.ui.commodities:setOrigin (0.0, self.commodities_y)
	self.frame_y = 1.0 - sz_y + geoname_card_sz_y + ui.config.vspacing

	--if not h then
		self.ui.commodities:removeTabAt (0)
	--end

	ui.destroy (geoname_frame)
	--ui.destroy (w)
end

function view:set_map_interactive (interactive)
	self.map_interaction = interactive
	self.ui.map_tile:setEnabled (interactive)
end

function view:title()
	return ""
end

function view:icon()
	return nil
end

function view:has_title_background()
	return true
end

function view:animate_map (t, duration, partial)
	local card_easing = anim.bezier(0.5,-0.15,1,-0.15)
	local f_card = function (t)
		local r = card_easing(t, duration)
		return r
	end

	local f_footer = function (t)
		local k = 0.8
		local d = 1.05
		local a = (d - k) / (k * k - k)
		return a * t * t + (1 - a) * t
	end

	local card_t = math.min (t, 0.7) / 0.7
	local footer_t = (math.max (t, 0.5) - 0.5) / 0.5

	--print(self.commodities_y)

	local y = self.editor_offset + (1 - self.commodities_y) * f_card (card_t)
	self.ui.commodities:setOrigin (0.0, y)
	self.ui.geoname_frame:setOrigin (ui.config.hspacing * 2, y + self.commodities_y)
	self.ui.context_payoff_frame:setOrigin(ui.config.hspacing * 2, y + self.commodities_y - self.ui.context_payoff_frame_offset)

	if not partial then
		local footer_dw = (self.extended_footer_size.x - self.default_footer_size.x) * (1.0 - f_footer (footer_t))
		local footer_dh = (self.extended_footer_size.y - self.default_footer_size.y) * (1.0 - f_footer (footer_t))
		self.ui.footer:setSize (self.default_footer_size.x + footer_dw, self.default_footer_size.y + footer_dh)
		self.ui.footer:setOrigin (0.0 - footer_dw * 0.5, self.ui.footer:getOrigin().y--[[ + self.editor_offset]])

		local top = underworld2.ui.status_panel.size.y
		local bottom = self.ui.footer:getOrigin().y
		local focus_y = (self.commodities_y + y - top) / (bottom - top) * 0.5--[[ + self.editor_offset]]
		
		self.ui.map_tile:setViewCenter(math.vec2 (0.5, focus_y))

		self.ui.map_tile:setGridOpacity (card_t)
	end
end

function view:set_commodities_card_offset (t, duration)
	local card_easing = anim.bezier (0.5,-0.15,1,-0.15)

	local y = self.editor_offset + (1 - self.commodities_y) * card_easing (t, duration)
	self.ui.commodities:setOrigin (0.0, y)
	self.ui.geoname_frame:setOrigin (ui.config.hspacing * 2, y + self.commodities_y)
	self.ui.context_payoff_frame:setOrigin(ui.config.hspacing * 2, y + self.commodities_y - self.ui.context_payoff_frame_offset)
end

function view:start_show_map_animation (animation_time)
	self.commodities_shown = false

	local elapsed = 0.0
	local animation_time = animation_time
	local start_y = self.ui.commodities:getOrigin().y + self.commodities_y
	local min_y = self.commodities_y
	self.animation_closure = function (dt)
		elapsed = elapsed + dt 

		if (elapsed > animation_time) then
			elapsed = animation_time
		end

		local t = (elapsed / animation_time)
		self:animate_map (t, animation_time)
		if elapsed >= animation_time then
			self.show_map_lock:lock()
			self.footer_lock:release()

			self:set_map_interactive (true)

			return false
		end
		return true
	end

	get_sound_pool():play ("ui/SFX_Menu_Close.wav")
end

function view:start_hide_map_animation (animation_time)
	self.commodities_shown = true

	local elapsed = 0.0
	local animation_time = animation_time
	local start_y = self.ui.commodities:getOrigin().y + self.commodities_y
	local min_y = self.commodities_y
	self.animation_closure = function (dt)
		elapsed = elapsed + dt

		if (elapsed > animation_time) then
			elapsed = animation_time
		end

		local t = (elapsed / animation_time)
		self:animate_map (1.0 - t, animation_time)
		if elapsed >= animation_time then
			self.footer_lock:lock()
			self.show_map_lock:release()

			self.map_events.market_opened (self)
			self:set_map_interactive (false)
			underworld2.TileDatabase():setCurrentEvent ("home")

			return false
		end
		return true
	end

	get_sound_pool():play ("ui/SFX_Menu_Woosh_Open.wav")
end

function view:show_map (animated)
	if self.animation_closure then
		self.animation_closure (1000)
		self.animation_closure = nil
	end
	if self.commodities_shown then
		self:start_show_map_animation (0.3)
		if not animated then
			self.animation_closure (1000)
			self.animation_closure = nil
		end
	end
end

function view:hide_map (animated)
	if self.animation_closure then
		self.animation_closure (1000)
		self.animation_closure = nil
	end
	if not self.commodities_shown then
		self:start_hide_map_animation (0.3)
		if not animated then
			self.animation_closure (1000)
			self.animation_closure = nil
		end
	end
end

function view:set_buying_mode()
	local mode = ui.CommoditiesCardMode_Buying
	self.card_mode = mode
	for i, widget in pairs (self.cards) do
		widget:setMode (mode)
	end
end

function view:set_selling_mode()
	local mode = ui.CommoditiesCard_DisplayModeSelling
	self.card_mode = mode
	for i, widget in pairs (self.cards) do
		widget:setMode (mode)
	end
end

function view:set_current_item (index)
	self.ui.commodities:selectTabAt (index - 1)
end

function view:start_offer_selection_tutorial()
	local card = self.cards[1]
	local btn = card:getOfferRow(0)
	ui.core.view_controller.instance:setTouchEventFilter (btn:mapToRoot (btn:getOrigin()), btn:getSize())

	local pos = btn:mapToRoot (btn:getOrigin()) + math.vec2 (btn:getSize().x * 0.75, btn:getSize().y * 0.5)
	return underworld2.tutorial:create_tap_view (pos, math.pi * 0.25)
end

function view:start_order_selection_tutorial()
	local card = self.cards[1]
	local btn = card:getOrderRow(0)
	ui.core.view_controller.instance:setTouchEventFilter (btn:mapToRoot (btn:getOrigin()), btn:getSize())

	local pos = btn:mapToRoot (btn:getOrigin()) + math.vec2 (btn:getSize().x * 0.75, btn:getSize().y * 0.5)
	return underworld2.tutorial:create_tap_view (pos, math.pi * 0.25)
end

function view:start_card_swipe_tutorial (dir)
	for i, v in ipairs (self.cards) do
		v:setEnabled (false)
	end

	local card = self.cards[dir == 0 and 1 or 2]
	local origin = card:mapToRoot (card:getOrigin())
	local size = card:getSize()
	local epos = origin + math.vec2 (size.x * 0.4, size.y * 0.5)
	local spos = origin + math.vec2 (size.x * 0.8, size.y * 0.5)
	return underworld2.tutorial:create_swipe_view (dir == 0 and spos or epos, dir == 0 and epos or spos, math.pi * 0.25)
end

function view:end_card_swipe_tutorial()
	for i, v in ipairs (self.cards) do
		v:setEnabled (true)
	end
end

function view:start_switch_to_selling_tutorial()
	local card = self.cards[1]
	local finger_pos = card:startSwitchToSellingTutorial()

	return underworld2.tutorial:create_tap_view (finger_pos, math.pi * 0.25)
end

function view:start_switch_to_buying_tutorial()
	local card = self.cards[1]
	local finger_pos = card:startSwitchToBuyingTutorial()

	return underworld2.tutorial:create_tap_view (finger_pos, math.pi * 0.25)
end

function view:init (location)
	local player = underworld2.player
	
	self.map_events = 
	{
		dealer_selected = function (self, pos)
			self:map_animate_translate (self.home_location:plane(), pos)
			underworld2.TileDatabase():setCurrentEvent ("dealer_trade")
		end,
		
		deal_started = function (self, pos)
		end,

		market_opened = function (self)
			underworld2.TileDatabase():setCurrentEvent ("home")
			self:map_animate_translate (self.ui.map_tile:getCenter(), self.home_location:plane())
		end
	}

	local bind_to_location = function (location) end

	if underworld2.ClientFeatures():getState ("map_scroll_lock", true) then
		local behaviour = ui.map.MapWidgetBindToLocation()
		bind_to_location = function (location) behaviour:bindToLocation (location) end

		self.ui.map_tile:setBehaviour (behaviour)
		
		events.connect (self.dealers_overlay:getFocusChangedSignalProxy(), function (location)
			bind_to_location (location)
		end)

		self.map_events.market_opened = function (self)
			underworld2.TileDatabase():setCurrentEvent ("home")
			bind_to_location (self.home_location:plane())
			self:map_animate_translate (self.ui.map_tile:getCenter(), self.home_location:plane())
		end
	else
		self.ui.map_tile:setBehaviour (ui.map.MapWidgetFreeScroll())
	end

	if underworld2.ClientFeatures():getState ("map_dealer_focus_on_deal", false) then
		self.map_events.dealer_selected = function (self, pos) end
		self.map_events.deal_started = function (self, pos)
			bind_to_location (pos)
			self:map_animate_translate (self.home_location:plane(), pos)
			underworld2.TileDatabase():setCurrentEvent ("dealer_trade")
		end
	end

	self.cards = { }

	events.connect (player:getQuestActivatedSignalProxy(), function (quest)
		self:on_quest_activated (quest)
	end)

	events.connect (player:getQuestAddedSignalProxy(), function (quest)
		self:on_quest_added (quest)
	end)

	events.connect (player:getStatsChangedSignalProxy(), function (p)
		self:on_player_stats_changed()
	end)

	events.connect (player:getCourierStateChangedSignalProxy(), function (courier)
		self:on_courier_state_changed (courier)
	end)

	events.connect (underworld2.Shops():getUpdatedSignalProxy(), function()
		self:on_shop_items_changed()
	end)

	events.connect (underworld2.ShopsNetwork():getNetworkChangedSignalProxy(), function()
		self:on_shops_network_changed()
	end)
	
	local tile_index = player:getTileInfo():getTileIndex()
	self.current_tile = math.vec2d (tile_index.x, tile_index.y)
	
	self:update_tile_scale()

	self:update_user_location (location:isValid() and location:getGeoPosition() or player:getLocation())
	self:on_shop_items_changed()

	for id, ptr in pairs (player:getQuests()) do
		local quest = ptr:get()
		self:on_quest_added (quest)
		if quest:isActivated() then
			self:on_quest_activated (quest)
		end
	end

	local couriers_count = player:getCouriersCount()
	for i = 0, couriers_count - 1 do
		local courier = player:getCourierAt (i)
		self:on_courier_state_changed (courier)
	end

	self:on_shops_network_changed()

	self.location_update_conn_id = events.connect (net.location_provider():getLocationUpdateSignalProxy(), function (location)
		self:location_updated (net.Location (location))
	end)
	
	self:on_player_stats_changed()

	--initial state
	self.commodities_shown = true
	self.footer_lock:lock()
	self:show_map (false)

	self:update_footer()

	if underworld2.tutorial:current_stage() == nil then
		self:hide_map()
	end
end

local function __update_bonus_events (self, player)
	local now = underworld2.server_time()
	local try_queue = function (bonus, end_time, view)
		if end_time <= now then return end

		local bonus_type = bonus:type()
		if not self.bonus_expired_events[bonus_type] then
			local timed_event_id = underworld2.timed_events:queue ("bonus_expired", end_time, function()
				self.bonus_expired_events[bonus_type] = nil
				local game_event
				game_event = underworld2.bonus_expired_event.create (function()
					if ui.core.view_controller:has_modal_views() then return false end
					view.show (function()
						game_event:finalize()
					end)
					return true
				end)
				underworld2.game_events:queue (game_event)
			end)
			self.bonus_expired_events[bonus_type] = underworld2.timed_events:get ("bonus_expired", timed_event_id)
		else
			self.bonus_expired_events[bonus_type].when = end_time
		end
	end

	if not self.bonus_expired_events then
		self.bonus_expired_events = { }
		underworld2.timed_events:create_queue ("bonus_expired")
	end

	local cash_bonus_ptr 		= player:getBonus ("CashBoostBonus")
	local rp_bonus_ptr 			= player:getBonus ("RespectBoostBonus")
	local police_protect_ptr 	= player:getBonus ("PoliceProtectionBonus")
	
	local cash_bonus 		= cash_bonus_ptr:get() and runtime.cast_to (cash_bonus_ptr:get(), "underworld2::ModifierBonus")
	local rp_bonus 			= rp_bonus_ptr:get() and runtime.cast_to (rp_bonus_ptr:get(), "underworld2::ModifierBonus")
	local police_protect 	= police_protect_ptr:get() and runtime.cast_to (police_protect_ptr:get(), "underworld2::PoliceProtectionBonus")

	if cash_bonus then try_queue (cash_bonus, cash_bonus:getEndTime (2.0), ui.wore_off_cash_boost) end
	if rp_bonus then try_queue (rp_bonus, rp_bonus:getEndTime (2.0), ui.wore_off_rp_boost) end
	if police_protect then try_queue (police_protect, police_protect:getEndTime(), ui.wore_off_bust_protection) end
end

function view:on_player_stats_changed()
	local player, cards = underworld2.player, self.cards
	for i, card in pairs (self.cards) do
		local commodity_id = card:getModel():getCommodityId()
		local storage = player:getStorage (commodity_id)
		card:setOffer (player:getOfferPrice (commodity_id))
		card:setOrder (player:getOrderPrice (commodity_id))
		card:setStorage (storage:getUsage(), storage:getCapacity())
	end

	self:update_footer()

	self:update_tile_scale()

	self:check_tutorial_trigger ({ "INITIAL", "PASSIVE_TRADING", "CREW", "HEAT" })

	__update_bonus_events (self, player)
end

function view:update_footer()
	local player = underworld2.player

	local equipment_unlock_level = underworld2.BalanceValuesDatabase():get ("equipmentUnlockLevel", 2)
	local crew_unlock_level = underworld2.BalanceValuesDatabase():get ("crewUnlockLevel", 4)

	self.ui.btn_equipment:setRequiredLevel (equipment_unlock_level)
	self.ui.btn_equipment:setLocked (player:getLevel() < equipment_unlock_level)
	self.ui.btn_crew:setRequiredLevel (crew_unlock_level)
	self.ui.btn_crew:setLocked (player:getLevel() < crew_unlock_level)

	self.ui.btn_crew:setNewItemsCount (underworld2.upgrades_tracker.has_new_items { level = player:getLevel(), category = "skills" } and 1 or 0)
	self.ui.btn_equipment:setNewItemsCount (underworld2.upgrades_tracker.has_new_items { level = player:getLevel(), category = "items" } and 1 or 0)
end

function view:check_tutorial_trigger (list)
	local active_trigger = underworld2.player:getActiveTrigger()
	if underworld2.tutorial:current_stage() or (#active_trigger == 0) then
		return false
	end

	local find = function (list, value)
		for i, v in pairs (list) do
			if v == value then 
				return true
			end
		end
		return false
	end

	if find (list, active_trigger) then
		local stages = tutorial.list[active_trigger]
		if stages then
			self:start_tutorial (active_trigger, stages)
			return true
		end
		print (string.format ("unknown tutorial requested: %s", active_trigger))
	end
	return false
end

function view:start_tutorial (trigger_name, stages)
	local e = underworld2.tutorial_event.create (function(e)
		if underworld2.ui.stack:top() ~= self or ui.core.view_controller:has_modal_views() then
			return false
		end

		underworld2.ui.status_panel:hide_profile_info()
		net.location_provider():lockUpdates()
		runtime.cast_to (self.user_location_pin:get(), "ui::map::UserLocationPin"):setRingsEnabled (false)
		underworld2.network.set_silent_retry (true)
		underworld2.tutorial:start (stages, trigger_name, function()
			underworld2.network.set_silent_retry (false)
			underworld2.player:setActiveTrigger ("")
			runtime.cast_to (self.user_location_pin:get(), "ui::map::UserLocationPin"):setRingsEnabled (true)
			net.location_provider():unlockUpdates()
			local location = net.location_provider():getLocation()
			self:location_updated (net.location_provider():getLocation(), true)
			e:finalize()
		end)

		return true
	end)

	underworld2.game_events:queue (e)
end

function view:on_courier_state_changed (courier)
	if courier:getState() == underworld2.CourierStateNotOccupied then
		self.dealers_overlay:removeCourier (courier:getId())
	else
		self.dealers_overlay:addCourier (courier)
	end
end

function view:lock_map_move()
	self.ui.map_tile:setMapMoveLocked (true)
end

function view:unlock_map_move()
	self.ui.map_tile:setMapMoveLocked (false)
end

function view:lock_map()
	self.ui.map_tile:setEnabled (false)
end

function view:unlock_map()
	self.ui.map_tile:setEnabled (true)
end


function view:update(dt)
	if self.animation_closure and not self.animation_closure (dt) then
		self.animation_closure = nil
	end
	self.ui.animations:update (dt)
	
	if underworld2.player then
		local now = underworld2.server_time()
		local exposure = underworld2.player:getPoliceExposure()

		--local heat_value 	= exposure:getHeatValue (now)
		local reset_price	= exposure:getResetPrice (now)
		self.ui.payoff:setPrice (tostring (reset_price))
	end
end

local get_pos = function(lon, lat, zoom)
	local n = math.pow(2, zoom)
	local x = ((lon + 180.0) / 360.0) * n
	local rlat = lat / 180.0 * math.pi
	local y = (1 - (math.log(math.tan(rlat) + 1.0 / math.cos(rlat)) / math.pi)) / 2.0 * n
	return x, y
end

local update_in_progress = false

local function _calc_line_circle_intersection (p1, p2, r)
	local dx = p1.x - p2.x
	local dy = p1.y - p2.y

	local a = dx * dx + dy * dy
	local b = -2 * (dx * p1.x + dy * p1.y)
	local c = p1.x * p1.x + p1.y * p1.y - r * r

	local d = b * b - 4 * a * c
	if d < 0 then 
		return nil, nil
	end
	return (-b + math.sqrt (d)) / (2 * a), (-b - math.sqrt (d)) / (2 * a)
end

function view:update_user_location (geopoint)
	local tile_info = underworld2.player:getTileInfo()
	self.ui.geoname:setText (tile_info:getCountryName().."/"..tile_info:getLocationName())
	
	if not self.home_location then
		self.home_location = geopoint
		self.map_events.market_opened (self)
	else
		self.home_location = geopoint
	end
	self.user_location_pin:setLocation (geopoint)
end

function view:location_updated (location, force)
	local geopoint
	if not location:isValid() then
		geopoint = underworld2.player:getLocation()
	else
		geopoint = location:getGeoPosition()
	end
	
	self:update_user_location (geopoint)
	
	if force then
		ui.event_view.show()
		underworld2.network.update_geo_position (location, function()
			ui.event_view.close()
		end)
		return
	end
	
	local _check_intersect = function (p1, p2, r)
		local t1, t2 = _calc_line_circle_intersection (p1, p2, r)
		if not t1 then
			return false
		end
		return (t1 >= 0 and t1 <= 1) or (t2 >= 0 and t2 <= 1)
	end

	local tile_info = underworld2.player:getTileInfo()
	
	local lon = geopoint.lon
	local lat = geopoint.lat
	
	if tile_info:get() then
		local center_x, center_y = get_pos (lon, lat, tile_info:getZoom())
		local tile_x, tile_y = math.floor (center_x), math.floor (center_y)
		
		if not self.current_tile then
			self.current_tile = math.vec2d (tile_x, tile_y)
		elseif (center_x < self.current_tile.x) or (center_y < self.current_tile.y) or (center_x > self.current_tile.x + 1) or (center_y > self.current_tile.y + 1) then
			-- check whether the new location data is accurate enough to switch to another tile
			local ha = location:getHorizontalAccuracy()
			local er = math.cos (lat / 180.0 * math.pi) * 6378.1 * 1000
			local r = math.pow (2, tile_info:getZoom()) * ha / (2 * math.pi * er)

			local center = math.vec2d (center_x, center_y)
			local p0 = self.current_tile - center
			local p1 = p0 + math.vec2d (0, 1)
			local p2 = p0 + math.vec2d (1, 1)
			local p3 = p0 + math.vec2d (1, 0)

			if _check_intersect (p0, p1, r) then
				return
			elseif _check_intersect (p1, p2, r) then
				return
			elseif _check_intersect (p2, p3, r) then
				return
			elseif _check_intersect (p3, p0, r) then
				return
			end
		end
		
		if (self.current_tile.x ~= tile_x) or (self.current_tile.y ~= tile_y) then 
			self.current_tile = math.vec2d (tile_x, tile_y)
			
			ui.event_view.show()
			underworld2.network.update_geo_position (location, function()
				ui.event_view.close()
			end)
		end
	end
end

local tutorial_collect_stage = { }

function view:update_tile_scale()
	if not underworld2.ClientFeatures():getState ("map_disable_hd_tiles", true) then
		if not underworld2.player or underworld2.player:getLevel() < 4 then
			self.ui.map_tile:setTileScale (4.0)
		else
			self.ui.map_tile:setTileScale (2.0)
		end
	else
		self.ui.map_tile:setTileScale (4.0)
	end
end

function view:on_quest_selected (id)
	ui.quests_view.show (id, function (view)
		if view.result == ui.quests_view.RESULT_COLLECT_REWARD then
			ui.event_view.show()
			local quest_ptr = view.selected_quest
			local quest_id = quest_ptr:get():getId()
			local levelup = check_levelup_routine (self, underworld2.player)

			underworld2.network.complete_quest (quest_id, function()
				ui.event_view.close()
				local quest = quest_ptr:get()--quest ptr is still alive

				local player = underworld2.player
				local reward = quest:getReward():get()
				
				local loot = { pos = self.dealers_overlay:getQuestLootEmmiterPos (quest:getId()) }

				local commodities_amount 	= reward:getCommoditiesAmount()
				local commodity_id 			= reward:getCommodityId()
				local rp 					= reward:getRespectPoints()
				local gold 					= reward:getGold()
				local cash 					= reward:getCash()
				if commodities_amount > 0 then
					loot.commodity = { amount = commodities_amount, id = commodity_id }
				end
				if rp > 0 then
					loot.rp = { amount = rp, cbfn = function() levelup:check() end }
					local rp_bonus = __eval_bonus (player, rp, "RespectBoostBonus")
					if rp_bonus then
						loot.rp_bonus = { amount = rp, cbfn = function() levelup:check() end }
					end
				end
				if gold > 0 then
					loot.gold = { amount = gold }
				end
				if cash > 0 then
					loot.cash = { amount = cash }
					local cash_bonus = __eval_bonus (player, cash, "CashBoostBonus")
					if cash_bonus then
						loot.cash_bonus = { amount = cash_bonus }
					end
				end
				local items = __get_map_drop_items (loot)
				for i, item in pairs (items) do
					self.dealers_overlay:dropLoot (item)
				end

				underworld2.player:removeQuest (quest_id)
				self.dealers_overlay:removeQuest (quest_id)
				--and now quest ptr is probably dead...
			end)
		end 
	end)
end

function view:on_deal_selected (id)
	local now = underworld2.server_time()

	local player 		= underworld2.player
	local courier 		= player:getCourier (id)
	local progress 		= courier:getProgressValue (now)
	local boost_price 	= courier:getBoostPrice (progress)

	self.dealers_overlay:collectLoot()
	
	--[[if underworld2.tutorial:current_stage() == tutorial.initial[15] then
		-- if user does not like reading...
		if courier:getEndTime() > now then
			courier:setEndTime (now)
			return
		end

		player:addRespectPoints (courier:getRespectPoints())
		player:addItemsToStorage (courier:getCommodityId(), courier:getItemsAmount())
		player:releaseCourier (id)
		underworld2.tutorial:next_stage()
		return
	elseif underworld2.tutorial:current_stage() == tutorial.initial[27] then
		get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")
		player:boostCourier (id)
		player:removeGold (boost_price)
		self.dealers_overlay:dealBoosted (id)
		underworld2.tutorial:next_stage()
		return
	elseif underworld2.tutorial:current_stage() == tutorial.initial[28] then
		player:addRespectPoints (courier:getRespectPoints())
		player:addMoney (courier:getMoney())
		player:releaseCourier (id)
		underworld2.tutorial:next_stage()
		return
	end]]


	local _boost_courier = function (courier)
		get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")
		player:boostCourier (courier:getId())
		player:removeGold (boost_price)
		player:lockUpdates()
		self.dealers_overlay:dealBoosted (courier:getId())

		network.online_controller.send_async
		{
			url = "/api/boost.do",
			body = 
			{
				name = "request",
				children = 
				{
					{ name = "type", text = 1 },
					{ name = "gold", text = boost_price },
					{ name = "id", text = courier:getId() }
				}
			},
			on_completed = function() player:unlockUpdates() end
		}

	end

	local _collect_items = function (courier, action_type)		
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		local sound_id = is_purchasing and "ui/SFX_Loot.wav" or "ui/SFX_Get_Money.wav"
		get_sound_pool():play (sound_id)
		
		local player 		= underworld2.player
		local commodity_id 	= courier:getCommodityId()
				
		local courier_id = courier:getId()

		local loot = { pos = self.dealers_overlay:getDealLootEmmiterPos (courier_id) }
		
		local free_courier = true
		local ditch = action_type == "DITCH"
		if not ditch then
			if is_purchasing then
				local amount = courier:getItemsAmount()
				loot.commodity = {
					id 		= commodity_id,
					amount 	= math.min (player:getStorage (commodity_id):getAvailable(), amount),
					cbfn 	= function() player:unlockUpdates() end
				}
				free_courier = amount == loot.commodity.amount
				
				player:addItemsToStorage (commodity_id, loot.commodity.amount)
				courier:setItemsAmount (amount - loot.commodity.amount)
			else
				local courier_money = courier:getMoney()
				local cash, cash_bonus = __eval_cash_amount_to_collect (player, courier_money)
				loot.cash = 
				{
					amount = cash,
					cbfn = function()
						player:addMoney (cash)
						player:unlockUpdates()
					end
				}
				if cash_bonus > 0 then
					loot.cash_bonus = 
					{
						amount = cash_bonus,
						cbfn = function()
							player:addMoney (cash_bonus)
							player:unlockUpdates()
						end
					}
				end
				free_courier = cash == courier_money
				courier:setMoney (courier_money - cash)
			end
		end

		local levelup = check_levelup_routine (self, player)

		local rp_earned = courier:getRespectPoints()
		if rp_earned > 0 then
			loot.rp = 
			{ 
				amount = rp_earned,
				cbfn = function()
					player:addRespectPoints (rp_earned)
					player:unlockUpdates()
					levelup:check()
				end
			}
			local rp_bonus = __eval_bonus (player, rp_earned, "RespectBoostBonus")
			if rp_bonus then
				loot.rp_bonus = 
				{
					amount = rp_bonus,
					cbfn = function()
						player:addRespectPoints (rp_bonus)
						player:unlockUpdates()
						levelup:check()
					end
				}
			end
			courier:setRespectPoints (0)
		end

		local fame_earned = 0
		if free_courier then
			fame_earned = courier:getFamePoints()
		end

		if fame_earned > 0 then
			loot.fame = 
			{ 
				amount = fame_earned,
				cbfn = function() player:unlockUpdates() end 
			}
			courier:setFamePoints (0)
		end
		

		local items = __get_map_drop_items (loot)
		player:lockUpdates()
		self.dealers_overlay:dealScaleAnimation (courier_id, 1.2, function()
			for i, item in pairs (items) do
				player:lockUpdates()
				self.dealers_overlay:dropLoot (item)
			end

			if free_courier then
				self.dealers_overlay:dealScaleAnimation (courier_id, 0.1, function()
					player:releaseCourier (courier_id)
					player:unlockUpdates()
				end)
			else
				self.dealers_overlay:dealScaleAnimation (courier_id, 0.7, function() 
					self.dealers_overlay:dealScaleAnimation (courier_id, 1.0, function()
						player:unlockUpdates() 
					end)
				end)
			end
		end)

		player:lockUpdates()

		network.online_controller.send_async 
		{
			url = "/api/completedeal.do",
			body = 
			{
				name = "request",
				children = 
				{
					{ name = "action", type = action_type },
					{ 
						name = "courier",
						children = { { name = "id", text = courier_id } }
					}
				}
			},
			on_completed = function (result) player:unlockUpdates() end
		}
	end
	
	local _premium_collect_items = function (courier, rent_price)
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		local sound_id = is_purchasing and "ui/SFX_Loot.wav" or "ui/SFX_Get_Money.wav"
		get_sound_pool():play (sound_id)
		
		local player 		= underworld2.player
		local commodity_id 	= courier:getCommodityId()
				
		local courier_id = courier:getId()

		local loot = { pos = self.dealers_overlay:getDealLootEmmiterPos (courier_id) }
		
		
		if is_purchasing then
			local amount = courier:getItemsAmount()
			loot.commodity = {
				id 		= commodity_id,
				amount 	= amount,
				cbfn 	= function() player:unlockUpdates() end
			}			
			player:addItemsToStorage (commodity_id, loot.commodity.amount)
			courier:setItemsAmount (amount - loot.commodity.amount)
		else
			local courier_money = courier:getMoney()
			local cash, cash_bonus = __eval_cash_amount_to_collect (player, courier_money, true)
			loot.cash = 
			{
				amount = cash,
				cbfn = function()
					player:addMoney (cash)
					player:unlockUpdates()
				end
			}
			if cash_bonus > 0 then
				loot.cash_bonus = 
				{
					amount = cash_bonus,
					cbfn = function()
						player:addMoney (cash_bonus)
						player:unlockUpdates()
					end
				}
			end
			courier:setMoney (courier_money - cash)
		end

		local levelup = check_levelup_routine (self, player)

		local rp_earned = courier:getRespectPoints()
		if rp_earned > 0 then
			loot.rp = 
			{ 
				amount = rp_earned,
				cbfn = function()
					player:addRespectPoints (rp_earned)
					player:unlockUpdates()
					levelup:check()
				end
			}
			local rp_bonus = __eval_bonus (player, rp_earned, "RespectBoostBonus")
			if rp_bonus then
				loot.rp_bonus = 
				{
					amount = rp_bonus,
					cbfn = function()
						player:addRespectPoints (rp_bonus)
						player:unlockUpdates()
						levelup:check()
					end
				}
			end
			courier:setRespectPoints (0)
		end

		local fame_earned = courier:getFamePoints()
		if fame_earned > 0 then
			loot.fame = 
			{ 
				amount = fame_earned,
				cbfn = function() player:unlockUpdates() end 
			}
			courier:setFamePoints (0)
		end
		

		local items = __get_map_drop_items (loot)
		player:lockUpdates()
		self.dealers_overlay:dealScaleAnimation (courier_id, 1.2, function()
			for i, item in pairs (items) do
				player:lockUpdates()
				self.dealers_overlay:dropLoot (item)
			end

			self.dealers_overlay:dealScaleAnimation (courier_id, 0.1, function()
				player:releaseCourier (courier_id)
				player:unlockUpdates()
			end)
		end)

		player:lockUpdates()
		network.online_controller.send_async
		{
			url = "/api/premiumCompleteDeal.do",
			body = 
			{
				name = "request",
				children = 
				{
					{ name = "courier", children = { { name = "id", text = courier_id}, { name = "completionCost", text = rent_price } } }
				}
			},
			on_completed = function() player:unlockUpdates() end
		}
	end

	local upvalue = {}

	local _show_rent_storage_dialogue = function (storage, courier)
		local rent_price 	= player:getRentPrice (courier)
		local v 			= ui.rent_modal.create (storage, courier, rent_price)
		ui.core.view_controller:show_modal_view (v)
		v.delegate_builder (
		{
			finished = function()
				ui.core.view_controller:close_modal_view (v)
				if v.result == ui.rent_modal.RESULT_CANCEL then
					return
				elseif v.result == ui.rent_modal.RESULT_DITCH then
					_collect_items (courier, "DITCH")
				elseif v.result == ui.rent_modal.RESULT_RENT then
					if player:getGold() < rent_price then
						get_sound_pool():play ("ui/SFX_Cash_Out.wav")
						ui.not_enough_gold.show(rent_price - player:getGold(), function()
							upvalue.show_rent_storage_dialogue(storage,courier)
						end )
						return
					else
						_premium_collect_items (courier, rent_price)
					end
				end
			end,

			on_back = function()
				ui.core.view_controller:close_modal_view (v)
			end
		}).build (v)

		get_sound_pool():play ("ui/SFX_Notice.wav")
	end
	
	upvalue.show_rent_storage_dialogue = _show_rent_storage_dialogue
	
	local _show_rent_vault_dialogue = function (level, ditch_amount, collected_amount)
		local bonus = player:getBonus ("CashBoostBonus")
		local modifier_bonus = bonus:get() and runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		local autoclose_at, event_id
		if modifier_bonus then
			local end_time = modifier_bonus:getClosestModifierEndTime (now)
			if now < end_time then autoclose_at = end_time end
		end

		local conv = underworld2.GoldConversionDatabase():get (level:getId())
		local rent_price 	= math.ceil (ditch_amount * conv:getRentPriceCoeff())
		local view = ui.core.view.builder (ui.vault_warning).build (level, ditch_amount, collected_amount, rent_price)
		ui.core.view_controller:show_modal_view (view)
		view.delegate_builder (
		{
			finished = function (view)
				if event_id ~= nil then underworld2.timed_events:unqueue (event_id) end

				ui.core.view_controller:close_modal_view (view)
				local player = underworld2.player
				if view.result == ui.vault_warning.RESULT_RENT then
					if rent_price <= underworld2.player:getGold() then
						_premium_collect_items (courier, rent_price)
					else
						get_sound_pool():play ("ui/SFX_Cash_Out.wav")
						ui.not_enough_gold.show(rent_price - underworld2.player:getGold(), function()
							upvalue.show_rent_vault_dialogue (level, ditch_amount, collected_amount)
						end )
					end
				elseif view.result == ui.vault_warning.RESULT_DITCH then
					_collect_items (courier, "DITCH")
				end
			end
		}).build (view)

		if autoclose_at then
			event_id = underworld2.timed_events:queue (autoclose_at, function()
				event_id = nil
				view:finalize()
			end)
		end
	end

	upvalue.show_rent_vault_dialogue = _show_rent_vault_dialogue

	if (now >= courier:getEndTime()) and (boost_price == 0) then
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		if is_purchasing then
			local storage_full = player:getStorage (courier:getCommodityId()):getAvailable() == 0
			if storage_full then
				_show_rent_storage_dialogue (player:getStorage (courier:getCommodityId()), courier)
			else
				_collect_items (courier, "COLLECT")
			end
		else
			local courier_money = courier:getMoney()
			local total_cash, total_bonus = __eval_cash_amount_to_collect (player, courier_money, true)
			local cash, cash_bonus 		  =	__eval_cash_amount_to_collect (player, courier_money, false)
			local collected_amount = cash + cash_bonus
			local ditch_amount = (total_cash + total_bonus) - collected_amount
			if collected_amount == 0 then
				local skill = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP"):getSkill ("WALLET")
				local level_id = player:getSkillLevel (skill:getId())
				local level = skill:getLevel (level_id)
				
				_show_rent_vault_dialogue (level, ditch_amount, collected_amount)
			else
				_collect_items (courier, "COLLECT")
			end
		end
	elseif (boost_price > 0) and (boost_price <= player:getGold()) then
		get_sound_pool():play ("ui/UI_Click_05.wav")
		_boost_courier (courier)
	else
		get_sound_pool():play ("ui/SFX_Cash_Out.wav")
		ui.not_enough_gold.show(boost_price - underworld2.player:getGold())
	end
	
end

local function show_quest_failed (quest)
	local e = underworld2.quest_expired_event.create (function(e)
		ui.quest_failed_view.show (quest, function()
			ui.event_view.show()
			local quest_id = quest:get():getId()
			underworld2.network.complete_quest (quest_id, function()
				e:finalize()
				ui.event_view.close()
			end)
		end)
		return true
	end)
	underworld2.game_events:queue (e)
end

function view:on_quest_activated (quest)
	local timed_quest = (quest:getClass():getName() == "TimedQuest") and runtime.cast_to (quest, "underworld2::quests::TimedQuest")
	if timed_quest then
		local quest_id = quest:getId()
		local ptr = underworld2.quests.QuestPtr (runtime.get_retained (quest))
		underworld2.timed_events:queue (timed_quest:getExpiresAt(), function()
			if ptr:get():isCompleted() then 
				return 
			end

			self.dealers_overlay:removeQuest (quest_id)
			underworld2.player:removeQuest (quest_id)
			
			if not ptr:get():isProgressStarted() then
				ui.event_view.show()
				underworld2.network.complete_quest (quest_id, function()
					ui.event_view.close()
				end)
				return
			end
			
			show_quest_failed (ptr)
		end)
	end
end

local function add_quest (self, quest_ptr)
	local quest = quest_ptr:get()
	self.dealers_overlay:addQuest (quest_ptr)
	if quest:isActivated() then
		self:on_quest_activated (quest)
	end
end

function view:on_quest_added (quest)
	local ptr = underworld2.quests.QuestPtr (runtime.get_retained (quest))
	local available_from = quest:getAvailableFrom()
	local now = underworld2.server_time()
	
	if now >= available_from then
		add_quest (self, ptr)
	else
		underworld2.timed_events:queue (available_from, function()
			add_quest (self, ptr)
		end)
	end
end

function view:on_shops_network_changed()
	self.dealers_overlay:reloadDealersNetwork()
end

function view:map_animate_translate (p1, p3)
	local p2 = self.ui.map_tile:adjustZoomForRegion (p1, p3)
	if (p2.z > self.ui.map_tile:getZoom()) then
		self.ui.map_tile:moveTo (p3, 15.0)
	else
		self.ui.map_tile:moveTo (p2)
		self.ui.map_tile:moveTo (p3, 15.0)
	end
end


local function build_show_offer_editor_animation (self, card, editor, handler_hidden, handler_complete)
	return anim.sequence {
		anim.lerp {
			from = 0.0, to = 1.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = function() 
				card:setDealEditor (editor)
				self.editor_offset = card:getEditorOffset()
				if handler_hidden then handler_hidden() end
			end
		},
		anim.lerp {
			from = 1.0, to = 0.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = handler_complete
		}
	}
end

local function build_show_order_editor_animation (self, card, editor, handler_hidden, handler_complete)
	return anim.sequence {
		anim.lerp {
			from = 0.0, to = 1.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = function() 
				card:setDealEditor (editor)
				self.editor_offset = card:getEditorOffset()
				if handler_hidden then handler_hidden() end
			end
		},
		anim.lerp {
			from = 1.0, to = 0.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = handler_complete
		}
	}
end

local function build_close_card_editor_animation (self, card, handler)
	return anim.lerp {
		from = 0.0, to = 1.0, time = 0.3,
		object = self,
		property = view.set_commodities_card_offset,
		on_finished = function()
			self.editor_offset = 0.0
			card:closeDealEditor()
			if handler then handler() end
		end
	}
end

local function build_show_commodities_animation (self, handler)
	return anim.lerp {
		from = 1.0, to = 0.0, time = 0.3,
		object = self,
		property = view.set_commodities_card_offset,
		on_finished = handler
	}
end

local function build_cancel_card_editor_animation (self, card)
	return anim.sequence {
		anim.lerp {
			from = 0.0, to = 1.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = function()
				self.editor_offset = 0.0
				card:closeDealEditor()
			end
		},
		anim.lerp {
			from = 1.0, to = 0.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset
		}
	}
end

function view:on_card_offer_selected (w, index)
	if not index:isValid() then return end
	if not self.ui.animations:empty() then return end

	if not underworld2.player:getFirstFreeCourier():get() then
		self:show_runners_busy()
		return
	end
	local shop = reflect.toTable (index:modelData())
	local commodity_id = index:model():getCommodityId()
	local storage = underworld2.player:getStorage (commodity_id)
	local capacity = storage:getCapacity()
	local available = storage:getAvailable()
	local amount = shop.selling_amount
	local heat_raise = underworld2.CommoditiesDatabase():get (commodity_id):getHeatRaise()

	local editor = w:createDealEditor (index, ui.DealEditorWidget_ModeBuy)
	editor:setAmountLimit (math.min (capacity, amount), available + 1)
	self.ui.commodities:lockPageMove()

	events.connect (editor:getFinishedSignalProxy(), function(editor, result, selected_amount)
		local finalize = function()
			self.ui.commodities:unlockPageMove()
			self.ui.hide_context_payoff()
			underworld2.ui.status_panel:set_heat_delta (0)
		end 
		
		if underworld2.tutorial:current_stage() == tutorial.initial[9] then
			self.ui.animations:push (build_close_card_editor_animation (self, w, function()
				finalize()
				underworld2.tutorial:next_stage()
			end))
			return
		end

		if result == ui.DealEditorWidget_ResultOk then
			self:on_card_purchase_items (w, commodity_id, shop, selected_amount, finalize)
		else
			finalize()
			self.ui.animations:push (build_cancel_card_editor_animation (self, w))
		end
	end)
	events.connect (editor:getAmountChangedSignalProxy(), function (editor, selected_amount)
		underworld2.ui.status_panel:set_heat_delta (heat_raise * selected_amount)
	end)

	self.map_events.dealer_selected (self, math.geopoint (shop.location.lon, shop.location.lat):plane())
	local a = build_show_offer_editor_animation (self, w, editor, 
		function()
			local now = underworld2.server_time()
			local player = underworld2.player
			local exposure = player:getPoliceExposure()

			local do_show_context_payoff = exposure:getHeatValue(now) >= underworld2.BalanceValuesDatabase():get ("contextDropHeatTreshold", 0.8)
			local police_protect = player:getBonus ("PoliceProtectionBonus")
			if do_show_context_payoff and police_protect:get() then
				local end_time = runtime.cast_to (police_protect:get(), "underworld2::PoliceProtectionBonus"):getEndTime()
				if end_time > now then
					do_show_context_payoff = false
				end
			end

			if do_show_context_payoff then
				self.ui.show_context_payoff()
			end
		end,
		function()
			if underworld2.tutorial:current_stage() == tutorial.initial[7] then
				underworld2.tutorial:next_stage()
			elseif underworld2.tutorial:current_stage() == tutorial.heat[3] then
				underworld2.tutorial:next_stage()
			end
		end)
	self.ui.animations:push (a)
end

function view:on_card_order_selected (w, index)
	if not index:isValid() then return end
	if not self.ui.animations:empty() then return end

	if not underworld2.player:getFirstFreeCourier():get() then
		self:show_runners_busy()
		return
	end

	local shop = reflect.toTable (index:modelData())
	local commodity_id = index:model():getCommodityId()
	local commodity = underworld2.CommoditiesDatabase():get (commodity_id)
	local storage = underworld2.player:getStorage (commodity_id)
	local heat_raise = commodity:getHeatRaise()

	if storage:getUsage() == 0 then
		w:animateEmptyStorage()
		return
	end

	local editor = w:createDealEditor (index, ui.DealEditorWidget_ModeSell)
	editor:setAmountLimit (math.min (storage:getUsage(), shop.purchase_amount))
	self.ui.commodities:lockPageMove()

	events.connect (editor:getFinishedSignalProxy(), function(editor, result, selected_amount)
		self.ui.commodities:unlockPageMove()
		self.ui.hide_context_payoff()
		underworld2.ui.status_panel:set_heat_delta (0)

		if underworld2.tutorial:current_stage() == tutorial.initial[24] then
			self.ui.animations:push (build_close_card_editor_animation (self, w, function()
				underworld2.tutorial:next_stage()
			end))
			return
		end

		if result == ui.DealEditorWidget_ResultOk then
			self:on_card_sell_items (w, commodity_id, shop, selected_amount)
		else
			self.ui.animations:push (build_cancel_card_editor_animation (self, w))
		end
	end)
	events.connect (editor:getAmountChangedSignalProxy(), function (editor, selected_amount)
		underworld2.ui.status_panel:set_heat_delta (heat_raise * selected_amount)
	end)

	self.map_events.dealer_selected (self, math.geopoint (shop.location.lon, shop.location.lat):plane())
	local a = build_show_order_editor_animation (self, w, editor,
		function()
			local now = underworld2.server_time()
			local player = underworld2.player
			local exposure = player:getPoliceExposure()

			local do_show_context_payoff = exposure:getHeatValue(now) >= underworld2.BalanceValuesDatabase():get ("contextDropHeatTreshold", 0.8)
			local police_protect = player:getBonus ("PoliceProtectionBonus")
			if do_show_context_payoff and police_protect:get() then
				local end_time = runtime.cast_to (police_protect:get(), "underworld2::PoliceProtectionBonus"):getEndTime()
				if end_time > now then
					do_show_context_payoff = false
				end
			end

			if do_show_context_payoff then
				self.ui.show_context_payoff()
			end
		end,
		function()
			if underworld2.tutorial:current_stage() == tutorial.initial[22] then
				underworld2.tutorial:next_stage()
			elseif underworld2.tutorial:current_stage() == tutorial.heat[3] then
				underworld2.tutorial:next_stage()
			end
		end)
	self.ui.animations:push (a)
end

function view:on_card_mode_changed (w, mode)
	if underworld2.tutorial:current_stage() == tutorial.initial[20] then
		underworld2.tutorial:next_stage()
	end

	self.card_mode = mode
	for i, widget in pairs (self.cards) do
		if w:getModel():getCommodityId() ~= widget:getModel():getCommodityId() then
			widget:setMode (mode)
		end
	end
end

function view:on_card_purchase_items (card, commodity_id, shop, amount, handler)
	local player = underworld2.player
	local courier = player:getFirstFreeCourier()
	if not courier:get() then
		show_message (L "KEY_DEAL_NO_COURIES")
		return
	end

	local storage = player:getStorage (commodity_id)
	local price = shop.selling_price

	if not is_player then
		price = player:getStats():labSellPrice (price, commodity_id)
	end

	if player:getMoney() < (amount * price) then
		get_sound_pool():play ("ui/SFX_Cash_Out.wav")
		ui.not_enough_cash.show (amount * price - player:getMoney(), false)
		return
	end

	local purchase_confirmed = function()
		handler()
		self.ui.animations:push (anim.lerp {
			from = 0.0, to = 1.0, time = 0.3,
			object = self,
			property = view.set_commodities_card_offset,
			on_finished = function()
				self.editor_offset = 0.0
				card:closeDealEditor()
				self:process_buy_items (shop, commodity_id, amount)
			end
		})
	end

	if (storage:getAvailable() < amount) then
		ui.deal_alert_view.show (commodity_id, amount, ui.format.duration_string (shop.duration), function (result)
			if result == ui.deal_alert_view.RESULT_BUY then
				purchase_confirmed()
			end
		end)
		return
	else
		purchase_confirmed()
	end
end

function view:on_card_sell_items (card, commodity_id, shop, amount)
	self.ui.animations:push (anim.lerp {
		from = 0.0, to = 1.0, time = 0.3,
		object = self,
		property = view.set_commodities_card_offset,
		on_finished = function()
			self.editor_offset = 0.0
			card:closeDealEditor()
			self:process_sell_items (shop, commodity_id, amount)
		end
	})
end

function view:on_card_edit_offer_price (card)
	if not self.ui.animations:empty() then return end

	local commodity_id = card:getModel():getCommodityId()
	local current_price = underworld2.player:getOfferPrice (commodity_id)
	local editor = card:createPriceEditor (current_price, true)

	local animate_editor = function (card, t) card:animatePriceEditor (t) end

	self.ui.commodities:lockPageMove()

	events.connect (editor:getFinishedSignalProxy(), function (editor, result)
		local update_price = result == ui.PriceEditorWidget_ResultOk
		local offer_price = editor:getPrice()
		local order_price = underworld2.player:getOrderPrice (commodity_id)
		self.ui.animations:push (anim.lerp
		{
			from = 1.0, to = 0.0, time = 0.15,
			object = card, property = animate_editor,
			on_finished = function()
				self.ui.commodities:unlockPageMove()
				card:closePriceEditor()
				if update_price then 
					self:process_update_items_price (commodity_id, offer_price, order_price)
				end
			end
		})
	end)

	self.ui.animations:push (anim.lerp
	{
		from = 0.0, to = 1.0, time = 0.2,
		object = card, property = animate_editor,
		on_finished = function()
			print "tutorial offer"
			if underworld2.tutorial:current_stage() == tutorial.passive_trading[12] then
				underworld2.tutorial:next_stage()
			end
		end
	})
	card:setPriceEditor (editor)
	animate_editor (card, 0.0)
end

function view:on_card_edit_order_price (card)
	if not self.ui.animations:empty() then return end

	local commodity_id = card:getModel():getCommodityId()
	local current_price = underworld2.player:getOrderPrice (commodity_id)
	local editor = card:createPriceEditor (current_price, false)

	local animate_editor = function (card, t) card:animatePriceEditor (t) end
	self.ui.commodities:lockPageMove()

	events.connect (editor:getFinishedSignalProxy(), function (editor, result)
		local update_price = result == ui.PriceEditorWidget_ResultOk
		local offer_price = underworld2.player:getOfferPrice (commodity_id)
		local order_price = editor:getPrice()
		self.ui.animations:push (anim.lerp
		{
			from = 1.0, to = 0.0, time = 0.15,
			object = card, property = animate_editor,
			on_finished = function()
				self.ui.commodities:unlockPageMove()
				card:closePriceEditor()
				if update_price then 
					self:process_update_items_price (commodity_id, offer_price, order_price)
				end
			end
		})
	end)

	self.ui.animations:push (anim.lerp
	{
		from = 0.0, to = 1.0, time = 0.2,
		object = card, property = animate_editor,
		on_finished = function()
			print "tutorial order"
			if underworld2.tutorial:current_stage() == tutorial.passive_trading[8] then
				underworld2.tutorial:next_stage()
			end
		end
	})
	card:setPriceEditor (editor)
	animate_editor (card, 0.0)
end

function view:on_card_animation_started (w)
	self.ui.commodities:lockPageMove()
end

function view:on_card_animation_finished (w)
	self.ui.commodities:unlockPageMove()
end

function view:process_update_items_price (commodity_id, offer_price, order_price)
	ui.event_view.show()
	underworld2.network.set_item_price (
	{
		id 					= commodity_id,
		purchase_price		= order_price,
		sell_price			= offer_price
	},
	ui.event_view.close)
end

function view:validate_deal_response (response, acceptable_results)
	local child_for_name = network.online_controller.child_for_name
	local result = child_for_name (response, "dealResult").text

	if not table.find (acceptable_results, result) then
		local msg = string.format ("Unexpected response: result=%s", tostring (result))
		return false, msg 
	end
	return true
end

function view:process_buy_items (shop, commodity_id, amount)
	local player = underworld2.player
	local courier = player:getFirstFreeCourier()
	local courier_id = courier:getId()
	local courier_tag = courier:getTag()

	local price = shop.selling_price
	if not shop.is_player then
		price = player:getStats():labSellPrice (price, commodity_id)
	end
		
	local levelup = check_levelup_routine (self, player)

	local finalize_with_error = function (text)
		self.ui.animations:push (build_show_commodities_animation (self, function()
			get_sound_pool():play ("ui/UI_Error.wav")
			show_message (text)
		end))
	end

	network.online_controller.send 
	{
		url = "/api/buy.do",
		body = 
		{
			name = "request",
			children =
			{
				{ 
					name = "owner", 
					children = 
					{ 
						{ name = "id", text = shop.owner_id },
						{ name = "isPlayer", text = tostring (shop.is_player) },
					} 
				},
				{
					name = "item",
					children = 
					{
						{ name = "id", text = commodity_id },
						{ name = "count", text = amount },
						{ name = "price", text = price }
					}
				},
				{ name = "courierId", text = courier_id }
			}
		},
		on_completed = function (response)
			local ok, message = self:validate_deal_response (response, {"SUCCESS", "POLICE_INTERCEPT", "NOT_ENOUGH_MONEY"})
			if not ok then
				finalize_with_error (message)
				return
			end

			self:process_buy_response (response, shop, commodity_id, price, amount, courier_tag, levelup)
		end,

		on_server_error = function (error)
			local acceptable = { [30] = L "KEY_DEAL_CONDITIONS_CHANGED", [31] = L "KEY_SHOP_NOT_AVAILABLE" }
			if acceptable[error.code] then
				finalize_with_error (acceptable[error.code])
				return true
			end
			return false
		end
	}
end

function view:process_buy_response (response, shop, commodity_id, price, amount, courier_tag, levelup)
	local deal_succeed = function()
		local profile = underworld2.ProfileDatabase():get (shop.owner_id)
		self.map_events.deal_started (self, profile:getLocation():plane())
		get_sound_pool():play ("ui/SFX_Runner_Start.wav")
		if underworld2.tutorial:current_stage() == tutorial.heat[4] then
			underworld2.tutorial:next_stage()
		end
	end

	local child_for_name = network.online_controller.child_for_name

	local result = child_for_name (response, "dealResult").text
	local chase  = child_for_name (response, "courierWasChased").text

	if result == "POLICE_INTERCEPT" then
		local bust_details = {
			commodity_id 	= commodity_id,
			amount 			= amount,
			cash 			= math.floor (amount * price),
			location 		= child_for_name (response, "location").text,
			courier_tag 	= courier_tag,
			is_purchasing	= true,
			payoff_price	= tonumber (child_for_name (response, "payOffPrice").text),
		}
		local fb_share = child_for_name (response, "freeEscapeAllowed").text == "true"
		self:show_chase_busted (bust_details, fb_share, function (bribed)
			if bribed then deal_succeed() end
			self.ui.animations:push (build_show_commodities_animation (self, function()
				levelup:check()
			end))
		end)
	elseif result == "NOT_ENOUGH_MONEY" then
		self.ui.animations:push (build_show_commodities_animation (self, function()
			get_sound_pool():play ("ui/SFX_Cash_Out.wav")
			ui.not_enough_cash.show (amount * shop.selling_price - underworld2.player:getMoney(), false)
		end))
	elseif chase == "true" then
		local location = child_for_name (response, "location").text
		local rp = tonumber (child_for_name (response, "chaseRpReward").text)
		self:show_chase_runaway (location, courier_tag, commodity_id, amount, rp, function()
			deal_succeed()
			self.ui.animations:push (build_show_commodities_animation (self, function()
				levelup:check()
			end))
		end)
	else
		deal_succeed()
		self.ui.animations:push (build_show_commodities_animation (self, function()
			levelup:check()
		end))
	end
end

function view:process_sell_items (shop, commodity_id, amount)
	local player = underworld2.player
	local courier = player:getFirstFreeCourier()
	local courier_id = courier:getId()
	local courier_tag = courier:getTag()

	local levelup = check_levelup_routine (self, player)

	local finalize_with_error = function (text)
		self.ui.animations:push (build_show_commodities_animation (self, function()
			get_sound_pool():play ("ui/UI_Error.wav")
			show_message (text)
		end))
	end

	network.online_controller.send 
	{
		url = "/api/sell.do",
		body = 
		{
			name = "request",
			children =
			{
				{ 
					name = "buyer", 
					children = 
					{ 
						{ name = "id", text = shop.owner_id },
						{ name = "isPlayer", text = tostring (shop.is_player) },
					} 
				},
				{
					name = "item",
					children = 
					{
						{ name = "id", text = commodity_id },
						{ name = "count", text = amount },
						{ name = "price", text = shop.purchase_price }
					}
				},
				{ name = "courierId", text = courier_id }
			}
		},
		on_completed = function (response)
			local ok, message = self:validate_deal_response (response, { "SUCCESS", "POLICE_INTERCEPT" })
			if not ok then
				finalize_with_error (message)
				return
			end

			self:process_sell_response (response, shop, commodity_id, amount, courier_tag, levelup)
		end,

		on_server_error = function (error)
			local acceptable = { [30] = L "KEY_DEAL_CONDITIONS_CHANGED", [31] = L "KEY_SHOP_NOT_AVAILABLE" }
			if acceptable[error.code] then
				finalize_with_error (acceptable[error.code])
				return true
			end
			return false
		end
	}
end

function view:process_sell_response (response, shop, commodity_id, amount, courier_tag, levelup)
	local deal_succeed = function()
		local profile = underworld2.ProfileDatabase():get (shop.owner_id)
		self.map_events.deal_started (self, profile:getLocation():plane())
		get_sound_pool():play ("ui/SFX_Runner_Start.wav")
		if underworld2.tutorial:current_stage() == tutorial.heat[4] then
			underworld2.tutorial:next_stage()
		end
	end
	local child_for_name = network.online_controller.child_for_name

	local result = child_for_name (response, "dealResult").text
	local chase  = child_for_name (response, "courierWasChased").text

	if result == "POLICE_INTERCEPT" then
		local bust_details = {
			commodity_id 	= commodity_id,
			amount 			= amount,
			cash 			= math.floor (amount * shop.purchase_price),
			location 		= child_for_name (response, "location").text,
			courier_tag 	= courier_tag,
			is_purchasing	= false,
			payoff_price	= tonumber (child_for_name (response, "payOffPrice").text)
		}
		self:show_chase_busted (bust_details, response.fb_share, function (bribed)
			if bribed then deal_succeed() end
			self.ui.animations:push (build_show_commodities_animation (self, function()
				levelup:check()
			end))
		end)
	elseif chase == "true" then
		local location = child_for_name (response, "location").text
		local rp = tonumber (child_for_name (response, "chaseRpReward").text)
		self:show_chase_runaway (location, courier_tag, commodity_id, amount, rp, function()
			self.ui.animations:push (build_show_commodities_animation (self, function()
				deal_succeed()
				levelup:check()
			end))
		end)
	else
		self.ui.animations:push (build_show_commodities_animation (self, function()
			deal_succeed()
			levelup:check()
		end))
	end
end

function view:on_shop_items_changed()
	local tinsert = table.insert
	local player, shops  = underworld2.player, underworld2.Shops()
	
	local skill_group = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP")
	local skill = skill_group:getSkill ("ITEM")
	local skill_level = underworld2.player:getSkillLevel (skill:getId())
	
	-- 1-st and 2-nd commodities are available by default
	local unlocked_commodities = { 1, 2 }
	for i = 1, skill_level do
		tinsert (unlocked_commodities, i + 1)
	end

	self.card_mode = self.card_mode or ui.CommoditiesCardMode_Buying

	for __, commodity_id in ipairs (unlocked_commodities) do
		local w = self.cards[commodity_id]
		if not w then
			w = self.ui.commodities:addCommodityTab()
			local tab = w:getParent()
			self.cards[commodity_id] = w
			w:setOrigin ((1.0 - w:getSize().x) * 0.5, tab:getSize().y - w:getSize().y)
			
			w:setModel (shops:get (commodity_id))
			
			local handlers = self.card_handlers

			events.connect (w:getOfferSelectedSignalProxy(), handlers.on_offer_selected)
			events.connect (w:getOrderSelectedSignalProxy(), handlers.on_order_selected)
			events.connect (w:getModeChangedSignalProxy(), handlers.on_mode_changed)
			events.connect (w:getAnimationStartedSignalProxy(), handlers.on_animation_started)
			events.connect (w:getAnimationFinishedSignalProxy(), handlers.on_animation_finished)
			events.connect (w:getEditOfferSignalProxy(), handlers.on_edit_offer_price)
			events.connect (w:getEditOrderSignalProxy(), handlers.on_edit_order_price)
		end

		w:setOffer (player:getOfferPrice (commodity_id))
		w:setOrder (player:getOrderPrice (commodity_id))
		local storage = player:getStorage (commodity_id)
		w:setStorage (storage:getUsage(), storage:getCapacity())

		w:setItemImage (commodity_images [commodity_id]) 

		w:setMode (self.card_mode)
	end 
end

function view:show_chase_busted (bust_details, fb_share, handler)
	local location = bust_details.location
	local courier_tag = bust_details.courier_tag
	local commodity_id = bust_details.commodity_id
	local amount = bust_details.amount
	ui.chase_view.show (location, courier_tag, commodity_id, amount, true, function()
		self:show_busted (bust_details, fb_share, handler)
	end)
end

function view:show_chase_runaway (location, courier_tag, commodity_id, amount, rp_reward, handler)
	ui.chase_view.show (location, courier_tag, commodity_id, amount, false, function()
		ui.chase_runaway_view.show (courier_tag, rp_reward, handler)
	end)
end

local function __do_payoff (context)
	ui.event_view.show()
	--underworld2.network.push_back
	network.online_controller.send
	{
		url = "/api/payOff.do",
		body = 
		{
			name = "request",
			type = context.type,
			children = 
			{
				{ name = "cost", text = context.price },
			}
		},
		on_completed = function()
			ui.event_view.close()
			context.on_completed()
		end
	}
end

function view:show_busted (bust_details, fb_share, handler)
	get_sound_pool():play ("ui/SFX_Heat_Drop.wav")
	get_sound_pool():play ("ui/SFX_Busted.wav")

	if self:check_tutorial_trigger ( { "BUSTING" } ) then
		handler (false)
		tutorial.bust_params = bust_details
		return
	end

	local upvalue = {}

	upvalue.show_modal = function ()
		local v = ui.busted_modal.create (bust_details, fb_share)
		ui.core.view_controller:show_modal_view (v)

		v.delegate_builder (
		{
			finished = function()
				ui.core.view_controller:close_modal_view (v)
				local payoff_price = bust_details.payoff_price
				if v.result == ui.busted_modal.RESULT_PAYOFF then
					if payoff_price <= underworld2.player:getGold() then 
						get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")
						__do_payoff {
							price = payoff_price,
							type = "gold",
							on_completed = handler and function() handler (true) end
						}
					else
						get_sound_pool():play ("ui/SFX_Cash_Out.wav")
						ui.not_enough_gold.show (payoff_price - underworld2.player:getGold(), function() upvalue.show_modal() end)
					end
				elseif v.result == ui.busted_modal.RESULT_SHARE then
					get_sound_pool():play ("ui/UI_Click_05.wav")
					__do_payoff {
						price = payoff_price,
						type = "fb_share",
						on_completed = handler and function() handler (true) end
					}
				elseif handler then
					handler (false)
				end
			end
		}).build (v)
	end

	upvalue.show_modal()
end

function view:moved_to_background()
end

function view:moved_to_foreground()
	self:update_footer()
end

function view:show_levelup (level_id)
	self:update_footer()

	get_sound_pool():play ("ui/SFX_LevelUp.wav")
	local v = ui.core.view.builder (ui.levelup).build (level_id)
	ui.core.view_controller:show_modal_view (v)
	v.delegate_builder (
	{
		finished = function (v)
			ui.core.view_controller:close_modal_view (v)
		end
	}).build (v)
end

function view:check_player_level()
	if not underworld2.player:isLevelupAvailable() then
		return
	end

	local v = ui.levelup.create (underworld2.player:getLevel())
	ui.core.view_controller:show_modal_view (v)
	v.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (v)
			if v.result == ui.levelup.RESULT_UPGRADE then
				local upgrades = ui.core.view.builder (ui.upgrades_view).build()
				underworld2.ui.stack:push_window (upgrades)
			end
		end
	}).build (v)
end

function view:destroy()
	if self.location_update_conn_id then
		events.disconnect (net.location_provider():getLocationUpdateSignalProxy(), self.location_update_conn_id)
		self.location_update_conn_id = nil
	end
end

function view:try_handle_back()
	if not self.ui.animations:empty() then
		return
	end
	if self.commodities_shown then
		local card = self.cards[(self.current_tab_index or 0) + 1]
		if card and card:processBackButton() then
			return true
		end
		self:show_map (true)
		return true
	end

	return false
end

function view:show_runners_busy()
	local v = ui.runners_busy.create ()
	ui.core.view_controller:show_modal_view (v)
	v.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (v)
 			if v.result == ui.runners_busy.RESULT_MANAGE then
				ui.couriers.show()
			end
		end
	}).build(v)
end

ui.main_view = ui.main_view or view
