require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"
require "ui/widgets/picture_with_backlight.lua"
require "ui/widgets/courier_widget_wrapper.lua"

require "anim.lua"

local view = setmetatable ({ }, { __index = ui.core.view } )
ui.couriers = view

local function __show_levelup (level_id)
	get_sound_pool():play ("ui/SFX_LevelUp.wav")
	local v = ui.core.view.builder (ui.levelup).build (level_id)
	ui.core.view_controller:show_modal_view (v)
	v.delegate_builder (
	{
		finished = function (v)
			ui.core.view_controller:close_modal_view (v)
			if v.result == ui.levelup.RESULT_UPGRADE then
				local upgrades = ui.core.view.builder (ui.crew_view).build()
				underworld2.ui.stack:push_window (upgrades)
			end
		end
	}).build (v)
end

local function check_levelup_routine (view, player)
	local prev_level_id = player:getLevel()
	return { 
		check = function()
			local new_level_id = player:getLevel()
			-- update upvalue, this will make sequenced check() calls behave consistently  
			prev_level_id = new_level_id
			if new_level_id > prev_level_id then
				view:finalize()
				__show_levelup (new_level_id)
			end
		end,
	}
end

local function __eval_cash_amount_to_collect (player, courier_money, premium)
	local modifier = 0.0
	local bonus 			= player:getBonus ("CashBoostBonus")
	if bonus:get() then
		local modifier_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		modifier = modifier_bonus:getCurrentModifierValue (underworld2.server_time())
	end

	if premium then
		local cash = courier_money
		local cash_bonus = courier_money * modifier
		--print (string.format ("courier_money=%d; cash=%d; cash_bonus=%d(premium)", courier_money, cash, cash_bonus))
		return cash, cash_bonus
	end

	local total_cash = courier_money + courier_money * modifier
	local wallet_available = math.max (0, player:getMaxMoney() - player:getMoney())
	local max_to_collect = math.min (wallet_available, total_cash)
	local cash = math.floor (max_to_collect / (modifier + 1))
	local cash_bonus = math.floor (cash * modifier)

	--print (string.format ("courier_money=%d; cash=%d; cash_bonus=%d(free)", courier_money, cash, cash_bonus))

	return cash, cash_bonus
end

--[[local function __test(courier_money, wallet, modifier)
	local total_cash = courier_money + courier_money * modifier
	local wallet_available = wallet
	local max_to_collect = math.min (wallet_available, total_cash)
	local cash = math.floor (max_to_collect / (modifier + 1))
	local cash_bonus = math.floor (cash * modifier)

	print (string.format ("courier_money=%d; cash=%d; cash_bonus=%d(free)", courier_money, cash, cash_bonus))
end

__test (764, 14332, 1.0)
__test (100, 0, 1.0)
__test (100, 0, 0.0)
__test (100, 100, 0.0)
__test (100, 100, 1.0)
__test (100, 50, 1.0)
__test (100, 3, 1.0)]]

local function __eval_bonus (player, value, bonus_type)
	local bonus	= player:getBonus (bonus_type)
	if bonus:get() then
		local modifier_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		local modifier = modifier_bonus:getCurrentModifierValue (underworld2.server_time())
		if modifier > 0 then
			return math.floor (value * modifier)
		end
	end
end

function view:back()
	self:finalize()
end

local collect = function (self, wrapper)
	local now = underworld2.server_time()

	local player 		= underworld2.player
	local courier		= wrapper.courier
	local progress 		= courier:getProgressValue (now)
	local boost_price 	= courier:getBoostPrice (progress)

	local levelup = check_levelup_routine (self, player)

	local _boost_courier = function (courier)
		get_sound_pool():play ("ui/SFX_Premium_Purchase.wav")
		player:boostCourier (courier:getId())
		player:removeGold (boost_price)
		player:lockUpdates()
		underworld2.network.boost ({ type = 1, id = courier:getId(), gold = boost_price }, function (result) 
			player:unlockUpdates()
		end)
	end

	local _collect_items = function (courier, action_type)
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		if mode == mode_collect then
			if is_purchasing then
				get_sound_pool():play ("ui/SFX_Loot.wav")
			else
				get_sound_pool():play ("ui/SFX_Get_Money.wav")
			end
		end
		
		local player 		= underworld2.player
		local commodity_id 	= courier:getCommodityId()
		
		local free_courier
		if action_type == "DITCH" then
			free_courier = true
		else		
			if is_purchasing then
				local amount			= courier:getItemsAmount()
				local collected_amount 	= math.min (player:getStorage (commodity_id):getAvailable(), amount)
				
				free_courier 			= amount == collected_amount
				
				player:addItemsToStorage (commodity_id, collected_amount)
				courier:setItemsAmount (amount - collected_amount)
			else
				local courier_money		= courier:getMoney()
				local collected_money, bonus_money = __eval_cash_amount_to_collect (player, courier_money, false)
				
				free_courier = collected_money == courier_money
				
				player:addMoney (collected_money + (bonus_money or 0))
				courier:setMoney (courier_money - collected_money)
			end
		end
		
		local rp_earned = courier:getRespectPoints()
		local rp_bonus = __eval_bonus (player, rp_earned, "RespectBoostBonus") or 0
		player:addRespectPoints (rp_earned + rp_bonus)
		
		if free_courier then
			local courier_id = courier:getId()
			player:releaseCourier (courier_id)
			player:lockUpdates()
			underworld2.network.complete_deal (courier_id, action_type, function() 
				player:unlockUpdates()
			end)
		else
			courier:setRespectPoints (0)
			
			player:lockUpdates()
			underworld2.network.complete_deal (courier:getId(), action_type, function() 
				player:unlockUpdates() 
			end)
		end	
		wrapper:on_courier_state_changed()
		levelup:check()		
	end
	
	local _premium_collect_items = function (courier, rent_price)
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		if mode == mode_collect then
			if is_purchasing then
				get_sound_pool():play ("ui/SFX_Loot.wav")
			else
				get_sound_pool():play ("ui/SFX_Get_Money.wav")
			end
		end
		
		local player 		= underworld2.player
		local commodity_id 	= courier:getCommodityId()
		
		if is_purchasing then
			local amount			= courier:getItemsAmount()
			player:addItemsToStorage (commodity_id, amount)
		else
			local money	= courier:getMoney()
			local bonus_money = __eval_bonus (player, money, "CashBoostBonus") or 0
			player:addMoney (money + bonus_money)
		end
		
		local rp_earned = courier:getRespectPoints()
		local rp_bonus = __eval_bonus (player, rp_earned, "RespectBoostBonus") or 0
		player:addRespectPoints (rp_earned + rp_bonus)
		
		local courier_id = courier:getId()
		player:releaseCourier (courier_id)
		player:lockUpdates()
		underworld2.network.premium_complete_deal (courier_id, rent_price, function() 
			player:unlockUpdates()
		end)
		wrapper:on_courier_state_changed()
		levelup:check()
	end

	local upvalue = {}

	local _show_rent_storage_dialogue = function (storage, courier)
		local rent_price 	= player:getRentPrice (courier)
		local instance 		= ui.rent_modal.create (storage, courier, rent_price)
		ui.core.view_controller:show_modal_view (instance)
		instance.delegate_builder (
		{
			finished = function(v)
				ui.core.view_controller:close_modal_view (v)
				if v.result == ui.rent_modal.RESULT_CANCEL then
					return
				elseif v.result == ui.rent_modal.RESULT_DITCH then
					_collect_items (courier, "DITCH")
				elseif v.result == ui.rent_modal.RESULT_RENT then
					if player:getGold() < rent_price then
						self:finalize()
						get_sound_pool():play ("ui/SFX_Cash_Out.wav")
						ui.not_enough_gold.show(rent_price - player:getGold(), function()
							view.show()
							upvalue.show_rent_storage_dialogue (storage,courier)
						end )
						return
					else
						_premium_collect_items (courier, rent_price)
					end
				end
			end,

			on_back = function(v)
				ui.core.view_controller:close_modal_view (v)
			end
		}).build (instance)
	end
	
	upvalue.show_rent_storage_dialogue = _show_rent_storage_dialogue
	
	local _show_rent_vault_dialogue = function (level, ditch_amount, collected_amount)
		local bonus = player:getBonus ("CashBoostBonus")
		local modifier_bonus = bonus:get() and runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		local autoclose_at, event_id
		if modifier_bonus then
			local end_time = modifier_bonus:getClosestModifierEndTime (now)
			if now < end_time then autoclose_at = end_time end
		end

		local conv = underworld2.GoldConversionDatabase():get (level:getId())
		local rent_price 	= math.ceil (ditch_amount * conv:getRentPriceCoeff())
		local instance = ui.core.view.builder (ui.vault_warning).build (level, ditch_amount, collected_amount, rent_price)
		ui.core.view_controller:show_modal_view (instance)
		instance.delegate_builder (
		{
			finished = function (v)
				if event_id ~= nil then underworld2.timed_events:unqueue (event_id) end

				ui.core.view_controller:close_modal_view (v)
				local player = underworld2.player
				if v.result == ui.vault_warning.RESULT_RENT then
					if rent_price <= player:getMaxMoney() then
						_premium_collect_items (courier, rent_price)
					else
						self:finalize()
						get_sound_pool():play ("ui/SFX_Cash_Out.wav")
						ui.not_enough_gold.show(rent_price - underworld2.player:getGold(), function()
							view.show()
							upvalue.show_rent_vault_dialogue (level, ditch_amount, collected_amount)
						end )
					end
				elseif v.result == ui.vault_warning.RESULT_DITCH then
					_collect_items (courier, "DITCH")
				end
			end,
			
			on_back = function (v)
				ui.core.view_controller:close_modal_view (v)
			end
		}).build (instance)

		if autoclose_at then
			event_id = underworld2.timed_events:queue (autoclose_at, function()
				event_id = nil
				instance:finalize()
			end)
		end
	end

	upvalue.show_rent_vault_dialogue = _show_rent_vault_dialogue

	if (now > courier:getEndTime()) and (boost_price == 0) then
		local is_purchasing = courier:getState() == underworld2.CourierStatePurchasing
		if is_purchasing then
			local storage_full = player:getStorage (courier:getCommodityId()):getAvailable() == 0
			if storage_full then
				_show_rent_storage_dialogue (player:getStorage (courier:getCommodityId()), courier)
			else
				_collect_items (courier, "COLLECT")
			end
		else
			local courier_money = courier:getMoney()
			local total_cash, total_bonus = __eval_cash_amount_to_collect (player, courier_money, true)
			local cash, cash_bonus 		  =	__eval_cash_amount_to_collect (player, courier_money, false)
			local collected_amount = cash + cash_bonus
			local ditch_amount = (total_cash + total_bonus) - collected_amount
			if collected_amount == 0 then
				local skill = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP"):getSkill ("WALLET")
				local level_id = player:getSkillLevel (skill:getId())
				local level = skill:getLevel (level_id)
				
				_show_rent_vault_dialogue (level, ditch_amount, collected_amount)
			else
				_collect_items (courier, "COLLECT")
			end
		end
	elseif (boost_price > 0) and (boost_price <= player:getGold()) then
		_boost_courier (courier)
	else
		wrapper.parent:finalize()
		get_sound_pool():play ("ui/SFX_Cash_Out.wav")
		ui.not_enough_gold.show(boost_price - underworld2.player:getGold(), function()
			view.show()
		end)
	end
end

function view.show()
	local instance = ui.core.view.builder (view).build()
	ui.core.view_controller:show_modal_view (instance)
	instance.delegate_builder (
	{
		finished = function (self)
			ui.core.view_controller:close_modal_view (self)
		end,
		on_back = function (self)
			ui.core.view_controller:close_modal_view (self)
		end
	}).build (instance)
end

function view:build_ui()
	local player = underworld2.player

	local widget = ui.Widget()
	widget:setOrigin (0.0, 0.0)
	widget:setSize (1.0, 1.0)
	
	local skill = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP"):getSkill ("COURIERS")
	
	local skill_level_id = player:getSkillLevel (skill:getId())
	
	local character_image = ui.Picture()
	character_image:setTexture (string.format ("ui/characters/ch_runner_modal_%d:filtered", skill_level_id))
	widget:addChild (character_image)
	
	local frame = ui.Frame()
	frame:setBordersTexture ("ui/modal_wide_window:filtered")
	frame:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.vspacing * 2.0, ui.config.vspacing * 2.0)
	
	local frame_layout = ui.VerticalLayout()
	frame_layout:setItemSpacing (ui.config.vspacing)

	local top_right_spacer = ui.Widget()
	top_right_spacer:setHorizontalSizePolicy(ui.SizePolicyFixed)
	
	do
		local top_layout = ui.HorizontalLayout()
		top_layout:setItemSpacing(ui.config.hspacing * 2)
		
		local ico = ui.widgets.picture_with_backlight.create ("ui/ico_runner_modal:filtered", "ui/vault/backlight_vault:filtered")
		top_layout:addWidget (ico)
		
		local lbl_center = ui.VerticalLayout()
		lbl_center:addWidget(ui.Widget())
		local title_label = ui.TextLabel()
		title_label:setTextStyle (ui.text_styles.modal_window_title.build())
		title_label:setText (L "KEY_COURIERS_RUNNERS")
		lbl_center:addWidget(title_label)
		lbl_center:addWidget(ui.Widget())
		
		top_layout:addItem (lbl_center)
		
		local close_button = ui.ImageButton()
		close_button:setTexture ("ui/btn_close:filtered")
		close_button:setSound ("ui/UI_Back.wav")
		close_button:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		events.connect (close_button:getClickedSignalProxy(), function() self:finalize() end)
		top_layout:addWidget (close_button)

		top_layout:addWidget(top_right_spacer)

		frame_layout:addItem (top_layout)
	end
	
	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	local sep_height = sep:getPreferredSize().y
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)
	
	local scroll = ui.ScrollWidget()
	scroll:setScrollType (ui.ScrollTypeVertical)
	scroll:setScrollBarTexture ("ui/scroll:filtered")
	scroll:setVerticalSizePolicy (ui.SizePolicyMaximum)
	
	frame_layout:addWidget (scroll)

	top_right_spacer:setSize(scroll:getLeftMargin() - ui.config.hspacing, 0)

	local scroll_layout = ui.VerticalLayout()
	scroll_layout:setItemSpacing (ui.config.vspacing)
	scroll:setLayout (scroll_layout)
	
	self.courier_wrappers = {} 

	local handler = function(wrapper, as_hire) 
		if not as_hire then
			collect (self, wrapper)
		else
			self:finalize()
			if underworld2.ui.stack:top():tag() ~= "ui.crew_view" then
				local v = ui.core.view.builder (ui.crew_view).build()
				underworld2.ui.stack:push_window (v)
			end
		end
	end

	local couriers_count = player:getCouriersCount()
	local skill = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP"):getSkill ("COURIERS")

	local content_height = -ui.config.vspacing
	local row_height = 0

	for i = 1, skill:getLevelsCount() do
		local courier = player:getCourierAt (i - 1)
		local skill_level = skill:getLevel ((i <= 2 and 1) or (i - 1))
		if skill_level and skill_level:get() then
			local wrapper, w, h = ui.widgets.courier_widget_wrapper.create (courier, skill_level, handler, i, self)
			scroll_layout:addWidget (wrapper.widget)
			table.insert (self.courier_wrappers, wrapper)

			content_height = content_height + ui.config.vspacing + h
			row_height = math.max(row_height, h)

			if not courier:get() then
				local unlock_level = skill_level:getLevelRequired()
				local available = underworld2.player:getLevel() >= unlock_level
				if not available then 
					break
				end
			end
		else
			break
		end
	end
	
	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)
	
	frame:setLayout (frame_layout)
	
	do
		local character_size = character_image:getPreferredSize()

		character_size.y = math.min(0.5, character_size.y)

		local frame_size = frame:getPreferredSize()
		
		local expected_frame_height = frame_size.y + content_height
		local max_frame_height = 0.5
		frame_size.y = math.min(max_frame_height, expected_frame_height)
		
		local sum_height = character_size.y + frame_size.y
		character_image:setOrigin ((1.0 - character_size.x) * 0.5, (1.0 - sum_height) * 0.5)
		character_image:setSize (character_size)
		
		frame:setOrigin (0.0, (1.0 - sum_height) * 0.5 + character_size.y)
		frame:setSize (1.0, frame_size.y)

		-- now, lets check - do we need to start the animation
		if expected_frame_height > (max_frame_height + row_height * 0.5) then
			local scroll_position = math.vec2(0,1)

			scroll:setRelativeScroll(scroll_position)
			scroll:setEnabled(false)

			self.animation = anim.lerp {
				from = 1.0,
				to   = 0.0,
				time = 0.15 * #self.courier_wrappers / 8,
				object = scroll,
				property = function(o,v)
					scroll_position.y = v
					o:setRelativeScroll(scroll_position)
				end,
				on_finished = function()
					self.animation = nil
					scroll:setEnabled(true)
				end
			}

			self.frames_passed = 0
		end
	end
	
	widget:addChild (frame)
	
	return widget
end

function view:destroy()
	for i, v in ipairs (self.courier_wrappers) do
		v:destroy()
	end
end

function view:update(dt)
	local now = underworld2.server_time()
	for i, v in ipairs (self.courier_wrappers) do
		v:update (now)
	end

	if self.animation then
		self.frames_passed = self.frames_passed + 1

		if self.frames_passed > 2 then
			self.animation:update(dt)
		end
	end
end

