require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"

require "ui/widgets/inbox_news_tab.lua"
require "ui/widgets/inbox_messages_tab.lua"
require "ui/widgets/inbox_global_tab.lua"
require "ui/widgets/inbox_cartel_tab.lua"
require "ui/widgets/inbox_private_chat.lua"

require "ui/animations/factory.lua"
require "anim/cubic_bezier.lua"

local view = setmetatable ( {
	TAB_GLOBAL 	= 1,
	TAB_RECENT 	= 2,
	TAB_NEWS 	= 0
 }, { __index = ui.core.view } )
ui.inbox = view

local builder = ui.core.view.builder(view)

function view.create (...)
	local v = builder.build(...)
	return v
end

local function _show_modal(...)
	local v = ui.inbox.create(...)
	v.delegate_builder (
	{
		on_back = function (view)
			v:finalize()
		end,

		finished = function (view)
			ui.core.view_controller:close_modal_view (v)
		end
	}).build (v)

	ui.core.view_controller:show_modal_view (v)
	return v
end

function view.show_chat (profile_id)
	local profile_db = underworld2.ProfileDatabase()
	local profile = profile_db:hasProfile (profile_id) and underworld2.ProfileDatabase():get (profile_id)
	_show_modal(function(v)
		v:set_active_category (view.TAB_RECENT)
		if profile then
			v:push_submenu (ui.widgets.inbox_private_chat.create (v, profile), true)
		end
	end, true)

	--[[elseif DEBUG then
		ErrorPrint (string.format ("Profile with id=%d not found", profile_id))
	end]]
end

function view.show_global()
	_show_modal(function(v)
		v:set_active_category (view.TAB_GLOBAL)
	end)
end

function view.show_recent()
	_show_modal(function(v)
		v:set_active_category (view.TAB_RECENT)
	end)
end

function view.show_news()
	_show_modal(function(v)
		v:set_active_category (view.TAB_NEWS)
	end)
end

local function build_header_widget (on_close)

	local widget= ui.Decorator()
	--W:setPlacingMode (ui.Picture_PlacingModeStretch)
	widget:setTexture("ui/top_hustler/tophustler_top_bg:filtered", ui.Decorator_StyleHStretcherNoBorders)
    widget:setVerticalSizePolicy (ui.SizePolicyMinimum)
	widget:setMargins (ui.config.hspacing * 2, ui.config.hspacing * 2, ui.config.vspacing * 2, ui.config.vspacing * 2)

	local layout = ui.HorizontalLayout()
	layout:setItemSpacing (ui.config.hspacing * 2)

	local pic = ui.Picture()
	pic:setTexture ("ui/messaging/ico_inbox_large:filtered")
	pic:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	layout:addWidget (pic)

	local title = ui.TextLabel()
	title:setTextStyle (ui.text_styles.modal_window_title.build())
	title:setText (L ("KEY_INBOX_WINDOW_TITLE"))
	layout:addWidget (title)

	local close = ui.ImageButton ()
	close:setTexture ("ui/btn_close:filtered")
	close:setSound ("ui/UI_Back.wav")
	close:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	events.connect (close:getClickedSignalProxy(), on_close)
	layout:addWidget (close)

	widget:setLayout (layout)

	return widget
end

function view:push_submenu (menu, hide_nav_bar)
	local stack = self.ui.menu_stack
	if #stack > 0 then
		stack[#stack].widget:removeFromParent()
	end
	table.insert (stack, menu)

	if hide_nav_bar then
		menu.widget:setOrigin (0, self.ui.header:getHeight())
		menu.widget:setSize (1, 1 - self.ui.header:getHeight())
		self.ui.navigation_bar:setHidden(true)
	else
		menu.widget:setOrigin (0, self.ui.header_size.y)
		menu.widget:setSize (1, 1 - self.ui.header_size.y)
		self.ui.navigation_bar:setHidden(false)
	end

	self.ui.root:addChild (menu.widget)
end

function view:pop_submenu()
	local stack = self.ui.menu_stack
	if #stack > 1 then
		local item = stack [#stack]
		item:on_close()
		self.ui.root:removeChild (item.widget)
		table.remove (stack)

		item = stack[#stack]
		self.ui.root:addChild (item.widget)

		self:refresh_active_category()
	end

	self.ui.navigation_bar:setHidden(false)
end 

local function build_buttons(parent, buttons)
	local navigation_bar = ui.Decorator()
	navigation_bar:setMargins(ui.config.hspacing*2,ui.config.hspacing*2,0,0)
	navigation_bar:setVerticalSizePolicy(ui.SizePolicyMinimum)
	navigation_bar:setTexture("ui/top_hustler/tophustler_nav_top_bg:filtered", ui.Decorator_StyleHStretcherNoBorders)

	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing(ui.config.hspacing*2)
	navigation_bar:setLayout(hlayout)

	local button_width = (1 - ui.config.hspacing*4 - ui.config.hspacing*4)/#buttons

	local list = {}

	for i, v in ipairs(buttons) do
		local btn = ui.TextButton()
		btn:setTexture ("ui/top_hustler/btn_glossy_blue:filtered")
		btn:setInactiveTexture ("ui/top_hustler/btn_glossy_disabled:filtered")
		btn:setTextStyle (ui.text_styles.button_regular.build())
		btn:setText (v.name)
		btn:setTag(v.tag)
		btn:setHorizontalSizePolicy(ui.SizePolicyFixed)
		btn:setSize(button_width, 1)
		btn:setInactive(true)

		events.connect(btn:getClickedSignalProxy(), function(sender)
			parent:set_active_category(sender:getTag())
		end)

		hlayout:addWidget(btn)

		list[v.tag] = btn
	end

	navigation_bar:setSize(navigation_bar:getPreferredSize())

	return navigation_bar, list
end

function view:refresh_active_category()
	local active_category = self.active_category
	if not active_category then return end
	local tab = self.ui.tabs[active_category]
	if tab and tab.refresh then tab:refresh() end
end

function view:set_active_category(category)
	local buttons = self.ui.buttons
	local active_category = self.active_category
	if category == active_category then return end
	if active_category then
		buttons[active_category]:setInactive(true)
	end

	buttons[category]:setInactive(false)

	self.active_category = category
	self.ui.tabs[category]:activated()
	self.ui.tab:selectTabAt (category)
end

function view:update(dt)
	self.animations:update(dt)
end

function view:clear_layout()
	local layout = self.ui.layout
	while layout:getItemAt (0) do
		layout:removeItemAt (0)
	end
end

function view:set_loading()
	self:clear_layout()

	local layout = self.ui.layout

	local waiting_widget = ui.WaitingWidget()
	waiting_widget:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	waiting_widget:setAlignment(ui.AlignmentHorizontalCenter)

	local hlayout = ui.HorizontalLayout()
	hlayout:addWidget(ui.Widget())
	hlayout:addWidget(waiting_widget)
	hlayout:addWidget(ui.Widget())

	local wrapper_widget = ui.Widget()
	wrapper_widget:setHorizontalSizePolicy(ui.SizePolicyMaximum)
	wrapper_widget:setVerticalSizePolicy(ui.SizePolicyMaximum)
	wrapper_widget:setLayout(hlayout)

	--layout:addWidget(ui.Widget())
	layout:addWidget(wrapper_widget)
	layout:setHorizontalPolicy(ui.SizePolicyMaximum)
	--layout:addWidget(ui.Widget())
end

function view:build_ui(cb, hide_nav_bar)

	self.ui = { menu_stack = { }, tabs = {} }
	self.animations = ui.animations.factory {}

	local root = ui.Widget()
	root:setOrigin (0, 0)
	root:setSize (1, 1)
	self.ui.root = root

	local header = build_header_widget (function() self:finalize() end)

	header:setOrigin (0, 0)
	header:setSize (math.vec2(1, header:getPreferredSize().y))
	self.ui.header = header
	root:addChild (header)

	local layout = ui.VerticalLayout()
	local navigation_bar, buttons = build_buttons(self, {
		{
			tag = view.TAB_NEWS,
			name = L"KEY_INBOX_NAVIGATION_NEWS"
		},
		{
			tag = view.TAB_GLOBAL,
			name = L"KEY_INBOX_NAVIGATION_GLOBAL"
		},
		{
			tag = view.TAB_RECENT,
			name = L"KEY_INBOX_NAVIGATION_MESSAGES"
		}
	})

	self.ui.navigation_bar = navigation_bar
	self.ui.buttons = buttons

	self.ui.header_size = math.vec2 (1, header:getHeight()+navigation_bar:getHeight())

	navigation_bar:setOrigin(0, header:getHeight())
	root:addChild(navigation_bar)

	--layout:addWidget (navigation_bar)

	self.animations:bind("top_animation", ui.animations.build {
		from = -self.ui.header_size.y,
		to = 0,
		duration = 0.3,
		parent = self,
		subfunction = function(t)
			return (anim.bezier(0.775, 0.645, 0.610, 1.25)) (t, 1)
		end,

		step = function(self, value)
			header:setOrigin(0, value)
			navigation_bar:setOrigin(0, header:getHeight()+value)
		end,

		stop = function(self)
			self.parent:clear_layout()

			local tab = ui.TabWidget()

			tab:lockPageMove()

			self.parent.ui.tab = tab
			local news_tab = ui.widgets.inbox_news_tab.create(self.parent)
			local global_tab = ui.widgets.inbox_global_tab.create(self.parent)
			local messages_tab = ui.widgets.inbox_messages_tab.create(self.parent)

			tab:addTab (news_tab.ui.root)
			tab:addTab (global_tab.ui.root)
			tab:addTab (messages_tab.ui.root)

			self.parent.ui.tabs[view.TAB_NEWS] = news_tab
			self.parent.ui.tabs[view.TAB_GLOBAL] = global_tab
			self.parent.ui.tabs[view.TAB_RECENT] = messages_tab

			layout:addWidget(tab)

			if cb then
				cb(self.parent)
			else
				self.parent:set_active_category (0)
			end
		end
	})


	local main_menu = {  
		widget = ui.Widget(),
		on_close = function () end
	}

	self.ui.layout = layout

	main_menu.widget:setLayout (layout)
	self:push_submenu (main_menu, hide_nav_bar)

	self:set_loading()

	self.animations:update(0)

	return root
end

function view:setCurrentTab (index)
	self:set_active_category(index)
end

function view:destroy()
	local stack = self.ui.menu_stack 
	-- the very last item will be destroyed by the view 
	stack[#stack]:on_close()
	for i = 1, (#stack - 1) do
		stack[i]:on_close()
		ui.destroy (stack[i].widget)
	end
	for i, tab in pairs (self.ui.tabs) do
		tab:on_close()
	end
end
