require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"

local view = setmetatable (
	{
		RESULT_SELL = 0,
		RESULT_CANCEL = 1 
	}, { __index = ui.core.view })
ui.item_sell_confirmation = view

local builder = ui.core.view.builder(view)

function view.create (item)
	local v = builder.build()
	v:set_item (item)
	return v
end

function view:back()
	self.result = view.RESULT_CANCEL
	self:finalize()
end

function view:build_ui()
	self.ui = { }

	local root = ui.Widget()
	root:setOrigin (0.0, 0.0)
	root:setSize (1.0, 1.0)

	local frame = ui.Frame()
	frame:setBordersTexture ("ui/modal_wide_window:filtered")
	frame:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.vspacing * 2.0, ui.config.vspacing * 2.0)

	self.frame = frame
	root:addChild (frame)

	return root
end

function view:set_item (item)
	self.frame:breakLayout()
	self.item = item

	local strings = underworld2.StringsDatabase()

	local frame_layout = ui.VerticalLayout()
	frame_layout:setItemSpacing (ui.config.vspacing)

	do

		local top_layout = ui.HorizontalLayout()
		top_layout:setItemSpacing (ui.config.hspacing)
		
		local ico = ui.widgets.picture_with_backlight.create ("ui/ico_warning_modal:filtered", "ui/backlight_busted:filtered")
		top_layout:addWidget (ico)
		
		local title_label = ui.TextLabel()
		title_label:setTextStyle (ui.text_styles.modal_window_title.build())
		title_label:setText (L "KEY_ITEM_SELL_CONFIRMATION")
		top_layout:addWidget (title_label)
		
		frame_layout:addItem (top_layout)

		do
			local close_button = ui.ImageButton()
			close_button:setTexture ("ui/btn_close:filtered")
			close_button:setSound ("ui/UI_Back.wav")
			close_button:setHorizontalSizePolicy (ui.SizePolicyMinimum)
			events.connect (close_button:getClickedSignalProxy(), function() 
				self.result = view.RESULT_CANCEL
				self:finalize()
			end)
			top_layout:addWidget (close_button)
		end

	end

	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	local sep_height = sep:getPreferredSize().y
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)

	do 
		local hlayout = ui.HorizontalLayout()
		hlayout:setItemSpacing (ui.config.hspacing)
		hlayout:setVerticalPolicy (ui.SizePolicyMinimum)

		hlayout:addWidget (ui.Widget())

		local badge = ui.NotificationBadge()
		badge:setPendingNotifications (item:getLevelRequired())
 		badge:setBadgeTexture ("ui/equipment/frame_item_level:filtered")
 		badge:setBadgeTextStyle (ui.text_styles.badge.build())
 		badge:setHorizontalSizePolicy (ui.SizePolicyMinimum)
 		hlayout:addWidget (badge)

 		local name_label = ui.TextLabel()
 		name_label:setText (strings:get (item:getNameKey()))
 		name_label:setTextStyle (ui.text_styles.normal_hl.build())
 		name_label:setWordWrap (false)
 		name_label:setHorizontalSizePolicy (ui.SizePolicyMinimum)
 		hlayout:addWidget (name_label)

 		hlayout:addWidget (ui.Widget())

 		frame_layout:addItem (hlayout)
	end

	do
		local hlayout = ui.HorizontalLayout()
		hlayout:addWidget (ui.Widget())

		local item_widget = ui.ItemWidget()
		item_widget:setItem (item)
		item_widget:setLocked (false)
		hlayout:addWidget (item_widget)

		hlayout:addWidget (ui.Widget())
		frame_layout:addItem (hlayout)
	end

	do
		local hlayout = ui.HorizontalLayout()
		hlayout:addWidget (ui.Widget())

		ui.widgets.item_widget.add_item_stats ( setmetatable( { widgets = { stats = {} } }, { __index = ui.widgets.item_widget } ), hlayout, item)

		hlayout:addWidget (ui.Widget())
		frame_layout:addItem (hlayout)
	end

	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)

	do
		local hlayout = ui.HorizontalLayout()
		hlayout:addWidget (ui.Widget())

		local sell_price = item:getSellPrice()

		local sell_button = ui.widgets.item_widget.create_price_btn (L"KEY_SELL", sell_price, "ui/ico_cash_small:filtered", ui.widgets.item_widget.btn_colors.sell)
		sell_button:setHorizontalSizePolicy (ui.SizePolicyFixed)
		sell_button:setWidth (math.max (0.5, sell_button:getPreferredSize().x))
		events.connect (sell_button:getClickedSignalProxy(), function()
			self.result = view.RESULT_SELL
			self:finalize()
		end)
		hlayout:addWidget (sell_button)

		hlayout:addWidget (ui.Widget())
		frame_layout:addItem (hlayout)

		local bonus = underworld2.player:getBonus ("CashBoostBonus")
		if bonus:get() then
			local mod_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
			local now = underworld2.server_time()
			local end_time = mod_bonus:getClosestModifierEndTime (now)
			modifier = mod_bonus:getCurrentModifierValue (now)
			if modifier > 0 then
				local bonus_label = ui.RichTextWidget()
				bonus_label:setText (L ("KEY_ITEM_SELL_CASH_BONUS_HINT", ui.format.digit (sell_price * modifier)))
				bonus_label:setDefaultTextStyle (ui.text_styles.small_hl.build())
				bonus_label:setDefaultHorizontalAlignment (1)
				frame_layout:addWidget (ui.Widget())
				frame_layout:addWidget (bonus_label)
			end
		end

	end

	local frame = self.frame

	frame:setLayout (frame_layout)

	local size = frame:getPreferredSize()
	local origin = math.vec2 (0, (1.0 - size.y) * 0.5)
	frame:setOrigin (origin)
	frame:setSize (math.vec2(1.0, size.y))
end

function view:update (dt)
	local autoclose_at = self.autoclose_at
	if not autoclose_at then return end

	if underworld2.server_time() >= autoclose_at then self:finalize() end
end
