require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"
require "ui/widgets/picture_with_backlight.lua"
require "ui/core/library.lua"

require "ui/widgets/quests/quest_state_controller.lua"
require "ui/widgets/quests/reward_controller.lua"
require "ui/widgets/quests/task_bring_commodities_controller.lua"
require "ui/widgets/quests/task_purchase_commodities_controller.lua"
require "ui/widgets/quests/task_sell_commodities_controller.lua"
require "ui/widgets/quests/task_earn_cash_controller.lua"

local view = setmetatable (
	{
		RESULT_CANCEL = 1,
		RESULT_COLLECT_REWARD = 2,
		RESULT_EXPIRED = 3
	}, { __index = ui.core.view } )
ui.quests_view = view

local lui = ui.core.lib

local sandbox =
{
	activate = function (quest_id, handler)
		ui.event_view.show()
		underworld2.network.activate_quest (quest_id, function()
			ui.event_view.close()
			handler()
		end)
	end,

	give = function (quest_id, task_id, amount, handler)
		ui.event_view.show()
		underworld2.network.update_quest_task (quest_id, task_id, amount, function()
			ui.event_view.close()
			handler()
		end)
	end,

	skip = function (quest_id, task_id, price, handler)
		ui.event_view.show()
		underworld2.network.skip_quest_task (quest_id, task_id, price, function()
			ui.event_view.close()
			handler()
		end)
	end,

	complete_quest = function (quest_id, handler)
		ui.event_view.show()
		underworld2.network.complete_quest (quest_id, function()
			ui.event_view.close()
			handler()
		end)
	end
}

function view:back()
	self.result = view.RESULT_CANCEL
	self:finalize()
end

local function __set_quest_scroll (self, scroll_x)
	local scroll = math.vec2 (0, 0)
	local count = #self.quest_controllers
	if count > 1 then
		scroll = math.vec2 ((scroll_x - 1) / (count - 1), 0)
	end
	self.ui.scroll:setRelativeScroll (scroll)
end

local function show_view (quests, selected_quest_id, on_finished)
	local v = ui.core.view.builder (view).build (quests, selected_quest_id)
	ui.core.view.delegate_builder (
	{
		finished = function()
			ui.core.view_controller:close_modal_view (v)
			on_finished(v)
		end
	}).build (v)
	ui.core.view_controller:show_modal_view (v)
end

function view.show (quest_id, on_finished)
	local quests = underworld2.player:getQuests()
	local now = underworld2.server_time()
	local quest_count = 0
	for i, q in pairs (quests) do
		if q:get():getAvailableFrom() > now then
			quests[i] = nil
		end
	end
	local _, quest = next (quests)
	quest = quest_id and quests[quest_id] or quest
	if not quest then
		underworld2.ui.show_message ("No active quests")
		return
	end
	local quest_id = quest:get():getId()
	show_view (quests, quest_id, on_finished)
end

local function __update_navigation (self)
	self.ui.prev:setHidden (self.current_quest_index == 1)
	self.ui.next:setHidden (self.current_quest_index == #self.quest_controllers)
end

local function __update_autoclose_timer (self, quest)
	if (quest:getClass():getName() == "TimedQuest") and quest:isActivated() and not quest:isCompleted() then
		local timed_quest = runtime.cast_to (quest, "underworld2::quests::TimedQuest")
		local expires_at = timed_quest:getExpiresAt()
		if not self.expires_at or expires_at < self.expires_at then
			self.expires_at = expires_at
		end
	end
end

local function __process_quest_activation (self, controller)
	local ptr = controller.ptr
	local state = ui.quest_state_controller.get_quest_state (ptr)
	local anim = anim.parallel
	{
		--anim.sequence
		--{
			anim.delay { time = 0.0, on_finished = function() self.reward_controller:set_quest (ptr) end },
		--}
		controller:build_character_animation(),
		controller:build_state_transition_animation (state)
	}
	if not controller.quest:isActivated() then
		ui.event_view.show()
		underworld2.network.push_back {
			url = "/api/quests/activate.do",
			body = 
			{
				name = "request",
				children = { { name = "questId", text = controller.quest:getId() } }
			},
			on_completed = function (result)
				ui.event_view.close()
				self.anim:push (anim)
				__update_autoclose_timer (self, controller.quest)				
			end,
			on_server_error = function (error)
				ui.event_view.close()
				self.result = view.RESULT_CANCEL,
				self:finalize()
				underworld2.ui.show_message ("Unknown error\n Code:%d\nDescription: %s", error.code, error.description)
			end
		}
	else
		self.anim:push (anim)
	end
end

local function __switch_to_quest (self, index)
	if not self.anim:empty() or (self.current_quest_index == index) then return end

	local from 	= self.quest_controllers [self.current_quest_index]
	local to 	= self.quest_controllers [index]
	local anim = anim.sequence
	{
		from:build_hide_details_animation(),
		anim.delay
		{ 
			time = 0.0,
			on_finished = function()
				to:prepare_to_show (true)
			end
		},
		anim.lerp {
			from = self.current_quest_index, to = index, time = 0.2,
			object = self,
			property = __set_quest_scroll,
			on_finished = function()
				self.current_quest_index = index
				__update_navigation (self)
				__process_quest_activation (self, to)
			end
		},
	}
	self.anim:push (anim)
end

local function __check_current_quest_state (self)
	local controller = self.quest_controllers[self.current_quest_index]

	local new_state = ui.quest_state_controller.get_quest_state (controller.ptr)
	local old_state = controller.state
	if new_state ~= old_state then
		local anim = controller:build_state_transition_animation (old_state, new_state)
		self.anim:push (anim)
	end
end

local function __on_task_skip (self, controller)
	local price = controller.task:getSkipPrice()
	if underworld2.player:getGold() < price then
		self.result = RESULT_CANCEL
		self:finalize()
		get_sound_pool():play ("ui/SFX_Cash_Out.wav")
		ui.not_enough_gold.show (price - underworld2.player:getGold())
		return
	end

	local quest_id = controller.associated_quest_id
	local task_id  = controller.task:getId()

	sandbox.skip (quest_id, task_id, price, function()
		local anim = controller:on_task_state_updated()
		if anim then self.anim:push (anim) end
		__check_current_quest_state (self)
	end)
end

local function __update_all_tasks (self)
	for i, v in ipairs (self.quest_controllers) do
		v:update_tasks()
	end
end

local function __on_task_give (self, controller)
	local quest_id = controller.associated_quest_id
	local task_id  = controller.task:getId()

	sandbox.give (quest_id, task_id, controller.amount, function()
		local anim = controller:on_task_state_updated()
		if anim then self.anim:push (anim) end
		
		__update_all_tasks (self)
		__check_current_quest_state (self)
	end)
end

local __task_connectors = 
{
	["underworld2::quests::BringCommoditiesTask"] = function (self, parent, controller)
		local quest_id = parent.quest:getId()
		controller.associated_quest_id = quest_id
		controller.skip_task:connect (self, __on_task_skip)
		controller.give_items:connect (self, __on_task_give)
	end,
	["underworld2::quests::PurchaseCommoditiesTask"] = function (self, parent, controller)
		local quest_id = parent.quest:getId()
		controller.associated_quest_id = quest_id
		controller.skip_task:connect (self, __on_task_skip)
	end,
	["underworld2::quests::SellCommoditiesTask"] = function (self, parent, controller)
		local quest_id = parent.quest:getId()
		controller.associated_quest_id = quest_id
		controller.skip_task:connect (self, __on_task_skip)
	end,
	["underworld2::quests::EarnCashTask"] = function (self, parent, controller)
		local quest_id = parent.quest:getId()
		controller.associated_quest_id = quest_id
		controller.skip_task:connect (self, __on_task_skip)
	end,
}

local function __connect_task (self, parent, controller)
	local task = controller.task
	local connector = __task_connectors [task:getClass():getScopedName()]
	if connector then
		connector (self, parent, controller)
	end
end

local function __animate_screen_appear (self, state)
	local set_button_scale = function (self, t)
		self.ui.close:setScale (t)
		self.ui.prev:setScale (t)
		self.ui.next:setScale (t)
	end

	local quest_controller = self.quest_controllers[self.current_quest_index]

	set_button_scale (self, 0.6)

	local anim = anim.sequence
	{
		--insert delay, this will give time to TaskManager to complete all resource loading related tasks
		anim.delay 
		{ 
			time = 0.0,
			on_finished = function() 
				self.ui.close:setHidden (false)
				__update_navigation (self)
			end
		},
		anim.parallel
		{
			anim.sequence 
			{
				anim.lerp
				{
					from = 0.6, to = 1.1, time = 0.1,
					object = self, property = set_button_scale
				},
				anim.lerp
				{
					from = 1.1, to = 1.0, time = 0.1,
					object = self, property = set_button_scale,
				}
			},
			quest_controller:build_character_animation(),
			quest_controller:build_state_transition_animation (state)
		}
	}
	self.anim:push (anim)
end

function view:build_ui (quests, selected_quest_id)
	self.anim = anim.queue()

	local widget = ui.load_ui ("ui/modal/quests.ui")
	self.ui = ui.build_children_map (widget)

	events.connect (self.ui.close:getClickedSignalProxy(), function()
		self.result = view.RESULT_CANCEL
		self:finalize()
	end)

	events.connect (self.ui.next:getClickedSignalProxy(), function()
		__switch_to_quest (self, self.current_quest_index + 1)
	end)
	
	events.connect (self.ui.prev:getClickedSignalProxy(), function()
		__switch_to_quest (self, self.current_quest_index - 1)
	end)

	self.current_quest_index = nil

	self.quest_controllers = { }

	local autoclose_at

	for quest_id, ptr in pairs (quests) do
		if not self.current_quest_index and selected_quest_id == quest_id then
			self.current_quest_index = #self.quest_controllers + 1
		end

		local quest = ptr:get()

		__update_autoclose_timer (self, quest)

		local controller = ui.quest_state_controller.create (ptr)

		self.ui.scroll:addChild (controller.ui.root)

		controller.continue:connect (self, function (self, controller)
			local old_state = ui.quest_state_controller.MODE_FIRST_TIME_SHOW
			local new_state = ui.quest_state_controller.MODE_IN_PROGRESS
			local anim = controller:build_state_transition_animation (old_state, new_state)
			self.anim:push (anim)
		end)

		controller.collect:connect (self, function (self, controller)
			self.selected_quest = controller.ptr
			self.result = view.RESULT_COLLECT_REWARD
			self:finalize()
		end)

		for i, v in ipairs (controller.task_controllers) do
			__connect_task (self, controller, v)
		end

		table.insert (self.quest_controllers, controller)
	end
	self.current_quest_index = self.current_quest_index or 1

	__set_quest_scroll (self, self.current_quest_index)

	self.quest_controllers[self.current_quest_index]:prepare_to_show (true)

	self.ui.close:setHidden (true)
	self.ui.prev:setHidden (true)
	self.ui.next:setHidden (true)

	local selected_quest_ptr = self.quest_controllers[self.current_quest_index].ptr
	local selected_quest = self.quest_controllers[self.current_quest_index].quest
	if not selected_quest:isActivated() then
		widget:setEnabled (false)
		--otherwise quests view will be on top of the event view
		underworld2.timed_events:queue (0.0, function()
			widget:setEnabled (true)
			ui.event_view.show()
			underworld2.network.push_back {
				url = "/api/quests/activate.do",
				body = 
				{
					name = "request",
					children = { { name = "questId", text = selected_quest:getId() } }
				},
				on_completed = function (result)
					ui.event_view.close()
					__animate_screen_appear (self, ui.quest_state_controller.MODE_FIRST_TIME_SHOW)
				end,
				on_server_error = function (error)
					ui.event_view.close()
					self.result = view.RESULT_CANCEL,
					self:finalize()
					underworld2.ui.show_message ("Unknown error\n Code:%d\nDescription: %s", error.code, error.description)
				end
			}
		end)
	else
		__animate_screen_appear (self, ui.quest_state_controller.get_quest_state (selected_quest_ptr))
	end

	self.reward_controller = ui.reward_controller.create()
	self.reward_controller.ui.root:setAnchor (math.vec2 (0.5, 0.5 - ui.config.vspacing * 2), math.vec2 (0.5, 1.0))
	self.reward_controller:set_quest (selected_quest_ptr)

	widget:addChild (self.reward_controller.ui.root)

	return widget
end

function view:on_destroy()

end

function view:update (dt)
	self.anim:update (dt)
	self.reward_controller:update (dt)

	if self.expires_at then
		local now = underworld2.server_time()
		if now >= self.expires_at then
			self.result = view.RESULT_EXPIRED
			self:finalize()
			return
		end
	end
end


