require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"
require "ui/widgets/picture_with_backlight.lua"
require "network/mailto.lua"
require "social/facebook_sharing.lua"

local view = setmetatable (
	{
		RESULT_GOTO_SHOP = 0,
		RESULT_CANCEL = 1 
	}, { __index = ui.core.view })
ui.server_error = view

local builder = ui.core.view.builder(view)

function view.create (...)
	local obj = builder.build(...)
	return obj
end

local messages = {
	[true] = L "KEY_SERVER_ERROR_MODAL_SERVER_UNAVAILBALE",
	[false] = L "KEY_SERVER_ERROR_MODAL_CRITICAL_ERROR"
}

local addresses = {
	[true] = {
		"anton@a-steroids.com", "dmitry@a-steroids.com", "vitaly.sverchinsky@a-steroids.com"
	},
	[false] = "support@underworld.zone"
}

local feedback_body = [[
Please do not remove this system information:
----------------------------------
Device-Type: %s
Device-OS: %s
Build: %i
Player ID: %s
Error: %s
----------------------------------

Tell us how we can help you here. You can also add screenshots or other attachments to the message.
]]

function view:back()
	self.result = view.RESULT_CANCEL
	self:finalize()
end

function view:build_ui (server_dead, error_message, connection_failed)
	self.ui = {}

	local strings = underworld2.StringsDatabase()

	local widget = ui.Widget()
	widget:setOrigin (0.0, 0.0)
	widget:setSize (1.0, 1.0)
	
	local character_image = ui.Picture()
	character_image:setTexture ("ui/characters/ch_advisor_modal:filtered")
	widget:addChild (character_image)

	local frame = ui.Frame()
	frame:setBordersTexture ("ui/modal_wide_window:filtered")
	frame:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.vspacing * 2.0, ui.config.vspacing * 2.0)
	
	local frame_layout = ui.VerticalLayout()
	frame_layout:setItemSpacing (ui.config.vspacing)

	do
		local top_layout = ui.HorizontalLayout()
		top_layout:setItemSpacing (ui.config.hspacing)
		
		local vl = ui.VerticalLayout()
		vl:addWidget(ui.Widget())
		local ico = ui.widgets.picture_with_backlight.create ("ui/ico_warning_modal:filtered", "ui/backlight_busted:filtered")
		vl:addWidget(ico)
		vl:addWidget(ui.Widget())
		top_layout:addItem (vl)
		
		local title_label = ui.TextLabel()
		title_label:setTextStyle (ui.text_styles.modal_window_title.build())
		title_label:setText (L "KEY_SERVER_ERROR_MODAL_TITLE")
		title_label:setRowsCount (1)
		top_layout:addWidget (title_label)
		
		frame_layout:addItem (top_layout)

		do
			local close_button = ui.ImageButton()
			close_button:setTexture ("ui/btn_close:filtered")
			close_button:setSound ("ui/UI_Back.wav")
			close_button:setHorizontalSizePolicy (ui.SizePolicyMinimum)
			close_button:setHidden(connection_failed and not server_dead or false)
			events.connect (close_button:getClickedSignalProxy(), function() 
				self.result = view.RESULT_CANCEL
				self:finalize()
			end)
			top_layout:addWidget (close_button)
		end
	end
	
	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	local sep_height = sep:getPreferredSize().y
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)

	do
		local text_frame = ui.Frame()
		text_frame:setBordersTexture ("ui/frame_modal_light:filtered")
		
		local margins = { x = text_frame:getLeftMargin() * 0.5, y = text_frame:getTopMargin() }
		text_frame:setMargins (margins.x, margins.x, margins.y, margins.y)
		
		local text_layout = ui.VerticalLayout()
		
		local title_label = ui.TextLabel()
		title_label:setTextStyle (ui.text_styles.modal_character_name.build())
		title_label:setText (L "KEY_ADVISOR_FULL")
		text_layout:addWidget (title_label)

		local message = connection_failed and L "KEY_SERVER_ERROR_MODAL_CONNECTION_FAILED" or messages[server_dead or false]

		local text_label = ui.TextLabel()
		text_label:setTextStyle (ui.text_styles.small_hl.build())
		text_label:setText (message)
		text_label:setRowsCount (2)
		text_layout:addWidget (text_label)

		text_frame:setLayout (text_layout)
		frame_layout:addWidget (text_frame)
	end

	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)

	do
		local v_sep = ui.Widget()
		v_sep:setSize(1.0, 0.002)
		v_sep:setVerticalSizePolicy (ui.SizePolicyFixed)
		frame_layout:addWidget (v_sep)

		local top_layout = ui.HorizontalLayout()
		top_layout:setItemSpacing (ui.config.hspacing)

		if not server_dead and not connection_failed then
			top_layout:addWidget (ui.Widget())

			local tag = underworld2.player and tostring(underworld2.player:id()) or os.device:getDeviceUIDString()
 
			local device = os.device:getDeviceIDString() --getDeviceNameString()
			local deviceOS = os.device:getDeviceOSVersionString()
			local build =  VERSION -- getDeviceAppVersionNumber()

			local btn = ui.TextButton()
			btn:setTexture ("ui/btn_lower_normal:filtered")
			btn:setTextStyle (ui.text_styles.button_regular.build())
			btn:setText (L "KEY_SERVER_ERROR_MODAL_BUTTON_SEND_REPORT")
			btn:setSize(0.5, 0.1)
			btn:setHorizontalSizePolicy (ui.SizePolicyFixed)
			btn:setTag (view.RESULT_GOTO_SHOP)
			top_layout:addWidget (btn)

			events.connect (btn:getClickedSignalProxy(), function()
				self:finalize()
				network.mailto {
					to = addresses[DEBUG or false],
					subject = "Underworld Support Request",
					body = string.format (feedback_body, device, deviceOS, build, tag, error_message or "unknown")
				}
			end)

			top_layout:addWidget (ui.Widget())
		elseif not server_dead then
			local btn = ui.TextButton()
			btn:setTexture ("ui/btn_lower_normal:filtered")
			btn:setTextStyle (ui.text_styles.button_regular.build())
			btn:setText (L "KEY_SERVER_ERROR_MODAL_BUTTON_TRY_AGAIN")
			--btn:setSize(0.5, 0.1)
			--btn:setHorizontalSizePolicy (ui.SizePolicyFixed)
			btn:setTag (view.RESULT_GOTO_SHOP)
			top_layout:addWidget (btn)

			local settings_btn = ui.TextButton()
			settings_btn:setTexture ("ui/btn_lower_normal:filtered")
			settings_btn:setTextStyle (ui.text_styles.button_regular.build())
			settings_btn:setText (L "KEY_SERVER_ERROR_SETTINGS")
			--settings_btn:setSize(0.5, 0.1)
			--settings_btn:setHorizontalSizePolicy (ui.SizePolicyFixed)
			settings_btn:setTag (view.RESULT_GOTO_SHOP)
			top_layout:addWidget (settings_btn)

			events.connect (btn:getClickedSignalProxy(), function()
				self:finalize()
			end)

			events.connect (settings_btn:getClickedSignalProxy(), function()
				os.device:openOSSettings(os.settings.network)

				self:finalize()
			end)
		else
			top_layout:addWidget (ui.Widget())

			local button = underworld2.facebook_sharing.create_button(L"KEY_SERVER_ERROR_MODAL_BUTTON_DETAILS", 0, function()
					self:finalize()
					os.device:openURL("http://underworld.zone/fb");
				end)

			button:setSize(0.5, 0.1)
			button:setHorizontalSizePolicy (ui.SizePolicyFixed)

			top_layout:addWidget (button)
			top_layout:addWidget (ui.Widget())
		end
		
		frame_layout:addItem (top_layout)

		v_sep = ui.Widget()
		v_sep:setSize(1.0, 0.002)
		v_sep:setVerticalSizePolicy (ui.SizePolicyFixed)
		frame_layout:addWidget (v_sep)
	end
	
	frame:setLayout (frame_layout)
	
	do
		local character_size = character_image:getPreferredSize()
		local frame_size = frame:getPreferredSize()

		local max_char_height = math.min (1.0 - frame_size.y - 0.05, character_size.y)
		
		character_size.x = character_size.x * max_char_height / character_size.y
		character_size.y = max_char_height
		
		character_image:setOrigin ((1.0 - character_size.x) * 0.5, 1.0 - character_size.y - frame_size.y - 0.05)
		character_image:setSize (character_size)
		character_image:setVerticalSizePolicy (ui.SizePolicyFixed)
		character_image:setHorizontalSizePolicy (ui.SizePolicyFixed)
		character_image:setPlacingMode(ui.Picture_PlacingModeStretch)
		
		frame:setOrigin (0.0, 1.0 - frame_size.y - 0.05)
		frame:setSize (1.0, frame_size.y)
	end
	
	widget:addChild (frame)

	return widget
end
