require "ui/core/view.lua"
require "ui/config.lua"
require "ui/text_styles.lua"
require "ui/format.lua"
require "ui/widgets/picture_with_backlight.lua"
require "ui/widgets/leaderboard_row.lua"
require "ui/modal/inbox.lua"
require "ui/animations/factory.lua"
require "social/facebook_sharing.lua"
require "social_ids.lua"
require "anim/cubic_bezier.lua"

local view = setmetatable (
	{
		CATEGORY_ALL_TIME = 1,
		CATEGORY_FRIENDS = 2
	}, { __index = ui.core.view })

local builder = ui.core.view.builder (view)
ui.top_hustlers = view

function view:back()
	self:finalize()
end

function view.create (...)
	local obj = builder.build(...)
	return obj
end

local function build_header(parent)
	local W = ui.Decorator()
	--W:setPlacingMode (ui.Picture_PlacingModeStretch)
	W:setTexture("ui/top_hustler/tophustler_top_bg:filtered", ui.Decorator_StyleHStretcherNoBorders)
    W:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, 0,0)
    W:setVerticalSizePolicy (ui.SizePolicyMinimum)

	local vlayout = ui.VerticalLayout()

	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing (ui.config.hspacing)

	local ico = ui.widgets.picture_with_backlight.create ("ui/icons/ico_fame_hustler_header:filtered", "ui/backlight_storagefull:filtered")
	hlayout:addWidget (ico)

	local title_label = ui.TextLabel()
	title_label:setTextStyle (ui.text_styles.modal_window_title.build())
	title_label:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalLeft)
	title_label:setText (L"TOP_HUSTLERS_TITLE")
	title_label:setHorizontalSizePolicy (ui.SizePolicyMaximum)
	hlayout:addWidget (title_label)

	local close = ui.ImageButton ()
	close:setTexture ("ui/btn_close:filtered")
	close:setSound ("ui/UI_Back.wav")
	close:setOrigin (0, 0)
	hlayout:addWidget (close)

	events.connect (close:getClickedSignalProxy(), function()
		parent:finalize()
	end)

	vlayout:addWidget(ui.Widget())
	vlayout:addItem(hlayout)
	vlayout:addWidget(ui.Widget())

	W:setLayout(vlayout)

	W:setSize (1, W:getPreferredSize().y)
	W:setOrigin (0, 0)

	return W
end

function view:set_active_category(category)
	local active_category = self.active_category
	if active_category == category then
		return
	end

	self:clear_layout()
	self.active_category = category

	local buttons = self.ui.category_buttons
	if active_category then
		buttons[active_category]:setInactive(true)
	end

	buttons[category]:setInactive(false)

	local bottom_animation = ui.animations.build {
		from = 1,
		to = 1 - self.ui.bottom_navigation:getHeight(),
		duration = 0.3,
		parent = self,
		subfunction = function(t)
			return (anim.bezier(0.775, 0.645, 0.610, 1.25)) (t, 1)
		end,

		step = function(self, value)
			self.parent.ui.bottom_navigation:setOrigin(0, value)
		end,

		stop = function()
			if category == view.CATEGORY_ALL_TIME then
				if #self.leaderboards == 0 then
					self:reload_leaderboard()
				else
					self:set_leaderboard(self.current_tab)
				end
			elseif category == view.CATEGORY_FRIENDS then
				if (#self.friends == 0) and (underworld2.social_ids.logged_in()) then
					self:reload_friends()
				else
					self:set_friends()
				end
			end
		end
	}

	self:set_loading()

	if category == view.CATEGORY_FRIENDS then
		bottom_animation.from, bottom_animation.to = bottom_animation.to, bottom_animation.from
	else
		bottom_animation.to = 1 - self.ui.bottom_navigation:getHeight()
	end

	self.animations:bind("bottom_animation", bottom_animation)
	self.animations:update(0)
end

local function build_categories(parent)
	local navigation_bar = ui.Decorator()
	navigation_bar:setMargins(ui.config.hspacing*2,ui.config.hspacing*2,0,0)
	navigation_bar:setVerticalSizePolicy(ui.SizePolicyMinimum)
	navigation_bar:setTexture("ui/top_hustler/tophustler_nav_top_bg:filtered", ui.Decorator_StyleHStretcherNoBorders)

	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing(ui.config.hspacing*2)
	navigation_bar:setLayout(hlayout)

	local button_width = (1 - ui.config.hspacing*4 - ui.config.hspacing*2)/2

	local all_time_btn = ui.TextButton()
	all_time_btn:setTexture ("ui/top_hustler/btn_glossy_blue:filtered")
	all_time_btn:setInactiveTexture ("ui/top_hustler/btn_glossy_disabled:filtered")
	all_time_btn:setTextStyle (ui.text_styles.button_regular.build())
	all_time_btn:setText (L "TOP_HUSTLERS_ALL_TIME")
	all_time_btn:setTag(view.CATEGORY_ALL_TIME)
	all_time_btn:setHorizontalSizePolicy(ui.SizePolicyFixed)
	all_time_btn:setSize(button_width, 1)

	local friends_btn = ui.TextButton()
	friends_btn:setTexture ("ui/top_hustler/btn_glossy_blue:filtered")
	friends_btn:setInactiveTexture ("ui/top_hustler/btn_glossy_disabled:filtered")
	friends_btn:setTextStyle (ui.text_styles.button_regular.build())
	friends_btn:setText (L "TOP_HUSTLERS_FRIENDS")
	friends_btn:setTag(view.CATEGORY_FRIENDS)
	friends_btn:setHorizontalSizePolicy(ui.SizePolicyFixed)
	friends_btn:setSize(button_width, 1)

	local buttons = {
		[view.CATEGORY_ALL_TIME] = all_time_btn,
		[view.CATEGORY_FRIENDS] = friends_btn
	}

	for k, v in pairs(buttons) do
		events.connect(v:getClickedSignalProxy(), function(sender)
			parent:set_active_category(sender:getTag())
		end)

		v:setInactive(true)
	end

	hlayout:addWidget(all_time_btn)
	hlayout:addWidget(friends_btn)

	navigation_bar:setSize(navigation_bar:getPreferredSize())

	return {
		widget = navigation_bar,
		buttons = buttons
	}
end

local place_icons = {
	[1] = "ui/top_hustler/wreath_gold:filtered",
	[2] = "ui/top_hustler/wreath_silver:filtered",
	[3] = "ui/top_hustler/wreath_bronze:filtered"
}

local place_labels = {
	[1] = L"TOP_HUSTLERS_PLACE_1ST",
	[2] = L"TOP_HUSTLERS_PLACE_2ND",
	[3] = L"TOP_HUSTLERS_PLACE_3RD"
}

local function build_winner_widget(parent, profile, place, rp, animate)
	local scale = 1.0
	if place ~= 1 then scale = 0.85 end

	local strings = underworld2.StringsDatabase()
	local level = underworld2.LevelsDatabase():get (profile:getLevel())

	if not level:get() then
		ErrorPrint (string.format ("can't find level %d description(player id=%d)", profile:getLevel(), profile:id()))
	end

	local root = ui.Widget()

	local W = ui.Widget()

	local vlayout = ui.VerticalLayout()
	--vlayout:setHorizontalPolicy(ui.SizePolicyMinimum)
	vlayout:setItemSpacing(ui.config.vspacing * 0.5)
	W:setLayout(vlayout)

	if place ~= 1 then
		local w = ui.Widget()
		w:setVerticalSizePolicy(ui.SizePolicyMaximum)
		vlayout:addWidget(w)
	end

	local name = ui.TextLabel()
	name:setTextStyle(ui.text_styles.small_hl.build())
	name:setElide(true)
	name:setTextAlignment(ui.AlignmentHorizontalCenter + ui.AlignmentVerticalCenter)
	--name:setHorizontalSizePolicy(ui.SizePolicyFixed)
	vlayout:addWidget(name)

	local animations = {}

	do
		local wrapper = ui.Widget()

		local hlayout = ui.HorizontalLayout()
		wrapper:setLayout(hlayout)
		hlayout:setItemSpacing (ui.config.hspacing)

		local badge = ui.NotificationBadge()
 		badge:setBadgeTexture ("ui/ico_level:filtered")
 		badge:setPendingNotifications (profile:getLevel())
 		badge:setBadgeTextStyle (ui.text_styles.badge.build())
 		badge:setHorizontalSizePolicy (ui.SizePolicyMinimum)
 		
		local level_label = ui.TextLabel ()
		level_label:setText (strings:get (level:getNameKey()))
		level_label:setTextStyle (ui.text_styles.hint_grey_small.build())
		--level_label:setElide (true)
		level_label:setWordWrap (false)

		hlayout:addWidget (ui.Widget())
		hlayout:addWidget (badge)
		hlayout:addWidget (level_label)
		hlayout:addWidget (ui.Widget())

		vlayout:addWidget (wrapper)
	end

	do
		local wrapper = ui.Widget()

		local rays = ui.Picture()
		rays:setTexture("ui/top_hustler/rays_top_hustler:filtered")
		rays:setHorizontalSizePolicy(ui.SizePolicyFixed)
		rays:setVerticalSizePolicy(ui.SizePolicyFixed)
		rays:setSize(rays:getPreferredSize())
		local rays_color = render.color(255,255,255,127)
		rays:setColor(rays_color)

		local rays_rotation_animation = ui.animations.build {
			from = 0,
			to = -math.pi * 2,
			duration = 10,
			times = -1,

			step = function(self, value)
				rays:setRotation(value)
			end
		}

		local rays_flash_animation = ui.animations.build {
			from = rays_color.a,
			to = 255,
			times = 2,
			oscillate = true,
			duration = 0.3,

			step = function(self, value)
				rays:setColor(render.color(255,255,255,value))
			end
		}

		if animate then
			wrapper:addChild(rays)

			animations["rays_rotation_animation"] = rays_rotation_animation
			animations["rays_flash_animation"] = rays_flash_animation
		end

		local ribbon_back = ui.Picture()
		ribbon_back:setTexture("ui/top_hustler/ribbon_top_background:filtered")
		ribbon_back:setSize(ribbon_back:getPreferredSize() * scale)
		ribbon_back:setPlacingMode(ui.Picture_PlacingModeStretch)
		ribbon_back:setHorizontalSizePolicy(ui.SizePolicyFixed)
		ribbon_back:setVerticalSizePolicy(ui.SizePolicyFixed)
		wrapper:addChild(ribbon_back)

    	local avatar = ui.AvatarWidget()
		avatar:setScale(0.7 * scale)
    	avatar:setProfile (profile)
    	avatar:setAvatarSize (underworld2.Avatar_SizeLarge)
		avatar:setSize(avatar:getPreferredSize())

		local player = underworld2.player
		local player_id = player:id()
		if player_id ~= profile:id() then
			events.connect (avatar:getClickedSignalProxy(), function()
				ui.inbox.show_chat (profile:id())
			end)
		end

		wrapper:addChild(avatar)

		local ribbon_front = ui.Picture()
		ribbon_front:setTexture("ui/top_hustler/ribbon_top_foreground:filtered")
		ribbon_front:setSize(ribbon_front:getPreferredSize() * scale)
		ribbon_front:setPlacingMode(ui.Picture_PlacingModeStretch)
		ribbon_front:setHorizontalSizePolicy(ui.SizePolicyFixed)
		ribbon_front:setVerticalSizePolicy(ui.SizePolicyFixed)
		wrapper:addChild(ribbon_front)

		local av_h = avatar:getHeight() * 1.1

		ribbon_back:setOrigin(0, av_h-ribbon_back:getHeight())
		ribbon_front:setOrigin((ribbon_back:getWidth()-ribbon_front:getWidth())/2, av_h-ribbon_back:getHeight()/2)
		avatar:setOrigin((ribbon_back:getWidth()-avatar:getWidth())/2, 0)

		local place_label = ui.TextLabel ()
		place_label:setText (place_labels[place])
		place_label:setTextStyle (ui.text_styles.winner_place_label.build())
		place_label:setWordWrap (false)
		place_label:setSize(place_label:getPreferredSize())
		place_label:setZoom(scale)
		place_label:setOrigin((ribbon_front:getWidth()-place_label:getWidth())/2,(ribbon_front:getHeight()-place_label:getHeight())/2)
		ribbon_front:addChild(place_label)

		local wreath = ui.Picture()
		wreath:setTexture(place_icons[place])
		wreath:setSize(wreath:getPreferredSize() * scale)
		wreath:setPlacingMode(ui.Picture_PlacingModeStretch)
		wreath:setHorizontalSizePolicy(ui.SizePolicyFixed)
		wreath:setVerticalSizePolicy(ui.SizePolicyFixed)
		wreath:setOrigin((ribbon_front:getWidth()-wreath:getWidth())/2,(ribbon_front:getHeight()-wreath:getHeight())/2)
		ribbon_front:addChild(wreath)

		wrapper:setHorizontalSizePolicy(ui.SizePolicyFixed)
		wrapper:setVerticalSizePolicy(ui.SizePolicyFixed)
		wrapper:setSize(ribbon_back:getWidth(), ribbon_front:getOrigin().y+(wreath:getHeight()-ribbon_front:getHeight()/2))

		rays:setOrigin((wrapper:getWidth()-rays:getWidth())/2,(wrapper:getHeight()-rays:getHeight())/2)

		local hlayout = ui.HorizontalLayout()

		hlayout:addWidget(ui.Widget())
		hlayout:addWidget(wrapper)
		hlayout:addWidget(ui.Widget())

		vlayout:addItem(hlayout)
	end

	do
		local hlayout = ui.HorizontalLayout()
		hlayout:setItemSpacing(ui.config.hspacing)

		local ico = ui.Picture()
		ico:setTexture("ui/icons/ico_fame_hustler_list:filtered")
		ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		ico:setVerticalSizePolicy (ui.SizePolicyMinimum)
		--ico:setSize(ico:getPreferredSize())

		local points = ui.TextLabel()
		points:setTextStyle(ui.text_styles.normal_hl.build())
		points:setText(rp)
		points:setWordWrap (false)

		hlayout:addWidget(ui.Widget())
		hlayout:addWidget(ico)
		hlayout:addWidget(points)
		hlayout:addWidget(ui.Widget())

		vlayout:addItem(hlayout)
	end

	--W:setSize(W:getPreferredSize())
	--W:setOrigin(0, 0)

	name:setText(profile:getNickname())
	--name:setWidth(W:getWidth())

	return {
		widget = W,
		animations = animations
	}
end

local function build_winners(parent)
	local W = ui.Widget()
	W:setMargins(0, 0, 0, 0)

	local hlayout = ui.HorizontalLayout()
	hlayout:setItemSpacing(ui.config.hspacing * 4)
	W:setLayout(hlayout)

	W:setOrigin(0,0)
	W:setSize(1.0, W:getPreferredSize().y)
	
	return {
		widget = W,
		layout = hlayout
	}
end

local function build_line(parent)
	local W = ui.Picture()
	W:setTexture("ui/top_hustler/line_white:filtered")
	W:setPlacingMode(ui.Picture_PlacingModeStretch)
	W:setSize(1.0-ui.config.hspacing*4, 0.004)
	W:setHorizontalSizePolicy(ui.SizePolicyFixed)
	W:setVerticalSizePolicy(ui.SizePolicyFixed)
	return W
end

local function build_facebook_row(parent, h_margin, login_function)
	local W = ui.Widget()

	local frame = ui.Frame()
	frame:setBordersTexture("ui/top_hustler/frame_tophustler:filtered")
	frame:setHorizontalSizePolicy(ui.SizePolicyFixed)
	frame:setRelativeSize(true)

	local vlayout = ui.VerticalLayout()
	vlayout:setItemSpacing(ui.config.vspacing*2)
	vlayout:addWidget(ui.Widget())
	--vlayout:addWidget(ui.Widget())

	local texts = {
	}

	local handler = function() end

	if underworld2.social_ids.logged_in() then
		texts.hint = L"TOP_HUSTLERS_INVITE_FRIENDS"
		texts.action = L"TOP_HUSTLERS_FACEBOOK_INVITE"

		handler = function()
			social.facebook.web_dialog("apprequests", {message = L"KEY_FB_SHARING_APP_NAME"},
				function(state, payload)
				end)
		end
	else
		texts.hint = L"TOP_HUSTLERS_CONNECT_WITH_FACEBOOK"
		texts.action = L"TOP_HUSTLERS_FACEBOOK_CONNECT"

		handler = function()
			underworld2.social_ids.login("FACEBOOK", function(successful, err)
				if not successful then
					return true
				else
					if login_function then
						login_function(parent)
					end
				end
			end)
		end
	end

	local vvlayout = ui.VerticalLayout()
	vvlayout:setItemSpacing(ui.config.vspacing)
	vvlayout:addWidget(ui.Widget())

	do
		local label = ui.TextLabel()
		label:setText(texts.hint)
		label:setTextStyle(ui.text_styles.normal_hl.build())
		label:setTextAlignment(ui.AlignmentHorizontalCenter)
		label:setHorizontalSizePolicy(ui.SizePolicyFixed)
		label:setWidth(0.8)
		--label:setVerticalSizePolicy(ui.SizePolicyMinimum)
		label:setRowsCount(label:getRowsForWidth(label:getWidth()))

		local hhlayout = ui.HorizontalLayout()
		hhlayout:addWidget(ui.Widget())
		hhlayout:addWidget(label)
		hhlayout:addWidget(ui.Widget())

		vvlayout:addItem(hhlayout)
	end

	do
		local button = underworld2.facebook_sharing.create_button(texts.action, 0, handler)
		button:setHorizontalSizePolicy(ui.SizePolicyFixed)
		button:setWidth(0.6)

		local hhlayout = ui.HorizontalLayout()
		hhlayout:addWidget(ui.Widget())
		hhlayout:addWidget(button)
		hhlayout:addWidget(ui.Widget())

		vvlayout:addItem(hhlayout)
	end

	vvlayout:addWidget(ui.Widget())

	local vvwidget = ui.Widget()
	vvwidget:setLayout(vvlayout)
	vvwidget:setVerticalSizePolicy(ui.SizePolicyMinimum)

	vlayout:addWidget(vvwidget)
	vlayout:addWidget(ui.Widget())
	--vlayout:addWidget(ui.Widget())

	frame:setLayout(vlayout)
	frame:setOrigin(h_margin,0)
	frame:setSize(1-h_margin*2, vvwidget:getPreferredSize().y + 0.06)

	W:addChild(frame)
	W:setHorizontalSizePolicy(ui.SizePolicyFixed)
	W:setVerticalSizePolicy(ui.SizePolicyFixed)
	W:setSize(frame:getSize())

	return {
		widget = W,
		frame = frame
	}
end

local function build_leaderboard(parent)
	local W = ui.ScrollWidget()
	W:setScrollType (ui.ScrollTypeVertical)
	W:setMinTouchArea (math.vec2(0,0))

	local vlayout = ui.VerticalLayout()
	W:setLayout (vlayout)
	vlayout:setItemSpacing(ui.config.vspacing)

	local h_margin = ui.config.hspacing*2
	vlayout:addWidget(ui.Widget())

	W:setOrigin(0,0)
	W:setSize(1.0-h_margin, W:getPreferredSize().y)

	return {
		widget = W,
		layout = vlayout
	}
end

function view:next_tab()
	local levels = #self.leaderboards
	if levels == 0 then return end

	local current_tab = self.current_tab+1
	if current_tab > levels then
		current_tab = 1
	end

	self:set_leaderboard(current_tab)
end

function view:prev_tab()
	local levels = #self.leaderboards
	if levels == 0 then return end

	local current_tab = self.current_tab-1
	if current_tab <= 0 then
		current_tab = levels
	end

	self:set_leaderboard(current_tab)
end

local active_indicator_texture = "ui/top_hustler/nav_indicator_active:filtered"
local inactive_indicator_texture = "ui/top_hustler/nav_indicator_passive:filtered"

local function build_nav_indicator(parent)
	local pic = ui.Picture()
	pic:setTexture(inactive_indicator_texture)
	pic:setSize(pic:getPreferredSize())
	return pic
end

local function build_bottom_navigation(parent)
	local W = ui.Picture()
	W:setTexture("ui/top_hustler/tophustler_nav_bottom_bg:filtered")
    W:setVerticalSizePolicy (ui.SizePolicyMinimum)
	W:setPlacingMode(ui.Picture_PlacingModeStretch)
	--W:setHorizontalSizePolicy (ui.SizePolicyFixed)
	W:setMargins(ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.hspacing,ui.config.hspacing)

	local hlayout = ui.HorizontalLayout()
	W:setLayout(hlayout)

	local left_btn = ui.ImageButton()
	left_btn:setTexture("ui/top_hustler/btn_arrow_left:filtered")
	hlayout:addWidget(left_btn)

	local widgets = {
		widget = W
	}

	do
		local vlayout = ui.VerticalLayout()
		vlayout:setHorizontalPolicy(ui.SizePolicyMaximum)

		local title = ui.TextLabel()
		title:setTextStyle(ui.text_styles.leaderboard_title.build())
		title:setText(L"KEY_LEADERBOARD_LOADING")
		title:setTextAlignment (ui.AlignmentHorizontalCenter)
		widgets.title = title

		--local tabs = ui.Widget()
		local tabs = ui.Widget()
		local tab_layout = ui.HorizontalLayout()
		tab_layout:setItemSpacing(ui.config.hspacing)

		tabs:setLayout(tab_layout)
		tabs:setHorizontalSizePolicy(ui.SizePolicyMinimum)

		local indicators = {}

		widgets.add_indicator = function()
			local indicator = build_nav_indicator(parent)
			tab_layout:addWidget(indicator)
			table.insert(indicators, indicator)
		end

		widgets.set_active_indicator = function(index)
			if index > #indicators then return end
			for i = 1, #indicators do
				local indicator = indicators[i]
				indicator:setTexture(inactive_indicator_texture)
			end

			local indicator = indicators[index]
			indicator:setTexture(active_indicator_texture)
		end

		local one_more_wrapper_layout = ui.HorizontalLayout()
		one_more_wrapper_layout:addWidget(ui.Widget())
		one_more_wrapper_layout:addWidget(tabs)
		one_more_wrapper_layout:addWidget(ui.Widget())

		vlayout:addWidget(title)
		vlayout:addItem(one_more_wrapper_layout)

		hlayout:addItem(vlayout)
	end

	local right_btn = ui.ImageButton()
	right_btn:setTexture("ui/top_hustler/btn_arrow:filtered")
	hlayout:addWidget(right_btn)

	events.connect(right_btn:getClickedSignalProxy(), function() parent:next_tab() end)
	events.connect(left_btn:getClickedSignalProxy(), function() parent:prev_tab() end)

	W:setSize (1, W:getPreferredSize().y)
	W:setOrigin (0, 0)

	return widgets
end

function view:refresh()
	if self.active_category == view.CATEGORY_ALL_TIME then
		self:set_leaderboard(self.current_tab)
	else
		self:set_friends()
	end
end

function view:clear_layout()
	self.animations:cancel("rays_rotation_animation")
	self.animations:cancel("rays_flash_animation")

	local layout = self.ui.leaderboard_layout
	while layout:getItemAt (0) do
		layout:removeItemAt (0)
	end

	local winners_layout = self.ui.winners_layout
	while winners_layout:getItemAt (0) do
		winners_layout:removeItemAt (0)
	end

	self.ui.top_wrapper:setHidden(true)
	self.ui.leaderboard:setRelativeScroll(math.vec2(0,0))
end

function view:reload_leaderboard()
	underworld2.network.leaderboards (function (leaderboards)
		if not self:valid() then return end

		self.leaderboards = leaderboards

		for i = 1, #leaderboards do
			self.ui.add_indicator()
		end

		if self.active_category == view.CATEGORY_ALL_TIME then
			self:set_leaderboard(self.current_tab)
		end
	end)
end

function view:reload_friends_with_ids(ids)
	underworld2.network.friends (#ids, ids, function (friends)
		if not self:valid() then return end

		self.friends = friends

		if self.active_category == view.CATEGORY_FRIENDS then
			self:set_friends()
		end
	end)
end

function view:reload_friends()
	if self.active_category == view.CATEGORY_FRIENDS then
		social.facebook.graph {
			method = "GET",
			path = "me/friends",
			read_permissions = { "user_friends" },
			parameters = {
			},
			callback = function(operation)
				if not self:valid() then return end
				if not operation:hasFailed() then
					local friends = operation:getLuaResponse().data
					local ids = {}

					for i = 1, #friends do
						table.insert(ids, friends[i].id)
					end

					self:reload_friends_with_ids(ids)
				else
				end
			end
		}
	end
end

function view:set_winners(data)
	local player = underworld2.player
	local player_id = player:id()
	local layout = self.ui.winners_layout

	-- winners
	layout:addWidget(ui.Widget())

	local winners = {}

	for i = 1, #data do
		if i > 3 then break end
		local row = data[i]
		local profile = row.player

		local widget = build_winner_widget(self, profile, i, row.score, (profile:id() == player_id))
		table.insert(winners, widget)
	end

	if #winners > 1 then
		winners[1], winners[2] = winners[2], winners[1]
	elseif #winners == 0 then
		player:retain()
		local widget = build_winner_widget(self, underworld2.ProfilePtr(player:get()), 1, player:getRespectPoints(), true)
		table.insert(winners, widget)
	end

	for i = 1, #winners do
		local winner = winners[i]
		layout:addWidget(winner.widget)
		for k, v in pairs(winner.animations) do
			self.animations:bind(k, v)
		end
	end

	layout:addWidget(ui.Widget())

	self.ui.top_wrapper:setHidden(false)
	self.ui.top_wrapper:setSize(self.ui.top_wrapper:getPreferredSize())
end

function view:set_friends()
	local h_margin = ui.config.hspacing*2
	local layout = self.ui.leaderboard_layout
	local data = self.friends
	local list = data.rows or {}
	local leaderboard_row = ui.widgets.leaderboard_row

	self:clear_layout()
	self:set_winners(list)

	local p_id = underworld2.player:id()
	local p_row = nil

	self.ui.leaderboard:setOrigin(0, self.ui.top_wrapper:getOrigin().y + self.ui.top_wrapper:getHeight())
	self.ui.leaderboard:setSize(self.ui.leaderboard:getWidth(), 1 - self.ui.leaderboard:getOrigin().y)

	if underworld2.social_ids.logged_in() then
		for i = 4, #list do
			local row = list[i]
			local profile = row.player

			if i == 11 then
				local wrapper = ui.Widget()
				wrapper:setMargins(h_margin,0,ui.config.vspacing/2,0)

				local line = build_line(self)

				wrapper:setSize(1.0, line:getHeight()+wrapper:getTopMargin())
				wrapper:setHorizontalSizePolicy(ui.SizePolicyFixed)
				wrapper:setVerticalSizePolicy(ui.SizePolicyFixed)

				wrapper:addChild(line)
				layout:addWidget(wrapper)
			end

			local widget = leaderboard_row.create (profile, row.rank, row.score, h_margin)
			layout:addWidget (widget)

			if profile:id() == p_id then
				p_row = widget
			end
		end

		local facebook_row = build_facebook_row(self,h_margin, self.refresh)
		layout:addWidget(facebook_row.widget)
		layout:addWidget(ui.Widget())
	else
		local facebook_row = build_facebook_row(self, h_margin, self.refresh)
		facebook_row.frame:setSize(facebook_row.frame:getWidth(), 1 - self.ui.leaderboard:getOrigin().y - ui.config.vspacing)
		facebook_row.widget:setSize(facebook_row.frame:getSize())
		layout:addWidget(facebook_row.widget)
	end

	layout:evaluate()

	if p_row then
		local max = self.ui.leaderboard:getMaxScroll().y
		local y = (p_row:getOrigin().y+(p_row:getHeight()-self.ui.leaderboard:getHeight())*0.5)/max
		self.ui.leaderboard:setRelativeScroll(math.vec2(0, y))
	end
end

function view:set_loading()
	local layout = self.ui.leaderboard_layout

	local waiting_widget = ui.WaitingWidget()
	waiting_widget:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	waiting_widget:setAlignment(ui.AlignmentHorizontalCenter)

	local hlayout = ui.HorizontalLayout()
	hlayout:addWidget(ui.Widget())
	hlayout:addWidget(waiting_widget)
	hlayout:addWidget(ui.Widget())

	local wrapper_widget = ui.Widget()
	wrapper_widget:setHorizontalSizePolicy(ui.SizePolicyMaximum)
	wrapper_widget:setVerticalSizePolicy(ui.SizePolicyMaximum)
	wrapper_widget:setLayout(hlayout)

	--layout:addWidget(ui.Widget())
	layout:addWidget(wrapper_widget)
	layout:setHorizontalPolicy(ui.SizePolicyMaximum)
	--layout:addWidget(ui.Widget())

	self.ui.leaderboard:setOrigin(0, self.ui.top_wrapper:getOrigin().y)
	self.ui.leaderboard:setSize(self.ui.leaderboard:getWidth(), 1 - self.ui.leaderboard:getOrigin().y - self.ui.bottom_navigation:getHeight())
end

function view:set_leaderboard(level)
	local h_margin = ui.config.hspacing*2
	local leaderboard_row = ui.widgets.leaderboard_row
	local data = self.leaderboards[level]
	local list = data.rows or {}
	local layout = self.ui.leaderboard_layout

	self.ui.title:setText(data.name)
	self.current_tab = level
	self:clear_layout()

	self:set_winners(list)

	self.ui.leaderboard:setOrigin(0, self.ui.top_wrapper:getOrigin().y + self.ui.top_wrapper:getHeight())
	self.ui.leaderboard:setSize(self.ui.leaderboard:getWidth(), 1 - self.ui.leaderboard:getOrigin().y - self.ui.bottom_navigation:getHeight())

	local p_id = underworld2.player:id()
	local p_row = nil

	for i = 4, #list do
		local row = list[i]
		local profile = row.player

		if i == 11 then
			local wrapper = ui.Widget()
			wrapper:setMargins(h_margin,0,ui.config.vspacing/2,0)

			local line = build_line(self)

			wrapper:setSize(1.0, line:getHeight()+wrapper:getTopMargin())
			wrapper:setHorizontalSizePolicy(ui.SizePolicyFixed)
			wrapper:setVerticalSizePolicy(ui.SizePolicyFixed)

			wrapper:addChild(line)
			layout:addWidget(wrapper)
		end

		local widget = leaderboard_row.create (profile, row.rank, row.score, h_margin)
		layout:addWidget (widget)

		if profile:id() == p_id then
			p_row = widget
		end
	end

	local facebook_row = build_facebook_row(self,h_margin, self.refresh)
	layout:addWidget(facebook_row.widget)
	layout:addWidget(ui.Widget())

	self.ui.set_active_indicator(level)

	layout:evaluate()

	if p_row then
		local max = self.ui.leaderboard:getMaxScroll().y
		local y = (p_row:getOrigin().y+(p_row:getHeight()-self.ui.leaderboard:getHeight())*0.5)/max
		self.ui.leaderboard:setRelativeScroll(math.vec2(0, y))
	end
end

function view:build_ui(...)
	self.ui = {}
	self.leaderboards = {}
	self.animations = ui.animations.factory {}
	self.current_tab = 1
	self.friends = {}

	local widget = ui.Widget()
	self.ui.widget = widget
	widget:setOrigin (0.0, 0.0)
	widget:setSize (1.0, 1.0)

	local header = build_header(self)
	self.ui.header = header
	widget:addChild(header)

	local categories = build_categories(self)
	self.ui.category_buttons = categories.buttons
	self.ui.categories = categories.widget
	widget:addChild(categories.widget)

	self.animations:bind("top_animation", ui.animations.build {
		from = -header:getHeight()-categories.widget:getHeight(),
		to = 0,
		duration = 0.3,
		subfunction = function(t)
			return (anim.bezier(0.775, 0.645, 0.610, 1.25)) (t, 1)
		end,

		step = function(self, value)
			header:setOrigin(0, value)
			categories.widget:setOrigin(0, header:getHeight() + value)
		end
	})

	local top_wrapper = ui.Widget()
	do
		local vlayout = ui.VerticalLayout()
		vlayout:setItemSpacing(ui.config.vspacing)

		top_wrapper:setLayout(vlayout)
		top_wrapper:setMargins(ui.config.hspacing*2, ui.config.hspacing*2, ui.config.vspacing, ui.config.vspacing)

		local winners = build_winners(self)
		vlayout:addWidget(winners.widget)
		self.ui.winners_layout = winners.layout

		local middle_line = build_line(self)
		vlayout:addWidget(middle_line)

		top_wrapper:setSize(1, top_wrapper:getPreferredSize().y)
		top_wrapper:setOrigin(0, header:getHeight()+categories.widget:getHeight())

		widget:addChild(top_wrapper)
	end

	self.ui.top_wrapper = top_wrapper
	self.ui.top_wrapper:setHidden(true)

	local leaderboard = build_leaderboard(self)
	widget:addChild(leaderboard.widget)
	self.ui.leaderboard = leaderboard.widget
	self.ui.leaderboard_layout = leaderboard.layout

	local bottom_navigation = build_bottom_navigation(self)

	widget:addChild(bottom_navigation.widget)

	self.ui.bottom_navigation = bottom_navigation.widget
	self.ui.title = bottom_navigation.title
	self.ui.add_indicator = bottom_navigation.add_indicator
	self.ui.set_active_indicator = bottom_navigation.set_active_indicator

	self:set_active_category(view.CATEGORY_ALL_TIME)

	self.animations:update(0)

	return widget
end

function view:update (dt)
	self.animations:update(dt)
end

function view.show(...)
	local v = view.create(...)

	ui.core.view.delegate_builder(
	{
		finished = function (view)
			ui.core.view_controller:close_modal_view (view)
		end
	}).build (v)

	ui.core.view_controller:show_modal_view (v)
end