require "ui/core/view.lua"
require "ui/config.lua"
require "ui/format.lua"
require "ui/widgets/gold_to_cash_widget.lua"
require "ui/widgets/picture_with_backlight.lua"

local view = setmetatable ({ }, { __index = ui.core.view } )
ui.vault = view

function view.show (skill_level)
  local instance = ui.core.view.builder (view).build (skill_level)
  ui.core.view.delegate_builder {
    finished = function (self)
      ui.core.view_controller:close_modal_view (self)
    end
  }.build (instance)
  ui.core.view_controller:show_modal_view (instance)
  return instance
end

function view:back()
	--self.result = view.RESULT_CANCEL
	self:finalize()
end

local function current_timespan(timestamp)
	local timespan = timestamp - underworld2.server_time()
	if timespan < 0 then timespan = 0 end
	return timespan
end

local function build_boosts_widget(boosts)
	local widget = ui.Frame()
	widget:setBordersTexture ("ui/frame_modal_light:filtered")
	widget:setMargins (ui.config.hspacing * 2, ui.config.hspacing * 2, ui.config.vspacing, ui.config.vspacing)

	local vlayout = ui.VerticalLayout()
	vlayout:setItemSpacing(ui.config.vspacing)
	widget:setLayout(vlayout)

	local widgets = {}

	for name, data in pairs(boosts) do
		local boost_widget = ui.Widget()

		local hlayout = ui.HorizontalLayout()
		hlayout:setItemSpacing(ui.config.hspacing)

		local ico_cash = ui.load_ui ("ui/cash_bonus.ui")
		local ico_cash_size = ico_cash:getPreferredSize()
		ico_cash:findChild ("ribbon"):setWidth (ico_cash_size.x)
		ico_cash:findChild ("ribbon_offset_bottom"):setHeight (ico_cash_size.y * 0.05)
		runtime.cast_to (ico_cash:findChild ("modifier"), "ui::TextLabel"):setText (data.value)
		hlayout:addWidget (ico_cash)

		local param = ui.TextLabel()
		param:setTextStyle (ui.text_styles.modal_character_name.build())
		param:setText ( data.text )
		param:setHorizontalSizePolicy(ui.SizePolicyMaximum)
		hlayout:addWidget (param)

		local ico = ui.Picture()
		ico:setTexture ("ui/ico_timer:filtered")
		hlayout:addWidget (ico)

		local param_value = ui.TextLabel()
		param_value:setTextStyle (ui.text_styles.tutorial_text.build())
		param_value:setText ( "" )
		param_value:setWordWrap (false)
		hlayout:addWidget (param_value)

		boost_widget:setLayout(hlayout)

		vlayout:addWidget(boost_widget)

		widgets[name] = {
			name_label = param,
			time_label = param_value,
			widget = boost_widget
		}
	end

	local hint = ui.RichTextWidget()
	hint:setDefaultTextStyle (ui.text_styles.tutorial_text.build())
	hint:setText (L "KEY_WALLET_BONUS_HINT")
	--hint:setDefaultVerticalAlignment (ui.AlignmentVerticalCenter)
	--hint:setDefaultHorizontalAlignment (ui.AlignmentHorizontalLeft)
	vlayout:addWidget (hint)

	widgets.container = widget

	return widgets
end

function view:update(dt)
	-- boosts
	local boosts = self.boosts
	if boosts then
		local timespan = 0

		for name, data in pairs(boosts) do
			local layout = self.ui.boosts[name]
			local duration = current_timespan(data.timestamp)
			if duration > timespan then timespan = duration end

			if duration > 0 then
				layout.time_label:setText(ui.format.duration_string(duration))
				layout.widget:setHidden(false)
			else
				layout.widget:setHidden(true)
			end
		end

		local container = self.ui.boosts.container
		if timespan > 0 then
			container:setHidden(false)
			container:setHeight(container:getPreferredSize().y)
			container:setOrigin(container:getOrigin().x, self.ui.frame:getOrigin().y - container:getHeight() - ui.config.vspacing*2)
		else
			container:setHidden(true)
		end
	end
end

function view:build_ui (skill_level)
	self.ui = {}

	local now = underworld2.server_time()

	local player = underworld2.player
	local max_money = player:getMaxMoney()
	local current_money = math.min (max_money, player:getMoney())
		
	local widget = ui.Widget()
	widget:setOrigin (0.0, 0.0)
	widget:setSize (1.0, 1.0)
	
	local character_image = ui.Picture()
	character_image:setTexture (string.format ("ui/characters/ch_accountant_modal_%d:filtered", skill_level:getId()))
	widget:addChild (character_image)

	do
		local boosts = { }

		local bonus_ptr = underworld2.player:getBonus ("CashBoostBonus")
		if bonus_ptr:get() then
			local bonus = runtime.cast_to (bonus_ptr:get(), "underworld2::ModifierBonus")
			for i, modifier in pairs (bonus:getModifiers()) do
				if now > modifier.StartTime and now < modifier.EndTime then
					local value = ui.format.modifier (modifier.Value)
					boosts[value] = { text = L("KEY_CASH_BOOST_TITLE", value), value = value, timestamp = modifier.EndTime }
				end
			end
			if next (boosts) then
				local widgets = build_boosts_widget(boosts)
				widget:addChild(widgets.container)

				self.boosts = boosts
				self.ui.boosts = widgets
			end		
		end
	end
	
	local frame = ui.Frame()
	frame:setBordersTexture ("ui/modal_wide_window:filtered")
	frame:setMargins (ui.config.hspacing * 2.0, ui.config.hspacing * 2.0, ui.config.vspacing * 2.0, ui.config.vspacing * 2.0)
	
	local frame_layout = ui.VerticalLayout()
	frame_layout:setItemSpacing (ui.config.vspacing)
	
	do
		local top_layout = ui.HorizontalLayout()
		
		local ico = ui.widgets.picture_with_backlight.create ("ui/vault/ico_vault:filtered", "ui/vault/backlight_vault:filtered")
		top_layout:addWidget (ico)
		
		local title_label = ui.TextLabel()
		title_label:setTextStyle (ui.text_styles.modal_window_title.build())
		title_label:setText (L "KEY_VAULT")
		top_layout:addWidget (title_label)
		
		local close_button = ui.ImageButton()
		close_button:setTexture ("ui/btn_close:filtered")
		close_button:setSound ("ui/UI_Back.wav")
		close_button:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		events.connect (close_button:getClickedSignalProxy(), function() self:finalize() end)
		top_layout:addWidget (close_button)
		
		frame_layout:addItem (top_layout)
	end
	
	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	local sep_height = sep:getPreferredSize().y
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)
	
	do
		local progress_bar = ui.ProgressBar()
		progress_bar:setTexture ("ui/bar_modal_cash:filtered")
		progress_bar:setValue (math.min (1.0, current_money/max_money))
		progress_bar:setTextStyle (ui.text_styles.normal_hl.build())
		progress_bar:setText (ui.format.digit (current_money))
		
		local progress_layout = ui.HorizontalLayout()
		progress_layout:addWidget (ui.Widget())
		
		local cash_ico = ui.Picture()
		cash_ico:setTexture ("ui/ico_cash:filtered")
		cash_ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		progress_layout:addWidget (cash_ico)
		
		progress_bar:setLayout (progress_layout)
		
		frame_layout:addWidget (progress_bar)
	end
	
	do
		local hlayout = ui.HorizontalLayout()
		
		hlayout:addWidget (ui.Widget())
		
		local max_label = ui.TextLabel()
		max_label:setTextStyle (ui.text_styles.normal_hl.build())
		max_label:setTextAlignment (ui.AlignmentHorizontalCenter + ui.AlignmentVerticalCenter)
		max_label:setText (L("KEY_VAULT_MAX") .. ":")
		max_label:setWordWrap (false)
		max_label:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		hlayout:addWidget (max_label)
		
		local ico = ui.Picture()
		ico:setTexture ("ui/ico_cash_small:filtered")
		ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		hlayout:addWidget (ico)
		
		local max_value_label = ui.TextLabel()
		max_value_label:setTextStyle (ui.text_styles.normal_hl.build())
		max_value_label:setTextAlignment (ui.AlignmentHorizontalCenter + ui.AlignmentVerticalCenter)
		max_value_label:setText (ui.format.digit (max_money))
		max_value_label:setWordWrap (false)
		max_value_label:setHorizontalSizePolicy (ui.SizePolicyMinimum)
		hlayout:addWidget (max_value_label)
		
		hlayout:addWidget (ui.Widget())
		
		frame_layout:addItem (hlayout)
	end
	
	do
		--local conv = underworld2.GoldConversionDatabase():get (skill_level:getId())
		
		local m = math.min (max_money, current_money) / max_money
		
		local handler = function (wallet_size, cash, gold)
			self:finalize()
      
			local player = underworld2.player
	    
			if gold > player:getGold() then
				get_sound_pool():play ("ui/SFX_Cash_Out.wav")
				ui.not_enough_gold.show (gold - player:getGold(), function()
					view.show (skill_level)
				end)
				return
			end
          
			ui.event_view.show()
			underworld2.network.fill_wallet (wallet_size, cash, gold, function (success)
				ui.event_view.close()
				if not success then
					view.show (skill_level)
				end
			end)
		end

		local cash2gold = underworld2.BalanceSettingsGoldDatabase()
	
		do
			local cash = max_money - current_money
			local gold = cash2gold:get (cash)
			local params = { value = 100, wallet_size = current_money, cash = cash, gold = gold, ico_id = "ui/vault/ico_vault:filtered", handler = handler }
			frame_layout:addWidget (ui.widgets.gold_to_cash_widget.create (params))
		end

		do
			local cash = math.min (math.floor (max_money * 0.5), max_money - current_money)
			local gold = cash2gold:get (cash)
			local params = { value = 50, wallet_size = current_money, cash = cash, gold = gold, ico_id = "ui/vault/ico_vault:filtered", handler = handler }
			frame_layout:addWidget (ui.widgets.gold_to_cash_widget.create (params))
		end

		do
			local cash = math.min (math.floor (max_money * 0.25), max_money - current_money)
			local gold = cash2gold:get (cash)
			local params = { value = 25, wallet_size = current_money, cash = cash, gold = gold, ico_id = "ui/vault/ico_vault:filtered", handler = handler }
			frame_layout:addWidget (ui.widgets.gold_to_cash_widget.create (params))
		end
	end
	
	local sep = ui.Picture()
	sep:setTexture ("ui/gold_line_large:filtered")
	sep:setHeight (sep_height)
	sep:setPlacingMode (ui.Picture_PlacingModeStretch)
	sep:setVerticalSizePolicy (ui.SizePolicyFixed)
	frame_layout:addWidget (sep)
	
	frame:setLayout (frame_layout)
	
	do
		local character_size = character_image:getPreferredSize()
		local frame_size = frame:getPreferredSize()
		
		local max_char_height = math.min (1.0 - frame_size.y, character_size.y)
		
		character_size.x = character_size.x * max_char_height / character_size.y
		character_size.y = max_char_height

		local character_image_origin = math.vec2((1.0 - character_size.x) * 0.5, 1.0 - character_size.y - frame_size.y)
		
		character_image:setOrigin (character_image_origin)
		character_image:setSize (character_size)
		character_image:setVerticalSizePolicy (ui.SizePolicyFixed)
		character_image:setHorizontalSizePolicy (ui.SizePolicyFixed)
		character_image:setPlacingMode(ui.Picture_PlacingModeStretch)

		self.ui.character_image = character_image

		frame:setOrigin (0.0, 1.0 - frame_size.y)
		frame:setSize (1.0, frame_size.y)

		self.ui.frame = frame

		-- boosts
		if self.boosts then
			local boosts = self.ui.boosts.container
			boosts:setSize(1.0 - ui.config.hspacing * 4, boosts:getPreferredSize().y)
			boosts:setOrigin(ui.config.hspacing * 2.0, 1.0 - frame_size.y - boosts:getHeight() - ui.config.vspacing*2)
		end
	end
	
	widget:addChild (frame)

	self:update(0)
	
	return widget
end

