
local controller = { }

function controller.create (params)
	local self = setmetatable ({ }, {__index = controller})

	local highlight = runtime.cast_to (ui.load_ui (params.ui), "ui::Picture")

	local hint = runtime.cast_to (highlight:findChild ("hint"), "ui::TextLabel")
	local progress = runtime.cast_to (highlight:findChild ("progress"), "ui::CashBar")
	local icon = runtime.cast_to (progress:findChild ("icon"), "ui::Picture")
	local text = runtime.cast_to (progress:findChild ("text"), "ui::TextLabel")
	local more = runtime.cast_to (progress:findChild ("more"), "ui::ImageButton")
	local more_hint = runtime.cast_to (more:findChild ("more_hint"), "ui::TextLabel")

	local ribbon = runtime.cast_to (icon:findChild ("ribbon"), "ui::Picture")
	local ribbon_offset = icon:findChild ("ribbon_offset_bottom")
	local modifier = runtime.cast_to (ribbon:findChild ("modifier"), "ui::TextLabel")

	local icon_size = icon:getImageSize()
	ribbon:setWidth (icon_size.x)
	ribbon_offset:setHeight (icon_size.y * 0.05)

	self.ui = 
	{
		highlight = highlight,
		hint = hint,
		progress = progress,
		icon = icon,
		text = text,
		more = more,
		more_hint = more_hint,
		modifier = modifier,
		ribbon = ribbon,
	}

	self.progress_limit = (params.width - more:getPreferredSize().x * 0.75) / params.width

	highlight:setWidth (params.width)

	self.display_value = params.display_value
	self.anim_queue = anim.queue()
	self.current_value = 0
	self.max_value = 0
	self.current_modifier = 1.0
	self.bonus_type = params.bonus_type
	self.modifier = 0.0

	return self
end

local highlight_keyframes = 
{
	t 		= {0.0, 0.5, 1.0},
	opacity = {0.0, 1.0, 0.0}
}

local icon_keyframes =
{
	t = { 0.0, 0.2, 0.5, 1.0 },
	scale = { 1.0, 1.0, 1.3, 1.0 }
}

local progress_keyframes = 
{
	t = { 0.0, 0.2, 0.8, 1.0},
	brightness = { 1.0, 2.0, 2.0, 1.0 }
}

local text_keyframes =
{
	t = { },
	brightness = { }
}

function controller:set_hint (text)
	self.ui.hint:setText (text)
end

function controller:set_more_hint (text)
	self.ui.more_hint:setText (text)
end

function controller:on_modifier_changed (modifier)
	self.modifier = modifier

	if modifier == 0.0 then
		self.ui.ribbon:setHidden (true)
		return
	end

	local anim_highlight = math.cspline (highlight_keyframes.t, highlight_keyframes.opacity)
	local anim_icon = math.cspline (icon_keyframes.t, icon_keyframes.scale)
	local anim = anim.lerp 
	{
		from = 0.0, to = 1.0, time = 1.5,
		object = self,
		property = function (self, t)
			self.ui.highlight:setColor (render.color (255, 255, 255, math.clamp (255 * anim_highlight (t), 0, 255)))
			self.ui.icon:setScale (anim_icon (t))
		end,
		on_finished = function()
			self.ui.ribbon:setHidden (false)
			self.ui.modifier:setText (ui.format.modifier (modifier + 1.0))
		end 
	}
	self.anim_queue:push (anim)
end

function controller:on_progress_changed (new_value, max_value)
	if (self.current_value == new_value) and (max_value == self.max_value) then
		return
	end

	local target_value = new_value
	local current_value = self.current_value
	
	self.current_value = target_value
	self.max_value = max_value

	local anim_progress = math.cspline (progress_keyframes.t, progress_keyframes.brightness)
	local anim = anim.lerp
	{
		from = 0, to = 1, time = 1.0,
		object = self,
		property = function (self, t)
			local value = math.lerp (current_value, target_value, t)
			local text = self.display_value (value, max_value)
			self.ui.progress:setValue ((value / max_value) * self.progress_limit)
			self.ui.progress:setBrightness (anim_progress (t))
			self.ui.text:setText (text)
			self.ui.text:setZoom (1 + math.sin (t * math.pi * 4) * 0.2)
		end,
	}
	self.anim_queue:push (anim)
end

function controller:update (dt)
	self.anim_queue:update (dt)

	local player = underworld2.player
	local now = underworld2.server_time()
	
	local bonus = player:getBonus (self.bonus_type)
	local modifier = 0.0
	local end_time
	local modifier_active = false
	if bonus:get() then
		local modifier_bonus = runtime.cast_to (bonus:get(), "underworld2::ModifierBonus")
		modifier = modifier_bonus:getCurrentModifierValue (now)
		modifier_active = modifier > 0.0
		end_time = modifier_bonus:getClosestModifierEndTime (now)
	end
	if self.modifier ~= modifier then
		self.modifier_end_time = end_time
		self:on_modifier_changed (modifier)
	elseif modifier_active and self.modifier_end_time ~= end_time then
		self.modifier_end_time = end_time
		self:on_modifier_changed (modifier)
	end
end

ui.widgets.animated_stat_bar_controller = controller

