require "ui/config.lua"
require "ui/text_styles.lua"
require "underworld2/couriers.lua"

local wrapper =
{
	STATE_AVAILABLE = 0,
	STATE_DELIVERY = 1,
	STATE_COMPLETE = 2,
	STATE_LOCKED = 3
}

local courier_icon_map = underworld2.couriers.avatars

local courier_rank_map =
{
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
}

local delivery_widget_width = 0.27
local button_size = 0.25

local courier_name_map = underworld2.couriers.names

local function create_delivery_widget (courier)
	local widget = ui.Widget()

	local vlayout = ui.VerticalLayout()

	do
		local hlayout = ui.HorizontalLayout()
		hlayout:setItemSpacing (ui.config.hspacing)

		hlayout:addWidget (ui.Widget())

		local commodity_id = courier:getCommodityId()


		local commodity_ico = ui.Picture ()
		hlayout:addWidget (commodity_ico)

		local items_label = ui.TextLabel()
		items_label:setTextStyle (ui.text_styles.small_hl.build())
		items_label:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalLeft)
		items_label:setWordWrap (false)

		hlayout:addWidget (items_label)

		if courier:getState() == underworld2.CourierStatePurchasing then
			local commodity = underworld2.CommoditiesDatabase():get (commodity_id)
			local commodity_name = underworld2.StringsDatabase():get (commodity:getNameKey())
			commodity_ico:setTexture (ui.config.commodity_small_icons[commodity_id])
			items_label:setText (string.format ("%d %s", courier:getItemsAmount(), commodity_name))
		else
			local total_cash = courier:getMoney()
			commodity_ico:setTexture ("ui/ico_cash_tiny:filtered")
			items_label:setText (ui.format.digit (total_cash) .. " Cash")
		end

		hlayout:addWidget (ui.Widget())

		vlayout:addItem (hlayout)
	end

	local progress_bar = ui.ProgressBar()
	progress_bar:setTexture ("ui/bar_delivery_side:filtered")
	progress_bar:setTextStyle (ui.text_styles.small_hl.build())
	vlayout:addWidget (progress_bar)

	progress_bar:setHorizontalSizePolicy(ui.SizePolicyFixed)
	local size = progress_bar:getSize()
	size.x = 0.35
	progress_bar:setSize(size)

	local w = { }
	w.widget = widget
	w.update = function (courier, server_time)
		local time_left = math.max (0, courier:getEndTime() - server_time)
		progress_bar:setValue (courier:getProgressValue (server_time))
		if (time_left == 0) then
			progress_bar:setText (L "KEY_COURIER_DONE")
		else
			progress_bar:setText (ui.format.duration_string (time_left))
		end
	end

	widget:setLayout (vlayout)

	return w
end

function wrapper.create (courier, skill_level, handler, index, dialog)

	local frame = ui.Frame()
	frame:setBordersTexture ("ui/frame_modal_light:filtered")

	local margins = { x = frame:getLeftMargin() * 0.5, y = frame:getTopMargin() * 0.5 }
	frame:setMargins (margins.x, margins.x, margins.y, margins.y)

	local frame_layout = ui.HorizontalLayout()
	frame_layout:setItemSpacing (margins.x)

	local courier_ico = ui.Picture()
	courier_ico:setTexture (courier_icon_map[index])
	courier_ico:setHorizontalSizePolicy (ui.SizePolicyMinimum)
	frame_layout:addWidget (courier_ico)

	do
		local vlayout = ui.VerticalLayout()
		vlayout:addWidget(ui.Widget())
		local courier_name_label = ui.TextLabel()
		courier_name_label:setTextStyle (ui.text_styles.normal_hl.build())
		courier_name_label:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalLeft)
		courier_name_label:setText (courier_name_map[index])
		vlayout:addWidget (courier_name_label)

		vlayout:addWidget(ui.Widget())
		frame_layout:addItem (vlayout)
	end
	frame:setVerticalSizePolicy (ui.SizePolicyMinimum)
	frame:setLayout (frame_layout)

	local size = frame:getPreferredSize()
	
	local placeholder = ui.Widget()
	placeholder:setSize(size)
	placeholder:setVerticalSizePolicy(ui.SizePolicyFixed)

	local ribbon = ui.Picture()
	ribbon:setTexture("ui/ribbon_new_character:filtered")

	placeholder:addChild(frame)

	local obj = setmetatable (
	{
		ico     = courier_ico,
		frame_layout = frame_layout,
		frame = frame,
		widget  = placeholder,
		courier = courier,
		handler = handler,
		parent  = dialog,
		skill_level = skill_level,
		layout_passed = false,
		ribbon = ribbon
	}, { __index = wrapper })

	if not courier:get() then
		obj:set_state (wrapper.STATE_LOCKED)
	else
		obj.courier_state_conn_id = events.connect (underworld2.player:getCourierStateChangedSignalProxy(), function (courier)
			if courier:getId() == obj.courier:getId() then
				obj:on_courier_state_changed()
			end
		end)
		obj:on_courier_state_changed()
	end



	return obj, size.x, size.y
end

function wrapper:on_courier_state_changed()
	local courier = self.courier
	if courier:getState() == underworld2.CourierStateNotOccupied then
		self:set_state (wrapper.STATE_AVAILABLE)
	else
		local now = underworld2.server_time()
		if now >= courier:getEndTime() then
			self:set_state (wrapper.STATE_COMPLETE)
		else
			self:set_state (wrapper.STATE_DELIVERY)
		end
	end
end

function wrapper:update (server_time)
	if not self.layout_passed then
		local size = self.widget:getSize()
		self.frame:setSize(size)

		local ribbon = self.ribbon

		local ribbon_size = ribbon:getPreferredSize()

		local virtual_size_scale = ribbon_size.x / ribbon_size.y

		ribbon_size.y = size.y * 0.7
		ribbon_size.x = ribbon_size.y * virtual_size_scale

		local x_offset = 3 / 73 * ribbon_size.x
		local y_offset = 3 / 73 * ribbon_size.y

		ribbon:setSize(ribbon_size)
		ribbon:setOrigin(math.vec2(-x_offset, size.y - ribbon_size.y + y_offset))

		ribbon:setPlacingMode(ui.Picture_PlacingModeStretch)

		local tag_label = ui.TextLabel()

		local text_x_offset = ribbon_size.x * 9 / 73
		local text_y_offset = ribbon_size.y * 31 / 74

		tag_label:setTextStyle (ui.text_styles.courier_tag.build())
		tag_label:setWordWrap (false)
		tag_label:setText (L "KEY_COURIER_NEW")
		tag_label:setHidden(false)
		tag_label:setRotation (-math.pi * 0.25)

		local tag_label_size = tag_label:getPreferredSize()
		tag_label:setOrigin (text_x_offset, text_y_offset)
		tag_label:setSize (tag_label_size)


		ribbon:addChild (tag_label)

		self.widget:addChild(ribbon)

		self.layout_passed = true
	end

	if self.state == wrapper.STATE_DELIVERY then
		local t = self.courier:getProgressValue (server_time)

		self.delivery_progress.update (self.courier, server_time)
		self.now_button:setPrice (self.courier:getBoostPrice (t))
		if self.courier:getProgressValue (server_time) == 1.0 then
			self:set_state (wrapper.STATE_COMPLETE)
		end
	end
end

function wrapper:set_state (state)
	if self.state_cleaner then
		self.state_cleaner (self)
		self.state_cleaner = nil
	end

	self.state = state

	self.ico:setCustomEffect("")
	self.ribbon:setHidden(true)

	if state == wrapper.STATE_AVAILABLE then
		local spacer = ui.Widget()
		local state_label = ui.TextLabel()
		state_label:setTextStyle (ui.text_styles.small_hl.build())
		state_label:setText (L "KEY_COURIER_AVAILABLE")
		state_label:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalCenter)

		self.frame_layout:addWidget (spacer)
		self.frame_layout:addWidget (state_label)
		self.state_cleaner = function (self)
			self.frame_layout:removeWidget (spacer)
			self.frame_layout:removeWidget (state_label)
		end
	elseif state == wrapper.STATE_LOCKED then
		local spacer = ui.Widget()
		self.frame_layout:addWidget (spacer)
		local state_label, hire_button

		local unlock_level = self.skill_level:getLevelRequired()
		local available = underworld2.player:getLevel() >= unlock_level

		if not available then
			state_label = ui.TextLabel()
			state_label:setTextStyle (ui.text_styles.small_hl.build())
			state_label:setText (L ("KEY_COURIER_REACH_LEVEL", unlock_level))
			state_label:setTextAlignment (ui.AlignmentVerticalCenter + ui.AlignmentHorizontalCenter)
				self.frame_layout:addWidget (state_label)
		else
			local colors = ui.button_colors.green

			hire_button = ui.UniversalButton()
			hire_button:setText (L "KEY_COURIER_HIRE")
			hire_button:setTextStyle (ui.text_styles.button_regular.build())
			hire_button:setTexture ("ui/equipment/btn_64_universal:filtered")
			hire_button:setStretch (ui.UniversalButton_StretchHorizontal + ui.UniversalButton_StretchVertical)
			hire_button:setPrimaryColor (ui.Button_ButtonStateNormal, colors.primary0)
			hire_button:setSecondaryColor (ui.Button_ButtonStateNormal, colors.secondary0)
			hire_button:setPrimaryColor (ui.Button_ButtonStatePressed, colors.primary1)
			hire_button:setSecondaryColor (ui.Button_ButtonStatePressed, colors.secondary1)

			hire_button:setSize(button_size,0)
			hire_button:setHorizontalSizePolicy(ui.SizePolicyFixed)

			events.connect (hire_button:getClickedSignalProxy(), function() self:handler(true) end)

			self.frame_layout:addWidget (hire_button)
		end

		self.state_cleaner = function (self)
			self.frame_layout:removeWidget (spacer)
			if not available then
				self.frame_layout:removeWidget (state_label)
			else
				self.frame_layout:removeWidget (hire_button)
			end
		end

		self.ico:setCustomEffect("grayscale")

		self.ribbon:setHidden(not available)
	elseif state == wrapper.STATE_DELIVERY then
		local now = underworld2.server_time()
		local delivery_progress = create_delivery_widget (self.courier)
		delivery_progress.update (self.courier, now)

		local now_button = ui.PriceButton()
		now_button:setStateTexture ("ui/btn_premium_side:filtered")
		now_button:setTextStyle (ui.text_styles.button_regular.build())
		now_button:setCurrencyIcon ("ui/ico_pbtn_gold:filtered")
		now_button:setPrice (self.courier:getBoostPrice (now))
		now_button:setText (L "KEY_COURIER_BOOST")
		events.connect (now_button:getClickedSignalProxy(), function() self.handler (self) end)

		self.delivery_progress = delivery_progress
		self.now_button = now_button

		now_button:setSize(button_size,0)
		now_button:setHorizontalSizePolicy(ui.SizePolicyFixed)

		self.frame_layout:addWidget (delivery_progress.widget)
		self.frame_layout:addWidget (now_button)
		self.state_cleaner = function (self)
			self.frame_layout:removeWidget (delivery_progress.widget)
			self.frame_layout:removeWidget (now_button)
		end
	elseif state == wrapper.STATE_COMPLETE then
		local delivery_progress = create_delivery_widget (self.courier)
		delivery_progress.update (self.courier, self.courier:getEndTime())

		local colors = ui.button_colors.green

		local collect_button = ui.UniversalButton()
		collect_button:setText (L "KEY_COURIER_COLLECT")
		collect_button:setTextStyle (ui.text_styles.button_regular.build())
		collect_button:setTexture ("ui/equipment/btn_64_universal:filtered")
		collect_button:setStretch (ui.UniversalButton_StretchHorizontal + ui.UniversalButton_StretchVertical)
		collect_button:setPrimaryColor (ui.Button_ButtonStateNormal, colors.primary0)
		collect_button:setSecondaryColor (ui.Button_ButtonStateNormal, colors.secondary0)
		collect_button:setPrimaryColor (ui.Button_ButtonStatePressed, colors.primary1)
		collect_button:setSecondaryColor (ui.Button_ButtonStatePressed, colors.secondary1)

		collect_button:setSize(button_size,0)
		collect_button:setHorizontalSizePolicy(ui.SizePolicyFixed)

		events.connect (collect_button:getClickedSignalProxy(), function() self.handler (self) end)

		self.frame_layout:addWidget (delivery_progress.widget)
		self.frame_layout:addWidget (collect_button)
		self.state_cleaner = function (self)
			self.frame_layout:removeWidget (delivery_progress.widget)
			self.frame_layout:removeWidget (collect_button)
		end
	end
end

function wrapper:destroy()
	if self.courier_state_conn_id then
		events.disconnect (underworld2.player:getCourierStateChangedSignalProxy(), self.courier_state_conn_id)
		self.courier_state_conn_id = nil
	end
end

ui.widgets.courier_widget_wrapper = wrapper
