require "lib/signal.lua"
require "ui/core/library.lua"
local lui = ui.core.lib

require "ui/widgets/quests/task_bring_commodities_controller.lua"
require "ui/widgets/quests/task_purchase_commodities_controller.lua"
require "ui/widgets/quests/task_sell_commodities_controller.lua"
require "ui/widgets/quests/task_earn_cash_controller.lua"

local controller = 
{
	MODE_FIRST_TIME_SHOW = 1,
	MODE_IN_PROGRESS = 2,
	MODE_COMPLETED = 3
}

local task_builders = 
{
	["underworld2::quests::BringCommoditiesTask"] 		= ui.task_bring_commodities_controller,
	["underworld2::quests::PurchaseCommoditiesTask"] 	= ui.task_purchase_commodities_controller,
	["underworld2::quests::SellCommoditiesTask"] 		= ui.task_sell_commodities_controller,
	["underworld2::quests::EarnCashTask"] 				= ui.task_earn_cash_controller,
}

local frame_bounce_keyframes =
{
	t = { 0.0, 0.4, 0.7, 0.85, 1.0 },
	y = { 0.0, 0.3, 1.0, 0.98, 1.0 }
}

local item_bounce_keyframes =
{
	t = { 0.0, 0.4, 0.7, 0.8, 1.0 },
	y = { 0.0, 0.3, 1.0, 0.7, 1.0 }
}

local function __load_tasks (quest)
	local ret = { }
	local tasks = quest:getTasks()
	for task_id, task in pairs (tasks) do
		local builder = task_builders[task:getClass():getScopedName()]
		if builder then
			local controller = builder.create (task)
			table.insert (ret, controller)
		end
	end
	return ret
end

local function __set_state (self, state)
	local quest = self.quest
	local story = quest:getStory():get()
	local text = state == controller.MODE_COMPLETED and story:getQuestCompletedText() or story:getQuestBriefText()
	self.ui.story:setText (text)
	self.state = state
end

local function __build_hide_tasks_animation (self)
	local height = self.ui.tasks:getPreferredSize().y
	local tasks = self.task_controllers
	return anim.lerp
	{
		from = height, to = 0, time = 0.2,
		object = self,
		property = function (self, t)
			self.ui.tasks:setHeight (t)
		end
	}
end

local function __build_item_drop_animation (self, index, item)
	if not item then return nil end
	local bounce = math.cspline (item_bounce_keyframes.t, item_bounce_keyframes.y)
	local height = item.ui.root:getHeight()
	return anim.lerp
	{
		from = 0, to = 1, time = 0.3,
		object = self,
		property = function (self, t)
			item.ui.root:setOffset (math.vec2 (0, -height * (index - 1) * bounce (1 - t)))
		end
	}
end

local function __build_show_tasks_animation (self)
	local height = self.ui.tasks:getPreferredSize().y
	local tasks = self.task_controllers
	local bounce = math.cspline (frame_bounce_keyframes.t, frame_bounce_keyframes.y)
	
	return anim.parallel
	{
		anim.lerp
		{
			from = 0, to = 1, time = 0.2,
			object = self,
			property = function (self, t)
				local y = bounce (t) * height
				self.ui.tasks:setHeight (y)
			end,
			on_finished = function() __set_state (self, controller.MODE_IN_PROGRESS) end
		},
		__build_item_drop_animation (self, 1, tasks[#tasks]),
		__build_item_drop_animation (self, 2, tasks[#tasks - 1]),
		__build_item_drop_animation (self, 3, tasks[#tasks - 2]),
	}
end

local function __build_show_collect_animation (self)
	local height = self.ui.collect_pane:getPreferredSize().y
	return anim.sequence
	{
		anim.delay
		{
			time = 0.0,
			on_finished = function()
				self.ui.collect:setHidden (true)
				self.ui.collect:setScale (0.6)
			end
		},
		anim.lerp
		{
			from = 0, to = height, time = 0.1,
			object = self,
			property = function (self, t)
				self.ui.collect_pane:setHeight (t)
			end,
			on_finished = function ()
				self.ui.collect:setHidden (false)
			end
		},
		anim.lerp
		{
			from = 0.6, to = 1.1, time = 0.1,
			object = self, 
			property = function (self, t) self.ui.collect:setScale (t) end
		},
		anim.lerp
		{
			from = 1.1, to = 1.0, time = 0.1,
			object = self,
			property = function (self, t) self.ui.collect:setScale (t) end,
			on_finished = function() __set_state (self, controller.MODE_COMPLETED) end
		}
	}
end

local function __build_hide_collect_animation (self)
	local height = self.ui.collect_pane:getPreferredSize().y
	return anim.sequence
	{
		anim.lerp
		{
			from = 1.0, to = 1.1, time = 0.1,
			object = self, 
			property = function (self, t) self.ui.collect:setScale (t) end
		},
		anim.lerp
		{
			from = 1.1, to = 0.6, time = 0.1,
			object = self,
			property = function (self, t) self.ui.collect:setScale (t) end,
			on_finished = function()
				self.ui.collect:setHidden (true)
			end
		},
		anim.lerp
		{
			from = height, to = 0, time = 0.1,
			object = self,
			property = function (self, t)
				self.ui.collect_pane:setHeight (t)
			end,
			on_finished = function ()
				self.ui.collect:setHidden (false)
			end
		}
	}
end

local function __build_show_continue_animation (self)
	self.ui.continue:setHidden (false)
	local height = self.ui.continue_pane:getPreferredSize().y
	return anim.sequence
	{
		--prepare
		anim.delay
		{
			time = 0.0,
			on_finished = function()
				self.ui.continue:setHidden (true)
				self.ui.continue:setScale (0.6)
			end
		},
		--animate details drop/show button
		anim.lerp
		{
			from = 0, to = height, time = 0.1,
			object = self,
			property = function (self, t)
				self.ui.continue_pane:setHeight (t)
			end,
			on_finished = function ()
				self.ui.continue:setHidden (false)
			end
		},
		-- button zoom in
		anim.lerp
		{
			from = 0.6, to = 1.1, time = 0.1,
			object = self, 
			property = function (self, t) self.ui.continue:setScale (t) end
		},
		-- button zoom out
		anim.lerp
		{
			from = 1.1, to = 1.0, time = 0.1,
			object = self,
			property = function (self, t) self.ui.continue:setScale (t) end,
			on_finished = function() __set_state (self, controller.MODE_FIRST_TIME_SHOW) end
		}
	}
end

local function __build_hide_continue_animation (self)
	local height = self.ui.continue_pane:getPreferredSize().y
	return anim.sequence
	{
		-- button zoom in
		anim.lerp
		{
			from = 1.0, to = 1.1, time = 0.1,
			object = self, 
			property = function (self, t) self.ui.continue:setScale (t) end
		},
		-- button zoom out/hide
		anim.lerp
		{
			from = 1.1, to = 0.6, time = 0.1,
			object = self,
			property = function (self, t) self.ui.continue:setScale (t) end,
			on_finished = function()
				self.ui.continue:setHidden (true)
			end
		},
		-- hide details
		anim.lerp
		{
			from = height, to = 0, time = 0.1,
			object = self,
			property = function (self, t)
				self.ui.continue_pane:setHeight (t)
			end,
		}
	}
end

function controller.create (quest)
	local self = setmetatable ({ }, { __index = controller } )
	self.geometry_update_required = signal()

	local widget = ui.load_ui ("ui/quests/quest_state.ui")
	self.ui = ui.build_children_map (widget)

	self.ptr = quest
	self.quest = quest:get()

	local story = self.quest:getStory():get()

	self.ui.character:setTexture (ui.config.quests.characters.get_resource_id ("modal", story:getCharacterId()))
	self.ui.character_name:setText (story:getCharacterName())

	self.task_controllers = __load_tasks (self.quest)

	for i, v in ipairs (self.task_controllers) do
		self.ui.tasks:addChild (v.ui.root)
	end

	__set_state (self, controller.get_quest_state (quest))

	self.continue = signal()
	self.collect = signal()

	events.connect (self.ui.continue:getClickedSignalProxy(), function() self.continue:notify (self) end)
	events.connect (self.ui.collect:getClickedSignalProxy(), function() self.collect:notify (self) end)

	return self
end

function controller:build_hide_details_animation()
	if self.state == controller.MODE_FIRST_TIME_SHOW then
		return __build_hide_continue_animation (self)
	elseif self.state == controller.MODE_IN_PROGRESS then
		return __build_hide_tasks_animation (self)
	elseif self.state == controller.MODE_COMPLETED then
		return __build_hide_collect_animation (self)
	end
end

function controller:build_show_character_animation()
end

function controller:build_hide_character_animation()
end

local anim_transitions = 
{
	[controller.MODE_FIRST_TIME_SHOW] = 
	{
		[controller.MODE_FIRST_TIME_SHOW] = function (self)
			return anim.sequence
			{
				anim.delay 
				{ 
					time = 0.0, 
					on_finished = function ()  
						__set_state (self, controller.MODE_FIRST_TIME_SHOW)
						self:prepare_to_show () 
					end 
				},
				__build_show_continue_animation (self)
			}
		end,

		[controller.MODE_IN_PROGRESS] = function (self)
			return anim.sequence
			{
				__build_hide_continue_animation (self),
				anim.delay 
				{ 
					time = 0.0, 
					on_finished = function ()  
						__set_state (self, controller.MODE_IN_PROGRESS)
						self:prepare_to_show () 
					end 
				},
				__build_show_tasks_animation (self),
			}
		end
	},
	[controller.MODE_IN_PROGRESS] = 
	{
		[controller.MODE_IN_PROGRESS] = function (self)
			return anim.sequence
			{
				anim.delay 
				{ 
					time = 0.0, 
					on_finished = function ()  
						__set_state (self, controller.MODE_IN_PROGRESS)
						self:prepare_to_show () 
					end 
				},
				__build_show_tasks_animation (self)
			}
		end,

		[controller.MODE_COMPLETED] = function (self)
			return anim.sequence
			{
				__build_hide_tasks_animation (self),
				anim.delay 
				{ 
					time = 0.0, 
					on_finished = function ()  
						__set_state (self, controller.MODE_COMPLETED)
						self:prepare_to_show () 
					end 
				},
				__build_show_collect_animation (self),
			}
		end
	},
	[controller.MODE_COMPLETED] =
	{
		[controller.MODE_COMPLETED] = function (self)
			return anim.sequence
			{
				anim.delay 
				{ 
					time = 0.0, 
					on_finished = function ()  
						__set_state (self, controller.MODE_COMPLETED)
						self:prepare_to_show () 
					end 
				},
				__build_show_collect_animation (self)
			}
		end
	}
}

function controller.get_quest_state (ptr)
	local quest = ptr:get()
	if not quest:isActivated() then return controller.MODE_FIRST_TIME_SHOW
	elseif not quest:isCompleted() then return controller.MODE_IN_PROGRESS
	end
	return controller.MODE_COMPLETED
end

function controller:build_state_transition_animation (from, to)
	to = to or from
	local transition = anim_transitions[from][to] (self)
	return transition
end

function controller:update_tasks ()
	for i, v in ipairs (self.task_controllers) do
		v:on_task_state_updated ()
	end
end

local CHARACTER_MAX_OFFSET_Y = 0.02

local character_anim_keyframes = 
{
	t = { 0.0, 0.6, 1.0},
	y = { 0.0, 0.95, 1.0}
}

function controller:build_character_animation ()
	local character_anim = math.cspline (character_anim_keyframes.t, character_anim_keyframes.y)
	return anim.lerp
	{
		from = 1, to = 0, time = 0.2,
		object = self,
		property = function (self, t)
			self.character_offset_y = character_anim (t) * CHARACTER_MAX_OFFSET_Y
			self.ui.character:setOffset (math.vec2 (0, self.character_offset_y))
		end
	}
end

function controller:prepare_to_show (switch_from_another)
	self.ui.continue_pane:setHidden (self.state ~= controller.MODE_FIRST_TIME_SHOW)
	self.ui.collect_pane:setHidden (self.state ~= controller.MODE_COMPLETED)
	self.ui.tasks:setHidden (self.state ~= controller.MODE_IN_PROGRESS)

	__set_state (self, controller.get_quest_state (self.ptr))

	if switch_from_another then
		--todo: set character
		self.character_offset_y = CHARACTER_MAX_OFFSET_Y
		self.ui.character:setOffset (math.vec2 (0, self.character_offset_y))
	end
end

ui.quest_state_controller = controller
