local M = {
	delivering = {},
	session = {}
}

local file_path = "/public/ui/equipment_view_state.lua"
-- Just to create the index for latter checks
local current

local function generate_state(player)
	current = {
		delivering = {},
		slots = {},
		stats = {},
		session = {slots = {}, skills = {}}
	}
end

function M.load_state(player)
	if not current then
		if file.exists(file_path) and file.is_empty(file_path) then
			current = dofile(file_path)
		else
			generate_state(player)
		end
	end

	return current
end

function M.flush_state()
	if current then
		file.write_string(file_path, table.dump(current))
	end
end

function M.session.slot(slot_id)
	if not current then M.load_state(player) end

	local slots = current.session.slots
	for i = 1, #slots do
		if slots[i] == slot_id then
			return i
		end
	end

	return 0
end

function M.session.skill(level)
	if not current then M.load_state(player) end

	local skills = current.session.skills
	for i = 1, #skills do
		if skills[i] == level then
			return i
		end
	end

	return 0
end

function M.session.clear(category)
	if not current then M.load_state(player) end

	current.session[category] = {}
end

function M.session.unlock_slots_for_level(level_id)
	if not current then M.load_state(player) end

	local db = underworld2.SlotsDatabase()
	local slots = current.session.slots

	for i = 1, 8 do
		local slot = db:get (i)
		if slot:getUnlockLevel() == level_id then
			table.insert(slots, i)
		end
	end
end

function M.session.unlock_skill_level(level)
	if not current then M.load_state(player) end

	local db = underworld2.SlotsDatabase()
	local skills = current.session.skills

	table.insert(skills, level)
end

function M.delivering.reset_delivered(player, slots)
	if not current then M.load_state(player) end

	for slot_id, widget in pairs(slots) do
		local item_id = M.delivering.get_item_id(player, slot_id)
		if item_id then
			local current_time = underworld2.server_time()
			local item_delivery_time = player:getItemDeliveryTime(item_id)
			if current_time >= item_delivery_time then
				current.delivering[slot_id] = nil
				player:setItemDelivered(item_id)
				item_id = -1
			end
		end

		if widget then
   			widget:setDeliveringItem(item_id)
		end
   	end

	M.flush_state()
end

function M.delivering.update_slots(player, slots)
	if not current then M.load_state(player) end

	local assigned = {}

	for slot_id, widget in pairs(slots) do
		assigned[widget:getDeliveringItemID()] = slot_id
	end

	local delivering = {}
	local delivering_native = player:getPendingItems()
	local items_db = underworld2.ItemsDatabase()
	local current_time = underworld2.server_time()

	for i = 0, delivering_native:size() - 1 do
   		local item_id = delivering_native[i]
		local delivery_time = player:getItemDeliveryTime(item_id)

   		if (not assigned[item_id]) and (current_time <= delivery_time) then
   			local item = items_db:getItem(item_id)
   			local slot_type = item:getSlotType()
   			table.insert(delivering, { 
   				item_id = item_id, 
   				slot_type = slot_type, 
   				time_left = current_time - delivery_time
   			})
   		end
   	end

   	table.sort(delivering, function(a,b) return a.time_left < b.time_left end)

   	local slots_db = underworld2.SlotsDatabase()

   	for slot_id, widget in pairs(slots) do
   		if widget:getDeliveringItemID() < 0 then
   			-- try to assign an item
   			local slot = slots_db:get(slot_id)
   			local slot_type = slot:getType()

   			for i,v in ipairs(delivering) do
   				if v.slot_type == slot_type and not assigned[v.item_id] then
   					current.delivering[slot_id] = v.item_id
   					assigned[v.item_id] = slot_id
   					break
   				end
   			end
   		end
   	end

   	for slot_id, widget in pairs(slots) do
		local item_id = M.delivering.get_item_id(player, slot_id)
		if assigned[item_id] == slot_id then
   			widget:setDeliveringItem(item_id)
		end
   	end

	M.flush_state()
end

function M.delivering.get_item_id(player, slot_id)
	if not current then M.load_state(player) end

	return current.delivering[slot_id] or -1
end

function M.delivering.set_item_id(player, slot_id, item_id)
	if not current then M.load_state(player) end
	current.delivering[slot_id] = item_id

	M.flush_state()
end

function M.delivering.set_best_item_id(player, slot_id, item_id)
	if not current then M.load_state(player) end
	local current_item = current.delivering[slot_id]

	if not current_item or current_item < 0 then
		current.delivering[slot_id] = item_id
		M.flush_state()
	else
		local current_time = underworld2.server_time()
		local current_item_left = player:getItemDeliveryTime(current_item) - current_time
		local new_item_left = player:getItemDeliveryTime(item_id) - current_time

		if new_item_left < current_item_left then
			current.delivering[slot_id] = item_id
			M.flush_state()
		end
	end
end

function M.delivering.mark_item_boosted(player, item_id)
	M.load_state(player)

	for i,v in pairs(current.delivering) do
		if v == item_id then
			current.delivering[i] = nil
			M.flush_state()
			return
		end
	end
end

underworld2.equipment_view_state = M
return M