require "social/jabber.lua"
require "underworld2/blacklist.lua"

local inbox = util.delegator ( { } )
		
local _storage = {
	contacts = { },
	new_conversations = 0
}

local _storage_path = "/home/inbox.db"
local _blacklist_path = "/home/inbox_blacklist.db"
local _do_flush = false

function _push (player_id, message, when)
	local contact = _storage.contacts[player_id]
	if not contact then
		_storage.contacts[player_id] = { pending = 0 }
		contact = _storage.contacts[player_id]
	end

	contact.last_msg = message
	contact.when = when
	contact.pending = contact.pending + 1

	inbox:notify ("conversation_updated", player_id, contact)

	_do_flush = true
end

function _send (player_id, message, when)
	_push (player_id, message, when)
end

function _receive (player_id, message, when)
	_push (player_id, message, when)

	_storage.new_conversations = _storage.new_conversations + 1
end

function inbox.init()
	if inbox.initialized then
		return
	end

	if file.exists (_storage_path) then
		_storage = dofile (_storage_path)
	end

	inbox.blacklist = underworld2.blacklist:new()
	inbox.blacklist:load(_blacklist_path)

	social.jabber:add_delegate (inbox)
	inbox.initialized = true
end

function inbox.add_to_blacklist(player_id)
	inbox.blacklist:set(player_id, true)
	inbox.blacklist:save(_blacklist_path)
end

function inbox.remove_from_blacklist(player_id)
	inbox.blacklist:set(player_id, nil)
	inbox.blacklist:save(_blacklist_path)
end

function inbox.is_blocked(player_id)
	return inbox.blacklist:get(player_id)
end

function inbox.blocked_contacts()
	return inbox.blacklist:all()
end

function inbox.finalize()
	if not inbox.initialized then return end

	local dump = table.dump (_storage)
	file.write_string (_storage_path, dump)

	_storage.contacts = { }
	_storage.new_conversations = 0
	_do_flush = false
	inbox.initialized = false

	--inbox.blacklist:save(_blacklist_path)

	social.jabber:remove_delegate (inbox)
	inbox:clear_delegates()
end

function inbox.clear()
	_storage.contacts = { }
	_storage.new_conversations = 0

	local dump = table.dump (_storage)
	file.write_string (_storage_path, dump)
	--inbox.blacklist:save(_blacklist_path)
	_do_flush = false
end

function inbox.mark_as_read (player_id)
	local contact = _storage.contacts[player_id]
	if contact and contact.pending > 0 then
		_storage.new_conversations = math.max (0, _storage.new_conversations - contact.pending)
		contact.pending = 0
		inbox:notify ("conversation_updated", player_id, contact)
		_do_flush = true
	end
end

function inbox.new_conversations()
	return _storage.new_conversations
end

function inbox.contacts()
	return _storage.contacts
end

function inbox.auth_state_changed(self, online)
	if not online and _do_flush then
		local dump = table.dump (_storage)
		file.write_string (_storage_path, dump)
		--inbox.blacklist:save(_blacklist_path)
		_do_flush = false
	end
end

local function parse_profile (stanza, profile_id)
	if not stanza then
		return false
	end

	local nickname = stanza.nickname
	local avatar_url = stanza.avatarUrl or ""
	local level = stanza.level or 1
	if not nickname then
		return false
	end

	local ptr = underworld2.Profile (profile_id, true)
	ptr:setNickname (nickname)
	ptr:setAvatarUrl (avatar_url)
	ptr:setLevel (tonumber (level))
	underworld2.ProfileDatabase():add (underworld2.ProfilePtr (runtime.disown (ptr)))
	
	return true
end

function inbox.stanza_received (self, stanza)
	if tostring (stanza.name) == "message" and tostring (stanza.type) == "chat" then
		local player_id = tonumber (string.sub (stanza.from, 1, string.find (stanza.from, "@", 1, true) - 1))
		if inbox.is_blocked(player_id) then return end

		if player_id then
			local profile_data = social.jabber.get_stanza_child (stanza, "profile")
			local exists = underworld2.ProfileDatabase():hasProfile (player_id)
			if not exists then
				if not parse_profile (profile_data, player_id) then
					return
				end
			end

			local msg = social.jabber.get_stanza_child (stanza, "body").children[1].text
			local delay = social.jabber.get_stanza_child (stanza, "delay")
			local when
			if delay then
				when = delay.stamp
			else
				when = social.jabber.date_time.delay_stamp()
			end
			
			-- TO DO: more elegant solution
			if (not delay) and (not inbox.silent) then
				get_sound_pool():play ("ui/SFX_Message_Get_Hard.wav")
			end

			_receive (player_id, msg, when)
		end
	end
end

function inbox.stanza_sent (self, stanza)
	if tostring (stanza.name) == "message" and tostring (stanza.type) == "chat" then
		get_sound_pool():play ("ui/message_send.wav")
		local player_id = tonumber (string.sub (stanza.to, 1, string.find (stanza.to, "@", 1, true) - 1))
		if inbox.is_blocked(player_id) then return end
		if player_id then
			local msg = social.jabber.get_stanza_child (stanza, "body").children[1].text
			_send (player_id, msg, social.jabber.date_time.delay_stamp())
		end
	end
end

underworld2.inbox = inbox
