
--[[
	category -> {
		last_visitied_level -> level serialized
	}
]]

local path = "/home/upgrades_tracker.sv"
local insert = table.insert

local tracker = {}

local _meta_default_level = { __index = function(t,k) return { last_visitied_level = 1 } end }

local categories = {}

local function merge_tables( ... )
	local args = { ... }
	local ret  = {}

	for _,arg in ipairs(args) do
		for __,value in ipairs(arg) do
			insert( ret, value )
		end
	end

	return ret
end

local function make_set( ... )
	local args = { ... }
	local ret  = {}

	for _,arg in ipairs(args) do
		for __,value in ipairs(arg) do
			ret[value] = true
		end
	end

	return ret
end


-- category -> { subcategory -> {level -> { items }} }
local upgrades_cache 

local function insert_upgrade( category, subcategory, level, id )
	category[subcategory] = category[subcategory] or {}
	local subcattable = category[subcategory]
	subcattable[level] = subcattable[level] or {}
	insert(subcattable[level], id)
end

local function build_items_cache()
	local items_db = underworld2.ItemsDatabase()
	local cpp_db_items = items_db:select (0,1000)

	for i=0,cpp_db_items:size()-1 do
		local item_id = cpp_db_items[i]
		local cached_info = items_db:getItem (item_id)
		insert_upgrade( upgrades_cache.items, cached_info:getSlotType(), cached_info:getLevelRequired(), item_id )
		categories.items[cached_info:getSlotType()] = categories.items[cached_info:getSlotType()] or {last_visitied_level = 1}
	end 	
end

local skill_ids = { "COURIERS", "ITEM", "WALLET" }

local function build_crew_cache()
	local skill_group = underworld2.SkillsDatabase():getGroup ("SMUGGLERSHIP")

	for i,s in ipairs(skill_ids) do
		local skill = skill_group:getSkill(s)

		local skill_id = skill:getId()
		
		for level_id = 1, skill:getLevelsCount() do	
			local level = skill:getLevel (level_id)
			if level:get() then
				local required_level = level:getLevelRequired() 
				insert_upgrade( upgrades_cache.skills, skill_id, required_level, level_id )
				categories.skills[skill_id] = categories.skills[skill_id] or {last_visitied_level = 1}
			end
		end	
	end
end

function tracker.init()
	categories	= {
		skills= setmetatable({}, _meta_default_level),
		items = setmetatable({}, _meta_default_level)
	}

	upgrades_cache = {
		skills = {},
		items = {}
	}

	build_items_cache()
	build_crew_cache()

	if file.exists(path) then
		local save_data = dofile(path)

		for cat_name, cat in pairs(save_data) do
			for scat_id,subcat in pairs(cat) do
				categories[cat_name] = categories[cat_name] or {}
				categories[cat_name][scat_id] = categories[cat_name][scat_id] or {}

				for key,data in pairs(subcat) do
					categories[cat_name][scat_id][key] = data
				end
			end
		end
	end
end

local function save()
	local save_data = {}
	
	for cat_name, cat in pairs(categories) do
		for scat_id,subcat in pairs(cat) do
			save_data[cat_name] = save_data[cat_name] or {}
			save_data[cat_name][scat_id] = save_data[cat_name][scat_id] or {}

			for key,data in pairs(subcat) do
				save_data[cat_name][scat_id][key] = data
			end
		end
	end

	file.write_string( path, table.dump(save_data) )
end

-- context -> {
-- 		level = player level,
--      [category = category],
--      [subcategory = subcategory]
-- }
-- Returns  a SET of new items ( i.e. id->true ), items count
function tracker.new_items( context )
	local ret = {}
	local count = 0

	for cat_name,category in pairs(categories) do
		if not context.category or (context.category == cat_name) then
			for scat_id,subcat in pairs(category) do
				if not context.subcategory or (context.subcategory == scat_id) then
					local cache = upgrades_cache[cat_name][scat_id]
					for level, ids in pairs(cache) do
						if level > subcat.last_visitied_level and level <= context.level then
							for ___,id in ipairs(ids) do 
								ret[id] = true
								count = count + 1
							end
						end
					end

					for ___,id in ipairs (subcat.freshly_delivered or {}) do
						ret[id] = true
						count = count + 1
					end
				end
			end
		end
	end

	return ret, count
end

-- context -> {
-- 		level = player level,
--      [category = category],
--      [subcategory = subcategory]
-- }
function tracker.has_new_items( context )
	local player = underworld2.player

	for cat_name,category in pairs(categories) do
		if not context.category or (context.category == cat_name) then
			for scat_id,subcat in pairs(category) do
				if not context.subcategory or (context.subcategory == scat_id) then
					local slot = underworld2.SlotsDatabase():get (scat_id)
					if (slot and slot:get()) and (slot:getUnlockLevel() <= player:getLevel()) then
						if subcat.freshly_delivered and #subcat.freshly_delivered > 0 then
							return true
						end

						local cache = upgrades_cache[cat_name][scat_id]
						if cache then
							for level, ids in pairs(cache) do
								if level > subcat.last_visitied_level and level <= context.level then
									return true
								end
							end
						end
					end
				end
			end
		end
	end

	return false
end

function tracker.add_delivered_item( category, subcategory, id )
	local cat = categories[category]
	local subcat = cat[subcategory]

	if not subcat then
		subcat = {}
		cat[subcategory] = subcat
	end

	subcat.freshly_delivered = subcat.freshly_delivered or {}
	table.insert(subcat.freshly_delivered, id)

	save()
	underworld2.ui.status_panel:on_new_items_list_updated()
end

function tracker.visited_subcategory( player, category, subcategory )
	local cat = categories[category]
	if cat then
		local subcat = cat[subcategory]
		if subcat then

			subcat.last_visitied_level = player:getLevel()
			save()
			underworld2.ui.status_panel:on_new_items_list_updated()
		end
	end
end

underworld2.upgrades_tracker = tracker